#-------------------------------------------------------------------------------
# Blender Plug-in - Author: Friedemann Seebass, Germany, 2020
# Licence: GPL V2.0.
#-------------------------------------------------------------------------------
#
# This file is a Python script to be used as add-on for the DCC software Blender
# version 2.8 and later. It enables the 3D input device "SpaceController" to
# move the scene and objects in Blender.
#
# Diese Datei ist ein Python-Skript, das als Add-on in der DCC-Software Blender
# verwendet werden soll. Es erlaubt das Bewegen der Szene und der Objekte in
# Blender mit dem 3D-Eingabegerät "SpaceController".
# 
#-------------------------------------------------------------------------------
# 
# 
# 
# 
#
# 
#
#
#
#
#
#                         Intentionally left blank.
#
#
#
#
#
#
#
#
#
#
#
#
#
#-------------------------------------------------------------------------------

import time
import datetime
import ctypes
import sys
import mathutils
import math # pi
import threading
import ctypes
import os
import traceback
import bpy
import re
import platform
import copy
from inspect import currentframe, getframeinfo
from bpy.props import StringProperty
from bpy.app.handlers import persistent

__author__ ="Friedemann Seebass"
__date__ ="$21.11.2020 14:15:00$"

bl_info = {
	"name": "SpaceController Blender Add-on",
	"description": "Adds support for SpaceControl's 3D input controllers to Blender.", 
	"author": "Friedemann Seebass",
	"version": (1, 0, 2),
	"blender": (2, 80, 0),
	"location": "View3D > SpaceController",
	"support": "COMMUNITY",
	"wiki_url": "http://www.friedemann-seebass.de/public_html/Software/misc/SpaceControllerBlenderAddOnDoc/SpaceControllerAddOnForBlenderReadMe.html",
	"category": "System"
	}

#-------------------------------------------------------------------------------

# Das Blender-Plug-in hat die folgende Grobstruktur:
# Es gibt eine Klasse BlenderPlugin, die ein Element der Klasse DaemonProxy und
# ein Element der Klasse BlenderProxy enthaelt.
#
#                         +-------------------------+
#                         |      BlenderPlugin      |
#                         +-------------------------+
#                           * 1        * 1        * 1
#                          /           |           \
#                       1 /            | n          \ 1
#         +--------------+    +-----------------+    +--------------+
#         | DaemonProxy  |    |     StdData     |    | BlenderProxy |
#         +--------------+    +-----------------+    +--------------+
#                 * 1                    1 |    \           * 1
#                 |                        |     \          |
#                 | 1                      |      \         | 1
#  +-----------------------------+         |     +------------+
#  | Dictionary of DeviceProxies |         |     +   Timer    +
#  +-----------------------------+         |     +------------+
#                              * 1         |
#                              |           |
#                              | n       1 |
#                            +---------------+     +--------+
#                            |  DeviceProxy  |---->| Thread |
#                            +---------------+     +--------+
#
# Legende:
#      abgeleitet --> von
#            Teil --* Ganzes
# benutzt/erzeugt --- wird benutzt/erzeugt
#

# Der DaemonProxy ist fuer die Kommunikation mit dem SC-Daemon zustaendig,
# der BlenderProxy fuer die Kommunikation mit Blender.
# Idealerweise weiss der DaemonProxy nichts ueber Blender-Strukturen und der
# BlenderProxy nichts vom SC-Daemon.
# Der DaemonProxy enthaelt ein Dictionary mit Elementen der Klasse DeviceProxy;
# Der Schluessel im Dictionary ist die Seriennummer eines Geraetes, der Wert
# ein DeviceProxy. Jeder DeviceProxy ist von der Python-Klasse Thread
# abgeleitet; der Thread holt sich die Kappendaten aus dem Geraet.
# Leider kann der Thread im DeviceProxy nicht direkt die Objekte in Blender
# manipulieren, weil das laut Blender-Doku nicht funktioniert. Daher legt jeder
# DeviceProxy die Kappendaten in einem Element der Klasse StdData ab. Das
# BlenderPlugin enthaelt eine Liste von StdData-Elementen, fuer jeden
# DeviceProxy eins.
# Der BlenderProxy startet nun seinerseits einen Endlos-Timer, der regelmaessig
# alle Elemente der StdData-Liste auf neue Eintraege durchsucht und neue
# Kappendaten in Bewegung in Blender umsetzt.
# Problem:
# Obwohl jeder DeviceProxy seinen eigenen Thread zum Abholen seiner Geraetedaten
# startet, reduziert sich die Geschwindigkeit beim Abholen, wenn mehrere
# Geraete angeschlossen werden. Bei einem Geraet werden wie ueblich problemlos
# alle 30 ms die Daten abgeholt, bei zwei Geraeten braucht man dafuer schon
# 60 ms. Folge: Die Auswertung laeuft massiv nach, wenn man die Senderate nicht
# drosselt. Die Ursache liegt in der SC-DLL, die aus gutem Grund die Routine
# doTheRequest() durch einen Mutex schuetzt. Das fuehrt dazu, dass ein Geraet
# erst dann wieder Daten bekommt, wenn alle anderen dran gewesen sind. Und
# wenn gerade keine Daten kommen, wartet jedes Geraet 30 ms bis es zurueckkehrt.
# Die vorlaeufige Loesung besteht darin, den DeviceProxy pruefen zu lassen,
# wie schnell die Kappendaten aufeinanderfolgen, und das Sende-Delay 
# entsprechend anzupassen. Eine endgueltige Loesung erforderte Aenderungen in
# der SC-DLL und damit wohl auch im Daemon, da dieser z. T. dieselben Dateien
# ("trncv_prot.cc" usw.) verwendet.
# fse, 07.08.2017
# fse, 23.08.2017: In doTheRequest() in sc_dll.cc wurde der Timeout von 40 ms
#                  auf 10 ms reduziert. Damit ist das Problem fuer bis zu vier
#                  Geraete einigermassen geloest.
# fse, 21.12.2020: Auf Blender 2.80 portiert (genauer: implementiert mit 
#                  Blender 2.91.0).
#
#-------------------------------------------------------------------------------

global gBlenderPlugin
global gIsUnregisterMenuEntry
global gCf
global gShortCuts  

gShortCuts = [] # Liste mit Shortcuts fuer unser Menue. Wir verwenden aber nur einen.

# Diese Variablennamen kommen in Parameterlisten von Funktionen vor und werden
# hier nur zugewiesen, damit der Obfuscator sie finden und ersetzen kann:
curFrame              = None
description           = None
isAutoFit             = None
isHandled             = None
isRegionView3dNeeded  = None
isToCheckForExistence = None
isUnregisterMenuEntry = None
msgTxt                = None
modus                 = None
sendDelay             = None
sharedData            = None
sharedStdDataList     = None
targetPntInWldSystem  = None

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Nur zur Abkuerzung (s. Kommentar bei log()):
gCf  = currentframe

# Schreibt den Text msgTxt auf die Konsole und setzt einen Header voran,
# bestehend aus Uhrzeit, Thread-ID, Zeilennummer, Klassenname und Funktionsname.
# curFrame wird vom Aufrufer uebergeben und enthaelt seinen currentFrame, den er
# durch Aufruf von currentFrame() holen muss. Damit die Aufrufe von log() etwas
# kompakter werden, wurde oben der Alias gCf() fuer currentFrame() eingefuehrt.
# fse, 27.07.17

def log(curFrame, obj, msgTxt):    
	return
	maxLocLen   = 30
	if curFrame == None:
		locHyphens = "-" * maxLocLen
		restHyphens = "-" * (80-25-maxLocLen)
		clm1 = "Class and Methods "
		clm1Len = len(clm1)
		clmFillLen = int((maxLocLen-clm1Len)/2)
		clm2 = " "*clmFillLen + clm1 + " "*clmFillLen
		print("--------------------------------------------------------------------------------")
		print("SpaceController Blender Plug-in log output opened at " + datetime.datetime.now().date().isoformat())
		print("--------------------------------------------------------------------------------")
		print("     Time   | TID |Line|" + clm2 + "| LogText")
		print("------------+-----+----+" + locHyphens + "+" + restHyphens)
	else:
		currentTime = datetime.datetime.now().time().isoformat()[:12]
		threadId    = threading.current_thread().ident
		frameinfo   = getframeinfo(curFrame)
		if obj == None:
			loc = "out of class"
		else:
			loc = obj.__class__.__name__ + "." + curFrame.f_code.co_name
		l = len(loc)
		if l > maxLocLen:
			loc = loc[l-maxLocLen:]
		msgStr = currentTime + "|" + str(threadId).rjust(5) + "|" + str(frameinfo.lineno).rjust(4) + "|" + loc.ljust(maxLocLen) + "|" + msgTxt
		print(msgStr)
	
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Die Klasse fuehrt den Thread aus, der regelmaessig prueft, ob die Verbindung
# zum Daemon noch besteht. Falls nicht, wird versucht sie aufzubauen.
# Ein Objekt der Klasse ist Member des DaemonProxy.
# fse, 28.07.17

class CheckingThread(threading.Thread):
	_isRunning   = False
	_isToRun     = True
	_isConnected = False
	_daemonProxy = None

	#----------------------------------------------------------------------------

	def __init__(self, sharedData, daemonProxy):
		log(gCf(), self, "Constructing ...")
		super().__init__()
		self._daemonProxy = daemonProxy
		log(gCf(), self, "Constructed.")

	#----------------------------------------------------------------------------

	def __del__(self):
		log(gCf(), self, "Invoked.")
		pass

	#----------------------------------------------------------------------------

	# Endlosschleife zur Pruefung, ob die Verbindung zum SC-Daemon noch besteht.
	# Falls nicht, wird versucht, sie wieder herzustellen.
	# fse, 04.08.17
	
	def run(self):
		log(gCf(), self, "Starting loop ...")
		cnt = 0
		self._isRunning = True
		while self._isToRun:
			self._daemonProxy.checkConnection()
#			log(gCf(), self, "cnt: {}".format(cnt))
#			cnt = cnt + 1
			time.sleep(1) # einmal in der Sekunde pruefen
		self._isRunning = False
		log(gCf(), self, "Loop ended.")

	#----------------------------------------------------------------------------

	# Beendet die Endlosschleife des Threads.
	# fse, 04.08.17
	
	def stop(self):
		cnt = 0
		self._isToRun = False
		while self._isRunning and cnt < 10:
			cnt += 1
			log(gCf(), self, "Waiting for stop ({})...".format(cnt))
			time.sleep(0.5)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Die Klasse repraesentiert ein SpaceControl-Geraet. Im wesentlichen stellt sie
# den Thread bereit, der die Kappendaten aus dem Geraet holt und in das 
# Shared-Memory-Objekt kopiert, aus dem sie der Timer in BlenderProxy holen
# kann.
# Ausserdem gibt es mit _selectionList[] eine Liste, in der die Blender-Objekte
# gespeichert sind, die ein Geraet exklusiv bewegt, wenn sie nicht leer ist.
# Bei leerer Liste bewegt ein Geraet immer die ganze Szene (sofern der
# ObjectMoveMode im Blender-Plug-in nicht aktiviert ist).
# fse, 28.07.17

class DeviceProxy(threading.Thread):
	_idx                 = -1
	_serialNo            = "unknown"
	_description         = "unknown"
	_isRunning           = False
	_isToRun             = True
	_daemonProxy         = None
	_blenderPlugin       = None
	_sharedStdData       = None # Datenbereich, in dem das Device die Kappendaten mit dem Blender-Proxy austauscht [s. getAndHandleData()]
	_selectionList       = []   # Liste der Blender-Objekte, die das Geraet exklusiv bewegt
	_region3d            = None # Region, in der der Mauszeiger war, als die _selectionList zugewiesen wurde
	_timeDifBuf          = []   # Liste, um Zeiten aufzuaddieren, scFetchStdData() braucht
	_isSendDelayAdjusted = False

	#----------------------------------------------------------------------------
	
	# Setzt die Region, die aktiv war/ist, waehrend die _selectionList
	# gesetzt wurde.
	# fse, 22.09.17
	
	def setRegion3d(self, region):
		self._region3d = region
		
	#----------------------------------------------------------------------------
	
	# Liefert die Region, die aktiv war/ist, waehrend die _selectionList
	# gesetzt wurde.
	# fse, 22.09.17
	
	def getRegion(self):
		return self._region3d

	#----------------------------------------------------------------------------

	# Setzt die Liste der Blender-Objekte, die das Device exklusiv bewegen soll.
	# fse, 22.09.17
	
	def setSelectionList(self, selectionList):
		self._selectionList = selectionList
		
	#----------------------------------------------------------------------------
	
	# Liefert die Liste der Blender-Objekte zurueck, die das Device gerade
	# exklusiv bewegt.
	# fse, 04.08.17
	
	def getSelectionList(self):
		return self._selectionList

	#----------------------------------------------------------------------------
	
	# Setzt das Flag, das steuert, ob der Reading-Thread die Anpassung des
	# Sende-Delays starten soll.
	# fse, 07.08.17

	def setIsSendDelayAdjusted(self, inPar):
		self._isSendDelayAdjusted = inPar

	#----------------------------------------------------------------------------

	def __init__(self, id, serialNo, description, daemonProxy):
		log(gCf(), self, "Constructing ...")
		super().__init__()
		self._idx           = id
		self._serialNo      = serialNo
		self._description   = description
		self._sharedStdData = StdData(devIdx=id, serNo=serialNo)
		self._daemonProxy   = daemonProxy
		self._blenderPlugin = self._daemonProxy._blenderPlugin
		self._blenderPlugin.addDevice(self._sharedStdData)
		log(gCf(), self, "Constructed: Device " + serialNo + ", _idx: " + str(self._idx) + ".")

	#----------------------------------------------------------------------------

	# Beendet den Reading-Thread des Device-Proxys.
	# fse, 04.08.17
	
	def stop(self):
		cnt = 0
		self._isToRun = False
		while self._isRunning and cnt < 10:
			cnt += 1
			log(gCf(), self, "Waiting for stop ({})...".format(cnt))
			time.sleep(0.5)

	#----------------------------------------------------------------------------

	# Der Reading-Thread des Device-Proxys. Mit scFetchStdData() werden
	# die Kappendaten geholt und in das Shared-Memory-Objekt self._sharedStdData
	# kopiert, aus dem der Timer des Blender-Proxys die Daten dann wieder
	# auslesen und verarbeiten kann.
	# Ausserdem wird gemessen, wie schnell die Kappendaten aufeinander folgen.
	# Wenn das langsamer ist als sie vom Geraet erzeugt werden, kommt es zum
	# Nachlaufen der Bewegung. Um das zu vermeiden, wird die Senderate mit der
	# Funktion adjustSendDelay() automatisch reduziert, wenn im
	# SpaceController-Menue der Schalter "Set Send Gap Automatically" gesetzt ist.
	# fse, 04.08.17
	# fse, 20.12.20: clock() replaced by perf_counter().

	def run(self):
		isDtaNull     = True
		isLastDtaNull = True
		log(gCf(), self, "Starting ...")
		x      = ctypes.c_short(0)
		y      = ctypes.c_short(0)
		z      = ctypes.c_short(0)
		a      = ctypes.c_short(0)
		b      = ctypes.c_short(0)
		c      = ctypes.c_short(0)
		traLmh = ctypes.c_int(0)
		rotLmh = ctypes.c_int(0)
		event  = ctypes.c_int(0)
		tvSec  = ctypes.c_long(0)
		tvUsec = ctypes.c_long(0)
		self._isRunning           = True
		self._sharedStdData.serNo = self._serialNo
		self._isSendDelayAdjusted = not self._blenderPlugin.isAutoSendGap
		actlTime                  = float(time.perf_counter())
		cnt                       = [0]        # Liste, damit als Returnwert verwendbar
		lastTime                  = [actlTime] # Liste, damit als Returnwert verwendbar
		self.retrieveAdvancedSettings() # um aktuellen _sendDelay zu bekommen
		while self._isToRun:
			if self._daemonProxy.getIsConnected():
				scStatus = self._blenderPlugin.scDll.scFetchStdData(self._idx, \
								ctypes.byref(x), \
								ctypes.byref(y), \
								ctypes.byref(z), \
								ctypes.byref(a), \
								ctypes.byref(b), \
								ctypes.byref(c), \
								ctypes.byref(traLmh), \
								ctypes.byref(rotLmh), \
								ctypes.byref(event), \
								ctypes.byref(tvSec), \
								ctypes.byref(tvUsec))
				self._daemonProxy.sendBlenderMode()
				if scStatus == 0:
					# Kappendaten in Shared-Memory kopieren:
					self._sharedStdData.x      = x     .value
					self._sharedStdData.y      = y     .value
					self._sharedStdData.z      = z     .value
					self._sharedStdData.a      = a     .value
					self._sharedStdData.b      = b     .value
					self._sharedStdData.c      = c     .value
					self._sharedStdData.traLmh = traLmh.value
					self._sharedStdData.rotLmh = rotLmh.value
					self._sharedStdData.event  = event .value
					self._sharedStdData.tvSec  = tvSec .value
					self._sharedStdData.tvUsec = tvUsec.value
					if isDtaNull:
						isLastDtaNull = True
					else:
						isLastDtaNull = False
					if self._sharedStdData.x == 0 and self._sharedStdData.y == 0 and self._sharedStdData.z == 0 and \
						self._sharedStdData.a == 0 and self._sharedStdData.b == 0 and self._sharedStdData.c == 0:
						isDtaNull = True
					else:
						isDtaNull = False
					self._sharedStdData.isDtaNull = isDtaNull
					self._sharedStdData.isLastDtaNull = isLastDtaNull
					self._sharedStdData.isHandled = False # fuer den BlenderProxy-Timer als zu bearbeitend markieren
#					if not isDtaNull:
#						self._sharedStdData.log(gCf())
					# Events bearbeiten:
					if self._sharedStdData.event == self._blenderPlugin.ADV_SETTINGS_CHANGED or self._sharedStdData.event == self._blenderPlugin.DEV_PARS_CHANGED:
						self.retrieveAdvancedSettings()
						self._sharedStdData.isHandled = True # um diese Events muss sich der Timer nicht kuemmern
					if self._sharedStdData.event == self._blenderPlugin.EVT_APPL_IN_FRGRND:
						self._daemonProxy.resetOldBlenderMode() # auf den Blender-Modus neu reagieren, wenn Blender wieder im Vordergrund
						self._sharedStdData.isHandled = True
					self._daemonProxy.sendBlenderMode()
					if not self._isSendDelayAdjusted and self._blenderPlugin.isAutoSendGap:
						self.adjustSendDelay(cnt, lastTime) # den Send-Delay neu anpassen, falls noetig
				else:
#					log(gCf(), self, "scStatus: " + str(scStatus))
					pass
			else:
				time.sleep(0.1)			

		self._isRunning = False
		log(gCf(), self, "Reading thread for device " + self._serialNo + "(" + str(self._idx) + ") ended.")

	#-------------------------------------- -------------------------------------
	
	# Holt den Move-Mode und den Auto-Fit-Status aus den Konfigurationsdaten.
	# fse, xx.xx.15
	# fse, 08.07.16: Jetzt wird auch der mIsAutoFit geholt.
	# fse, 04.08.17: Und auch noch mSendDelay.
	
	def retrieveAdvancedSettings(self):
		devIdx      = self._idx
		advSettings = SC_ADVANCED_SETTINGS()
		scStatus    = self._blenderPlugin.scDll.scGetAdvancedSettings(devIdx, ctypes.byref(advSettings))
		if scStatus == 0:
			self._blenderPlugin.setMoveMode (advSettings.mMoveMode )
			self._blenderPlugin.setIsAutoFit(advSettings.mIsAutoFit)
			self._sendDelay = advSettings.mSendDelay
			log(gCf(), self,   "sendDelay: " + str(self._sendDelay) +  ", moveMode: " + str(advSettings.mMoveMode) + ", isAutoFit: " + str(advSettings.mIsAutoFit))
		else:
			log(gCf(), self, "scGetAdvancedSettings() failed! scStatus: " + str(scStatus))
			pass

	#-------------------------------------- -------------------------------------
	
	# Passt den Send-Delay an die aktuell zu erzielende Datenrate an.
	# Die Funktion ermittelt mit time.clock() die aktuelle Zeit und vergleicht
	# sie mit der des letzten Aufrufs. Falls diese Zeit groesser ist als die
	# Zeit zwischen zwei Datenpakten kommt es zum Nachlaufen. Um das zu
	# verhindern, wird der Send-Delay des Geraetes angepasst.
	# Dazu werden 20 Messwerte aufeinanderfolgender Zeitdifferenzen in einem
	# Puffer gespeichert. Wenn die 20 Werte beisammen sind, werden sie sortiert,
	# und der Median wird als neuer Send-Delay verwendet und im Geraet 
	# eingestellt. Der Median hat den Vorteil, dass regelmaessig zu beobachtende
	# Aussreisser keine Rolle spielen.
	# fse, 07.08.17

	def adjustSendDelay(self, cnt, lastTime):
		maxElNum = 20
		hlfElNum = int(maxElNum/2)
		actlTime = float(time.clock())
		timeDif = actlTime - lastTime[0]
		log(gCf(), self, "idx: " + str(self._idx) + ", serialNo: " + self._serialNo + ", cnt: " + str(cnt[0]) + ", actlTime: " + str(actlTime) + ", lastTime: " + str(lastTime[0]) + ", timeDif: " + str(timeDif))
		if timeDif > 0.02:
			if cnt[0] < maxElNum:
				# erst einmal 20 Werte im Puffer sammeln:
				self._timeDifBuf.append(timeDif)
				cnt[0] += 1
			else:
				# dann den Median im Puffer als neuen Wert fuer sendDelay einstellen:
				self._timeDifBuf.sort()
				timeDif = self._timeDifBuf[hlfElNum] # Median der Werte im Puffer
#				log(gCf(), self, "_timeDifBuf: " + str(self._timeDifBuf))
				log(gCf(), self, "_idx: " + str(self._idx) + ", _serialNo: " + self._serialNo + ", timeDif: " + str(timeDif))
				self.setSendDelay(timeDif*1000)
				del self._timeDifBuf[:]
				cnt[0] = 0
				log(gCf(), self, "cnt: " + str(cnt[0]))
				self._isSendDelayAdjusted = True
		lastTime[0] = actlTime
		return

	#------------------------- ------------------------- ------------------------
	
	# Stellt im Geraet den uebergebenen Send-Delay-Werte ein.
	# fse, 07.08.17

	def setSendDelay(self, sendDelay):
		sd = ctypes.c_int(int(sendDelay))
		scStatus = self._blenderPlugin.scDll.scSetSendDelay(self._idx, ctypes.byref(sd), ctypes.byref(sd))
		if scStatus == 0:
			log(gCf(), self, "idx: " + str(self._idx) + ", serialNo: " + self._serialNo + ", sendDelay: " + str(sendDelay))
			pass
		else:
			log(gCf(), self, "idx: " + str(self._idx) + ", serialNo: " + self._serialNo + ", sendDelay: " + str(sendDelay) + ", scSetSendDelay() failed! scStatus: " + str(scStatus))
			pass

	#-------------------------------------- -------------------------------------
	
	# Stoesst fuer das Device die automatische Anpassung der
	# Senderate an, indem das Send-Delay erst einmal wieder auf 30 ms
	# herabgesetzt und das Flag zum Ausloesen der Datensammlung gesetzt wird.
	# fse, 07.08.17
	
	def triggerSendDelayAdjust(self):
		self.setSendDelay(30)             # Erst einmal wieder hohe Standard-Senderate einstellen ...
		self._isSendDelayAdjusted = False # ... und Flagge setzen, damit das Datensammeln wieder beginnt.
		log(gCf(), self, "idx: " + str(self._idx) + ", serialNo: " + self._serialNo)

	#----------------------------------------------------------------------------

	def setSelectionListAndRegion(self, selectionList, region):
		if self._selectionList:
			# in der Liste steht schon was => Liste loeschen:
			self._selectionList = []
			self._region3d = None
		else:
			# Liste leer => aktuelle Selektion zuweisen:
			self._selectionList = selectionList
			self._region3d = region

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Kopie der Struktur ScDevInfo aus spc_ctrlr.h.
# fse, xx.xx.14

class SC_DEV_INFO(ctypes.Structure):
	_fields_ = [ ("mId"      , ctypes.c_int    ),
	             ("mSerialNo", ctypes.c_char*16),  # Achtung: in spc_ctrlr.h Laenge als MAX_SERIALNO_LEN
					 ("mDescrptn", ctypes.c_char*64) ] # Achtung: in spc_ctrlr.h Laenge als MAX_DESCRPTN_LEN

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Kopie der Struktur ScAdvSens aus scDevpars.h.
# fse, xx.xx.14

class SC_ADV_SENS(ctypes.Structure):
	_fields_ = [ ("mXSens", ctypes.c_int),
					 ("mYSens", ctypes.c_int),
					 ("mZSens", ctypes.c_int),
					 ("mASens", ctypes.c_int),
					 ("mBSens", ctypes.c_int),
					 ("mCSens", ctypes.c_int) ]

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Kopie der Struktur ScAdvancedSettings aus scDevpars.h.
# fse, xx.xx.14

class SC_ADVANCED_SETTINGS(ctypes.Structure):
	_fields_ = [ ("mSens"            , SC_ADV_SENS   ),
					 ("mIsX"             , ctypes.c_bool ),
					 ("mIsY"             , ctypes.c_bool ),
					 ("mIsZ"             , ctypes.c_bool ),
					 ("mIsXRev"          , ctypes.c_bool ),
					 ("mIsYRev"          , ctypes.c_bool ),
					 ("mIsZRev"          , ctypes.c_bool ),
					 ("mXRemap"          , ctypes.c_int  ),
					 ("mYRemap"          , ctypes.c_int  ),
					 ("mZRemap"          , ctypes.c_int  ),
					 ("mIsA"             , ctypes.c_bool ),
					 ("mIsB"             , ctypes.c_bool ),
					 ("mIsC"             , ctypes.c_bool ),
					 ("mIsARev"          , ctypes.c_bool ),
					 ("mIsBRev"          , ctypes.c_bool ),
					 ("mIsCRev"          , ctypes.c_bool ),
					 ("mARemap"          , ctypes.c_int  ),
					 ("mBRemap"          , ctypes.c_int  ),
					 ("mCRemap"          , ctypes.c_int  ),
					 ("mMouseDirX"       , ctypes.c_int  ),
					 ("mMouseDirY"       , ctypes.c_int  ),
					 ("mMouseDirZ"       , ctypes.c_int  ),
					 ("mMouseDirA"       , ctypes.c_int  ),
					 ("mMouseDirB"       , ctypes.c_int  ),
					 ("mMouseDirC"       , ctypes.c_int  ),
					 ("mIsMouseMove"     , ctypes.c_bool ),
					 ("mIsMouseToRestore", ctypes.c_bool ),
					 ("mIsMouseAbs"      , ctypes.c_bool ),
					 ("mSendDelay"       , ctypes.c_int  ),
					 ("mAverageNum"      , ctypes.c_int  ),
					 ("mIsRemap"         , ctypes.c_bool ),
					 ("mTraLmBnd"        , ctypes.c_short),
					 ("mTraMhBnd"        , ctypes.c_short),
					 ("mRotLmBnd"        , ctypes.c_short),
					 ("mRotMhBnd"        , ctypes.c_short),
					 ("mIsToSendToFgWin" , ctypes.c_bool ),
					 ("mIsSpcFuncToSend" , ctypes.c_bool ),
					 ("mIsWheelAssShown" , ctypes.c_bool ),
					 ("mIsWsFuncToSend"  , ctypes.c_bool ),
					 ("mIsCapFuncToSend" , ctypes.c_bool ),
					 ("mAreChangesToSend", ctypes.c_bool ),
					 ("mBrightnessRed"   , ctypes.c_int  ),
					 ("mBrightnessGreen" , ctypes.c_int  ),
					 ("mAreDevLights"    , ctypes.c_bool ),
					 ("mRotCenterMode"   , ctypes.c_int  ),
					 ("mMoveMode"        , ctypes.c_int  ),
					 ("mIsRightHandMode" , ctypes.c_bool ),
					 ("mIsClock"         , ctypes.c_bool ),
					 ("mIsAutoFit"       , ctypes.c_bool ),
					 ("mLanguage"        , ctypes.c_int  ),
					 ("mPicture"         , ctypes.c_int  ) ]

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Der Daemon-Proxy uebernimmt die Kommunikation mit unserem SC-Daemon.
# Im Unterschied dazu ist der Blender-Proxy fuer die Kommunikation mit Blender
# zustaendig.
# Allen dreien gemeinsam ist die Liste _sharedStdDataList[], die zum Austausch
# der Kappendaten dient und in den Konstruktoren uebergeben wird.
# fse, xx.xx.14

class DaemonProxy:
	_blenderPlugin       = None
	_sharedStdDataList   = None
	_checkingThread      = None
	_daemonThread        = None
	_blenderMode         = "unknown"
	_oldBlenderMode      = ""
	_isInSendBlenderMode = False # eigene Semaphore, weil zur Einarbeitung in Python-Mechanismen zu faul
	_isConnected         = False
	_devNum              = 0     # Anzahl der angeschlossenen Devices
	_devDic              = {}    # Dictionary mit Seriennummer als Key und DeviceProxy als Value; Liste aller angeschlossenen Devices
	_isDeviceListModifying = False

	#----------------------------------------------------------------------------

	def __init__(self, blenderPlugin, sharedStdDataList):
		log(gCf(), self, "Constructing ...")
		self._sharedStdDataList = sharedStdDataList
		self._blenderPlugin     = blenderPlugin
		log(gCf(), self, "Constructed.")

	#----------------------------------------------------------------------------

	def __del__(self):
		log(gCf(), self, "Destructing ...")
		if self._daemonThread != None:
			self.stop()
		log(gCf(), self, "Destructed.")

	#----------------------------------------------------------------------------
	
	def getIsDeviceListModifying(self):
		return self._isDeviceListModifying

	#----------------------------------------------------------------------------
	
	def getIsConnected(self):
		return self._isConnected
	
	#----------------------------------------------------------------------------
	
	# Prueft, ob ein Device im Move-Mode ist, ob ihm also eine Liste von
	# Blender-Objekten zugewiesen wurde, die es exklusiv bewegen soll statt der
	# ganzen Szene.
	# fse, 04.08.17
	
	def getIsDeviceInMoveMode(self, serialNo):
		deviceProxy = self._devDic[serialNo]
		selectionList = deviceProxy.getSelectionList()
		isInMoveMode = len(selectionList) > 0
		return isInMoveMode
		
	#----------------------------------------------------------------------------
	
	def getIsRunning(self):
		if (self._checkingThread == None):
			return False
		else:
			return True

	#----------------------------------------------------------------------------
	
	# Prueft, ob sich ein Device mit der Seriennummer serialNo im Dictionary
	# self._devDic befindet.
	# fse, 04.08.17
	
	def isDeviceThere(self, serialNo):
		if serialNo in self._devDic:
			return True
		else:
			return False

	#----------------------------------------------------------------------------
	
	# Startet den Thread, der endlos die Anzahl der angesteckten Geraete prueft.
	# fse, 04.08.17
	
	def startDaemonProxy(self):
		log(gCf(), self, "Invoked.")
		if self._checkingThread == None:
			self._checkingThread = CheckingThread(self._blenderPlugin, self)
			self._checkingThread.start()
		else:
			log(gCf(), self, "CheckingThread already started!")
			pass

	#----------------------------------------------------------------------------
	
	# Beendet alle Threads des Daemon-Proxys, also alle Device-Reading-Threads
	# und den Checking-Thread, der die Anzahl der angesteckten Devices prueft.
	# fse, xx.xx.14
	# fse, 04.08.17: Es gibt jetzt fuer jedes Device einen Reading-Thread.

	def stop(self):
		for serialNo in self._devDic:
			deviceProxy = self._devDic[serialNo]
			if deviceProxy == None:
				log(gCf(), self, "DeviceProxy not started yet!")
				pass
			else:
				deviceProxy.stop()
				self._devDic[serialNo] = None
		if self._checkingThread == None:
			log(gCf(), self, "CheckingThread not started yet!")
			pass
		else:
			self._checkingThread.stop()
			del self._checkingThread
			self._checkingThread = None
			self._devNum = 0
			self._devDic.clear()
			del self._sharedStdDataList[:]

	#----------------------------------------------------------------------------
	
	# Prueft, ob das Plug-in mit dem SC-Daemon verbunden ist. Falls nicht, wird
	# versucht, die Verbindung herzustellen.
	# Ausserdem wird die Anzahl der angeschlossenen Geraete ueberwacht und bei
	# jeder Aenderung eine entsprechende Reaktion ausgeloest.
	# fse, 04.08.17
	
	def checkConnection(self):
		devNumCint     = ctypes.c_int(0)
		usedDevNumCint = ctypes.c_int(0)
		maxDevIdxCint  = ctypes.c_int(0)
		scStatus = self._blenderPlugin.scDll.scGetDevNum(ctypes.byref(devNumCint), ctypes.byref(usedDevNumCint), ctypes.byref(maxDevIdxCint))
		if scStatus == 0:
			devNum     = devNumCint.value
			usedDevNum = usedDevNumCint.value
			maxDevIdx  = maxDevIdxCint.value
			self._isConnected = True
#			log(gCf(), self, "devNum: " + str(devNum) + ", usedDevNum: " + str(usedDevNum) + ", maxDevIdx: " + str(maxDevIdx))
			if devNum != self._devNum:
				log(gCf(), self, "devNum: " + str(devNum) + ", self._devNum: " + str(self._devNum))
				self.modifyDeviceList(maxDevIdx)
		else:
			self._isConnected = False
			devNum = 0
			log(gCf(), self, "scGetDevNum() failed with scStatus=" + str(scStatus) + ". Try to connect with daemon ...")
			self._blenderPlugin.scDll.scDisconnect()
			if devNum != self._devNum:
				log(gCf(), self, "devNum: " + str(devNum) + ", self._devNum: " + str(self._devNum))
				self.modifyDeviceList(0)
			self._isConnected = self.connectToDaemon()
		self._devNum = devNum
	
	#-------------------------------------- -------------------------------------
	
	# Stellt Verbindung zum Daemon her.
	# Achtung: Der Name "Blender" wird im Daemon verschiedentlich benutzt bzw.
	# abgefragt. Bei Aenderung unbedingt auch dort anpassen (Suche nach
	# "Blender").
	# fse, 27.07.17

	def connectToDaemon(self):
		scStatus = self._blenderPlugin.scDll.scConnect2(False, ctypes.c_char_p(b"Blender"))
		log(gCf(), self, "scStatus: {}".format(scStatus))
		if (scStatus == 0):
			return True
		else:
			return False
	
	#-------------------------------------- -------------------------------------
	
	# Passt das Dictionary self._devDic der angeschlossenen Geraete an und loest
	# alle weiteren, noetigen Aktionen aus; z. B. wird das SpaceController-Menue
	# in Blender entsprechend angepasst.
	# fse, 04.08.17

	def modifyDeviceList(self, maxDevIdx):
		self._isDeviceListModifying = True
		log(gCf(), self, "----- Invoked. maxDevIdx: " + str(maxDevIdx))
		actDevSerNos = [] # Liste aller aktuell angeschlossenen Geraete
		newDevDic    = {} # Liste der neu angeschlossenen Geraete
		delDevDic    = {} # Liste der zu entfernenden Geraete
		scDevInfo    = SC_DEV_INFO()
		# FTDI Zeit zum Initialisieren der Geraete geben, sonst kann 
		# scGetDevInfo() scheitern:
		time.sleep(0.5)
		# Liste newDevDic mit neu angeschlossenen Geraeten aufbauen:
		for i in range(0, maxDevIdx+1):
			scStatus = self._blenderPlugin.scDll.scGetDevInfo(i, ctypes.byref(scDevInfo))
			if scStatus == 0:
				idx      = scDevInfo.mId
				serialNo = "".join(map(chr, scDevInfo.mSerialNo))
				descrptn = "".join(map(chr, scDevInfo.mDescrptn))
				actDevSerNos.append(serialNo)
				log(gCf(), self, "mId: " + str(id) + ", mSerialNo: " + serialNo + ", descrptn: " + descrptn)
				if serialNo in self._devDic:
					# Geraet ist schon in der Liste der angeschlossenen Geraete => ignorieren:
					pass
				else:
					# Geraet ist noch nicht in der Liste => Geraet erzeugen und
					# in Liste mit neuen Geraeten eintragen:
					deviceProxy = DeviceProxy(idx, serialNo, descrptn, self)
					newDevDic[serialNo] = deviceProxy
			else:
				log(gCf(), self, "scGetDevInfo() failed! scStatus: " + str(scStatus))
				pass
		# Pruefen, ob frueher vorhandene Geraete noch da sind und loeschen, falls nicht:
		for serialNo in self._devDic:
			if not serialNo in actDevSerNos:
				delDevDic[serialNo] = self._devDic[serialNo]
		for serialNo in delDevDic:
			deviceProxy = self._devDic[serialNo]
			deviceProxy.stop()
			self._blenderPlugin.removeDevice(deviceProxy._sharedStdData)
			del self._devDic[serialNo]
		# neue Geraete in Liste eintragen und deren Thread starten:
		for serialNo in newDevDic:
			deviceProxy = newDevDic[serialNo]
			log(gCf(), self, "Adding device " + deviceProxy._serialNo)
			self._devDic[serialNo] = deviceProxy
			deviceProxy.start()
		log(gCf(), self, "Actual Device List: -----")
		for serialNo in self._devDic:
			deviceProxy = self._devDic[serialNo]
			if self._blenderPlugin.isAutoSendGap:
				deviceProxy.triggerSendDelayAdjust()
			log(gCf(), self, "idx: " + str(deviceProxy._idx) + ", serialNo: " + deviceProxy._serialNo)
		log(gCf(), self, "                    ^^^^^")
		self._isDeviceListModifying = False
		log(gCf(), self, "----- Finished.")
		
	#----------------------------------------------------------------------------

	# Stoesst fuer alle angeschlossenen Geraete die automatische Anpassung der
	# Senderate an, indem die Send-Delays erst einmal wieder auf 30 ms
	# herabgesetzt und die Flags zum Ausloesen der Datensammlung gesetzt werden.
	# Wird aufgerufen, wenn man im SpaceController-Menue die Pruefung einschaltet.
	# fse, 07.08.17
	
	def triggerSendDelayAdjustForAllDevices(self):
		log(gCf(), self, "Invoked.")
		for serialNo in self._devDic:
			log(gCf(), self, serialNo)
			deviceProxy = self._devDic[serialNo]
			deviceProxy.triggerSendDelayAdjust()

	#----------------------------------------------------------------------------

	# Setzt die Liste der Objekte, die von einem Device exklusiv bewegt werden
	# sollen und die Region, die beim Anmelden der Bewegung aktiv war.
	# \param serialNo      Serienummer des Geraetes, dessen selectionList gesetzt
	#                      werden soll
	# \param selectionList die zu setzende Liste der Objekte, die vom Geraet
	#                      exklusiv bewegt werden sollen
	# \param region        die (Region, die zum Zeitpunkt des Setzens aktiv war
	# fse, 02.09.17
	
	def setDevProxySelectionListAndRegion(self, serialNo, selectionList, region):
		deviceProxy = self._devDic[serialNo]
		deviceProxy.setSelectionListAndRegion(selectionList, region)
		
	#----------------------------------------------------------------------------

	# Setzt den Blender-Modus, den wir an den Daemon uebertragen wollen.
	# Es hat sich herausgestellt, dass diese Methode beim Laden einer neuen
	# Blender-Datei genau dann aufgerufen wird, waehrend sendBlenderMode()
	# gerade ausgefuehrt wird. Folge: self._oldBlenderMode wurde dort auf 
	# ein anderes self._blenderMode gesetzt als es am Beginn der Methode war und
	# der neue Modus wurde nicht uebertragen.
	# self._blenderMode muesste also eigentlich durch eine Semaphore geschuetzt
	# werden. Da ich aber keine Lust habe, mich in die entsprechende
	# Python-Methodik einzulesen, wird die Semaphore hier durch die
	# While-Schleife nachgebildet. Hoffentlich tauchen nicht noch mehr solche
	# Faelle auf ...
	# fse, 23.01.15
	
	def setBlenderMode(self, modus):
		cnt = 0
		if self._isInSendBlenderMode:
			while self._isInSendBlenderMode and cnt < 30:
				cnt += 1
				log(gCf(), self, "Waiting for sendBlenderMode() to finish ({})...".format(cnt))
				time.sleep(0.05)
			# Aus nicht geklaerten Umstaenden kann das Plug-in ohne die folgende
			# Pause nach dem Start nicht den aktuellen Modus uebermitteln, sofern
			# mehr als ein Geraet angeschlossen ist; irgendein Thread ist dann
			# anscheinend noch beschaeftigt. 0.3 Sekunden reichen nicht
			# zur Heilung, daher 0.4 Sekunden Pause. Auch beim Laden einer
			# "*.blend"-Datei hat man dasselbe Problem, weil das Plug-in dabei
			# beendet und wieder gestartet wird:
			time.sleep(0.4)
		self._blenderMode = modus

	#----------------------------------------------------------------------------

	def resetOldBlenderMode(self):
		self._oldBlenderMode = ""
		
	#----------------------------------------------------------------------------
	
	# Sendet den Modus, in dem sich Blender befindet ("Default", "Skripting",
	# "Video Editing" usw.) an den Daemon.
	# fse, xx.xx.xx
	
	def sendBlenderMode(self):
		self._isInSendBlenderMode = True
#		log(gCf(), self, "_oldBlenderMode: " + str(self._oldBlenderMode) + ", _blenderMode: " + str(self._blenderMode))
		if self._blenderMode != self._oldBlenderMode:
			scStatus = self._blenderPlugin.scDll.scSetState2(ctypes.c_char_p(self._blenderMode.encode(encoding='ascii')))
			log(gCf(), self, "scStatus: " + str(scStatus) + ", _blenderMode: " + str(self._blenderMode))
			if scStatus == 0:
				self._oldBlenderMode = self._blenderMode
		self._isInSendBlenderMode = False
			
	#----------------------------------------------------------------------------

	def test(self):
		log(gCf(), self, "Invoked. xxx")
		self._blenderPlugin.test()

	#----------------------------------------------------------------------------

	def clean(self):
		log(gCf(), self, "Invoked.")
		del self._blenderPlugin

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Erzeugt das SpaceController-Menue in der Blender-Menuezeile.
# fse, 02.12.14 

class SpaceControlMenu(bpy.types.Menu):
	bl_label = "SpaceController"
	bl_idname = "OBJECT_MT_spacecontrol_menu"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def resetBlenderPlugin(self):
		self._blenderPlugin = None

	def draw(self, context):
		layout = self.layout
		layout.menu(RotCenterMenu.bl_idname)
		if self._blenderPlugin.isSlowDownMode:
			layout.operator("sc_op.slow_down_mode", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.slow_down_mode", icon = 'CHECKBOX_DEHLT')
		if self._blenderPlugin.isOnly3dViewActive:
			layout.operator("sc_op.only_3d_view", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.only_3d_view", icon = 'CHECKBOX_DEHLT')
		layout.separator()
		if self._blenderPlugin.getDevNum() > 1:
			layout.menu(BindDevicesMenu.bl_idname)
		if self._blenderPlugin.isObjectMoveMode:
			layout.operator("sc_op.object_move_mode", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.object_move_mode", icon = 'CHECKBOX_DEHLT')
		if self._blenderPlugin.isOmmMultiRotCenter:
			layout.operator("sc_op.multi_rotation_centers", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.multi_rotation_centers", icon = 'CHECKBOX_DEHLT')
		if self._blenderPlugin.isGenKeyframes:
			layout.operator("sc_op.gen_keyframe", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.gen_keyframe", icon = 'CHECKBOX_DEHLT')
		layout.separator()
		#if self._blenderPlugin.getIsThreadRunning():
		#	threadTxt = "Thread: on"
		#else:
		#	threadTxt = "Thread: off"
		#if self._blenderPlugin.getIsTimerRunning():
		#	timerTxt = "Timer: on"
		#else:
		#	timerTxt = "Timer: off"
		#startTxt = "Start (Current states: " + threadTxt + ", " + timerTxt + ")"
		startTxt = "Start"
		layout.operator("sc_op.start_plugin", text = startTxt)
		layout.operator("sc_op.stop_plugin")
		if self._blenderPlugin.getIsUnregisterMenuEntry():
			layout.separator()
			layout.operator("sc_op.unregister_plugin")
		
	def showMenu(self, context):
		layout = self.layout
		layout.menu(SpaceControlMenu.bl_idname)
		
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Distant Rotation Center: Higher
# Sensitivity' ausgefuehrt wird.
# fse, 16.01.15

class ReduceSensScPluginOp(bpy.types.Operator):
	""" Increases the device's sensitivity when the virtual camera is far away from the rotation center """
	bl_idname = "sc_op.slow_down_mode"
	bl_label  = "Distant Rotation Center: Higher Sensitivity"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.toggleReduceSensMode()
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Affect 3D View Areas Only'
# ausgefuehrt wird.
# fse, 10.03.15

class AffectOnly3dViewScPluginOp(bpy.types.Operator):
	""" If activated only Blender's 3D View areas are affected by the controller """
	bl_idname = "sc_op.only_3d_view"
	bl_label  = "Affect 3D View Areas Only"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.toggle3dViewMode()
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Set Send Gap Automatically'
# ausgefuehrt wird.
# fse, 16.12.14

class AutoSendGapScPluginOp(bpy.types.Operator):
	""" If activated the plug-in adjusts the device's send gap automatically to avoid lagging """
	bl_idname = "sc_op.auto_send_gap"
	bl_label  = "Set Send Gap Automatically"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.toggleAutoSendGap()
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Remember View When Bind Selection'
# ausgefuehrt wird.
# fse, 22.09.17

class RememberViewScPluginOp(bpy.types.Operator):
	""" If activated the plug-in remembers the last used (split) view when bind a selection to a device """
	bl_idname = "sc_op.remember_view"
	bl_label  = "Remember View When Bind Selection"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.toggleRememberView()
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Object Move Mode'
# ausgefuehrt wird.
# fse, 16.12.14

class ObjectMoveModeScPluginOp(bpy.types.Operator):
	""" Switches the Object Move Mode on and off """
	bl_idname = "sc_op.object_move_mode"
	bl_label  = "Object Move Mode"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.toggleObjectMoveMode()
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Multi Rotation Centers'
# ausgefuehrt wird.
# fse, 16.12.14

class MultiRotCentersOmmScPluginOp(bpy.types.Operator):
	""" If activated and object move mode is on and multiple objects are selected each object rotates around itself """
	bl_idname = "sc_op.multi_rotation_centers"
	bl_label  = "Multi Rotation Centers"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.toggleMultiRotCentersMmm()
		return {'FINISHED'}
	
	@classmethod
	def poll(self, context):
		if self._blenderPlugin.getIsObjectMoveMode():
			return True
		else:
			return False

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Generate Keyframes'
# ausgefuehrt wird.
# fse, 28.01.15

class GenKeyframesOmmScPluginOp(bpy.types.Operator):
	""" If activated keyframes are created automatically """
	bl_idname = "sc_op.gen_keyframe"
	bl_label  = "Generate Keyframes"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.toggleGenKeyframesOmm()
		return {'FINISHED'}
	
	@classmethod
	def poll(self, context):
		if self._blenderPlugin.getIsObjectMoveMode():
			return True
		else:
			return False

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Erzeugt das Untermenue mit der Rotationszentrenauswahl im SpaceController-Menue.
# fse, 17.12.14

class RotCenterMenu(bpy.types.Menu):
	bl_label = "Rotation Center"
	bl_idname = "OBJECT_MT_rotcenter_menu"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def draw(self, context):
		layout = self.layout
		if self._blenderPlugin.rotCenter == 'DEFAULT':
			layout.operator("sc_op.rc_default", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.rc_default", icon = 'CHECKBOX_DEHLT')
		if self._blenderPlugin.rotCenter == 'CURSOR':
			layout.operator("sc_op.rc_cursor", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.rc_cursor", icon = 'CHECKBOX_DEHLT')
		if self._blenderPlugin.rotCenter == 'SELECTION':
			layout.operator("sc_op.rc_selection", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.rc_selection", icon = 'CHECKBOX_DEHLT')
		
	def showMenu(self, context):
		layout = self.layout
		layout.menu(RotCenterMenu.bl_idname)
		
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts "Rotation Center/Default"
# ausgefuehrt wird.
# fse, 17.12.14

class RotCenterDefaultScPluginOp(bpy.types.Operator):
	""" If activated the rotation center is the one that is set in the SpaceControl driver's user interface """
	bl_idname = "sc_op.rc_default"
	bl_label  = "As Set In SpaceControl Panel"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.setRotCenterToDefault()
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts "Rotation Center/3D Cursor"
# ausgefuehrt wird.
# fse, 17.12.14

class RotCenterCursorScPluginOp(bpy.types.Operator):
	""" If activated the rotation center is the 3D view's cursor """
	bl_idname = "sc_op.rc_cursor"
	bl_label  = "3D Cursor"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.setRotCenterToCursor()
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts "Rotation Center/Selection"
# ausgefuehrt wird.
# fse, 17.12.14

class RotCenterSelectionScPluginOp(bpy.types.Operator):
	""" If activated the rotation center is the center of the current selection """
	bl_idname = "sc_op.rc_selection"
	bl_label  = "Selection"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.setRotCenterToSelection()
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Erzeugt das Untermenue zur Auswahl eines Geraetes.
# fse, 02.08.17

class BindDevicesMenu(bpy.types.Menu):
	bl_label = "Bind Selection to Device"
	bl_idname = "OBJECT_MT_devices_menu"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def draw(self, context):
		layout = self.layout
		for stdData in self._blenderPlugin.devDataList:
			idx    = stdData.devIdx
			opName = "sc_op.bind_device_" + str(idx)
			serNo = stdData.serNo
			isDeviceInMoveMode = self._blenderPlugin.daemonProxy.getIsDeviceInMoveMode(serNo)
			if isDeviceInMoveMode:
				layout.operator(opName, icon = 'CHECKBOX_HLT')
			else:
				layout.operator(opName, icon = 'CHECKBOX_DEHLT')
		layout.separator()
		if self._blenderPlugin.isRememberView:
			layout.operator("sc_op.remember_view", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.remember_view", icon = 'CHECKBOX_DEHLT')
		if self._blenderPlugin.isAutoSendGap:
			layout.operator("sc_op.auto_send_gap", icon = 'CHECKBOX_HLT')
		else:
			layout.operator("sc_op.auto_send_gap", icon = 'CHECKBOX_DEHLT')
		
	def showMenu(self, context):
		layout = self.layout
		layout.menu(DevicesMenu.bl_idname)
		
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# ausgefuehrt wird.
# fse, 02.08.17

class BindDevice0Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_0"
	bl_label  = "Device 0"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# fse, 02.08.17

class BindDevice1Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_1"
	bl_label  = "Device 1"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# fse, 02.08.17

class BindDevice2Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_2"
	bl_label  = "Device 2"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# fse, 02.08.17

class BindDevice3Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_3"
	bl_label  = "Device 3"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# ausgefuehrt wird.
# fse, 02.08.17

class BindDevice4Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_4"
	bl_label  = "Device 4"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# ausgefuehrt wird.
# fse, 02.08.17

class BindDevice5Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_5"
	bl_label  = "Device 5"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# ausgefuehrt wird.
# fse, 02.08.17

class BindDevice6Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_6"
	bl_label  = "Device 6"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# ausgefuehrt wird.
# fse, 02.08.17

class BindDevice7Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_7"
	bl_label  = "Device 7"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# ausgefuehrt wird.
# fse, 02.08.17

class BindDevice8Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_8"
	bl_label  = "Device 8"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken eines "Bind Selection to Device"-Untermenuepunkts
# ausgefuehrt wird.
# fse, 02.08.17

class BindDevice9Op(bpy.types.Operator):
	""" Binds all selected objects to the device with this serial no. for movement """
	bl_idname = "sc_op.bind_device_9"
	bl_label  = "Device 9"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		selectionList = list(bpy.context.selected_objects) # echte Kopie
		self._blenderPlugin.setSelectionList(self.bl_label, selectionList)
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Start Plug-in' ausgefuehrt
# wird.
# fse, 02.12.14

class StartScPluginOp(bpy.types.Operator):
	""" Starts the SpaceController Plug-in """
	bl_idname = "sc_op.start_plugin"
	bl_label  = "Start"
	_blenderPlugin = None
	
	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		log(gCf(), self, "SpaceController Plug-in started.")
#		bpy.utils.register_class(daemon_proxy.ModalDaemonTimerOp)
#		bpy.ops.wm.modal_spacecontrol_timer_operator()
		self._blenderPlugin.startPlugin()
		return {'FINISHED'}

	@classmethod
	def poll(self, context):
		if self._blenderPlugin.getIsTimerRunning() and self._blenderPlugin.getIsThreadRunning():
			return False
		else:
			return True

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Stop Plug-in' ausgefuehrt
# wird.
# fse, 02.12.14

class StopScPluginOp(bpy.types.Operator):
	""" Stops the SpaceController Add-on """
	bl_idname = "sc_op.stop_plugin"
	bl_label  = "Stop"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		ModalBlenderTimerOp.setTimerContext(ModalBlenderTimerOp, context) # Timer mit aktuellem Context versorgen
		self._blenderPlugin.stopPlugin()
		log(gCf(), self, "SpaceController Add-on stopped.")
		return {'FINISHED'}

	@classmethod
	def poll(self, context):
		if self._blenderPlugin.getIsTimerRunning() or self._blenderPlugin.getIsThreadRunning():
			return True
		else:
			return False

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Operator, der beim Anklicken des Menuepunkts 'Unregister Plug-in' ausgefuehrt
# wird.
# fse, 02.12.14

class UnregisterScPluginOp(bpy.types.Operator):
	""" Removes the SpaceController Add-on from Blender """
	bl_idname = "sc_op.unregister_plugin"
	bl_label  = "Unregister SpaceController Add-on"
	_blenderPlugin = None

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	def execute(self, context):
		self._blenderPlugin.stopPlugin()
		self._blenderPlugin.blenderProxy.unregisterAll()
		log(gCf(), self, "SpaceController unregistered.")
		return {'FINISHED'}

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Timer, der die Daten des SpaceControllers aus dem Daemon-Proxy holt und zu
# Blender transferiert.
# Laut Blender-Doku darf das nicht in einem eigenen Thread geschehen, sondern
# muss ueber einen Timer realisiert werden. Die Vorlage stammt aus den
# Blender-Skript-Templates.
# fse, 02.12.14

class ModalBlenderTimerOp(bpy.types.Operator):
	bl_idname = "wm.modal_spacecontrol_timer_operator"
	bl_label = "Modal SpaceControl Timer Operator"
	_blenderPlugin = None
	_cnt = 0
	_timerContext = None # wird zum externen Aufruf von cancel() gebraucht

	#----------------------------------------------------------------------------

	def setBlenderPlugin(self, blenderPlugin):
		self._blenderPlugin = blenderPlugin

	#----------------------------------------------------------------------------

	def getTimerContext(self):
		return self._timerContext

	#----------------------------------------------------------------------------

	def setTimerContext(self, context):
		self._timerContext = context

	#----------------------------------------------------------------------------

	def modal(self, context, event):
		if event.type == 'MOUSEMOVE':
			self._blenderPlugin.blenderProxy.setMousePos(event.mouse_x, event.mouse_y)
		if event.type == 'TIMER':
			self._cnt += 1;
			self._blenderPlugin.blenderProxy.getAndHandleData()
			#print(self._blenderPlugin.timerInterval)
#			print("ModalBlenderTimerOp.modal(): self._cnt: {}".format(self._cnt))
		return {'PASS_THROUGH'}

	#----------------------------------------------------------------------------

	def execute(self, context):
		winMan = context.window_manager
		timer = winMan.event_timer_add(self._blenderPlugin.timerInterval, window=context.window)
		self._blenderPlugin.setTimer(timer)
		winMan.modal_handler_add(self)
		self._blenderPlugin.setIsTimerRunning(True)
		# Context sichern. Er wird bei Cancel gebraucht, steht aber eigentlich
		# nur dann zur Verfuegung, wenn man das Plug-in per Button beendet. Wenn
		# es durch Deaktivieren des Add-Ons im Preferences-Dialog beendet werden
		# soll, gibt es keinen Context, wir nehmen daher immer den hier
		# gespeicherten:
		ModalBlenderTimerOp._timerContext = context
		return {'RUNNING_MODAL'}

	#----------------------------------------------------------------------------

	def cancel(self, context):
		timer = self._blenderPlugin.getTimer()
		context = ModalBlenderTimerOp._timerContext
		winMan = context.window_manager
		log(gCf(), self, "timer: {}".format(timer))
		winMan.event_timer_remove(timer)
		self._blenderPlugin.setIsTimerRunning(False)
		# Pruefen, wie das cancel() ausgeloest wurde:
		# Der Stacktrace ist normalerweise (beim Ausloesen durch unser "Stop"
		# im Menue oder durch Unchecken des Add-Ons im Preferences-Dialog)
		# laenger als drei Elemente. Ist er kuerzer, wurde das cancel() durch
		# Laden einer neuen *.blend-Datei oder durch das
		# Schliessen von Blender ausgeloest. Im zweiten Fall muss sich unser
		# Plug-in gesittet beenden:
		i = 0
		for line in traceback.format_stack():
			#print(line.strip())
			i += 1
		if i == 1:
			log(gCf(), self, "Timer cancelled by closing Blender.")
			self._blenderPlugin.signalTimerKilled() # Blender geschlossen => Plug-in beenden
		elif i == 3:
			pass
			log(gCf(), self, "Timer cancelled by loading file.")
			#self._blenderPlugin.signalTimerKilled() # auch Datei/Neu => Plug-in beenden
		elif i == 5:
			pass
			log(gCf(), self, "Timer cancelled by stopping add-on.")
		elif i == 8:
			pass
			log(gCf(), self, "Timer cancelled by unchecking add-on in preferences dialog.")
		else:
			pass
			log(gCf(), self, "Timer cancelled by unknown event (i={}).". format(i))

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Klasse zur Ausgabe einer Fehlermeldung.
# Vorlage: http://wiki.blender.org/index.php/Dev:2.5/Py/Scripts/Cookbook/Code_snippets/Interface
# fse, 21.01.15
# fse, 20.12.20: invoke_popup() vertraegt keinen height-Parameter mehr; 
#                "Failed = StringProperty()" durch "Failed: StringProperty()"
#                ersetzt.

class MessageOp(bpy.types.Operator):
	bl_idname = "error.message"
	bl_label = "Failed"
	Failed: StringProperty()

	def execute(self, context):
		self.report({'INFO'}, self.Failed)
		log(gCf(), self, self.Failed)
		return {'FINISHED'}

	def invoke(self, context, event):
		wMan = context.window_manager
		return wMan.invoke_popup(self, width=600)

	def draw(self, context):
		self.layout.label(text="SpaceControl Error Message (see console window for details)")
		rw = self.layout.row()
		rw.prop(self, "Failed")

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Repraesentiert eine Struktur mit den Standarddaten.
# Fuer jedes angeschlossene Geraet wird ein Element davon erzeugt, diese
# werden in der Liste _sharedStdData[] in BlenderPlugin abgelegt.
# fse, 31.07.17

class StdData:
	pass

	def __init__(self, devIdx=None, x=None, y=None, z=None, a=None, b=None, c=None, traLmh=None, rotLmh=None, event=None, tvSec=None, tvUsec=None, serNo=None):
		if devIdx == None:
			self.devIdx = -1
		else:
			self.devIdx = devIdx
		if x == None:
			self.x = 0
		else:
			self.x = x
		if y == None:
			self.y = 0
		else:
			self.y = y
		self.traLmh = traLmh
		if z == None:
			self.z = 0
		else:
			self.z = z
		self.traLmh = traLmh
		if a == None:
			self.a = 0
		else:
			self.a = a
		self.traLmh = traLmh
		if b == None:
			self.b = 0
		else:
			self.b = b
		self.traLmh = traLmh
		if c == None:
			self.c = 0
		else:
			self.c = c
		if traLmh == None:
			self.traLmh = 0
		else:
			self.traLmh = traLmh
		if rotLmh == None:
			self.rotLmh = 0
		else:
			self.rotLmh = rotLmh
		if event == None:
			self.event = -1
		else:
			self.event = event
		if tvSec == None:
			self.tvSec = 0
		else:
			self.tvSec = tvSec
		if tvUsec == None:
			self.tvUsec = 0
		else:
			self.tvUsec = tvUsec
		if serNo == None:
			self.serNo = "unknown"
		else:
			self.serNo = serNo
		
		self.isDtaNull     = True
		self.isLastDtaNull = True
		self.isHandled     = True

	def log(self, currentFrame):
#		if self.x != 0 or self.y != 0 or self.z != 0 or self.a != 0 or self.b != 0 or self.c != 0 or self.event != -1:
			log(currentFrame, self, str(self.devIdx) + ": " + str(self.x) + ", " + str(self.y) + ", " + str(self.z) + ", " + str(self.a) + ", " + str(self.b) + ", " + str(self.c) + ", " + str(self.traLmh) + ", " + str(self.rotLmh) + ", " + str(self.event) + ", " + str(self.tvSec) + ", " + str(self.tvUsec) + ", handled: " + str(self.isHandled))
			pass

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

# Klasse zur Kommunikation des Plug-ins mit Blender.
# fse, 04.12.14
# fse, 29.11.20: Matrix-Multiplikationsoperator "*" durch "@" ersetzt.

class BlenderProxy:
	_blenderPlugin     = None
	_sharedStdDataList = None

	# Punkte brauchen als vierte Koordinate eine 1, Vektoren eine 0; nur dann
	# funktionieren Koordinatentransformationen gemaess v'=_viewMat@v:
	_target           = mathutils.Vector((0, 0, 0, 1)) # Punkt, auf den die Kamera blickt, in Weltkoordinaten
	_ux               = mathutils.Vector((1, 0, 0, 0)) # Einheitsvektor in x-Richtung
	_uy               = mathutils.Vector((0, 1, 0, 0)) # Einheitsvektor in y-Richtung
	_uz               = mathutils.Vector((0, 0, 1, 0)) # Einheitsvektor in z-Richtung
	_origin           = mathutils.Vector((0, 0, 0, 1))
	_viewMat          = mathutils.Matrix()
	_region3d         = None
	_isHorizonLock    = False
	_isHoverMode      = False
	_isAutoFit        = False                          # das ist der Parameter mIsAutoFit aus den AdvancedSettings
	_doAutoFitDelay   = 0                              # wenn groesser 0 sagt dieses Flag, dass der Timer einmalig Fit ausfuehren soll
	_moveMode         = 0
	_dist             = 10
	_mouseX           = 0
	_mouseY           = 0
	_lastArea         = None
	_last3dViewArea   = None
	_lastRegion       = None
	_areaTypes        = [ 'VIEW_3D', 'CLIP_EDITOR', 'NODE_EDITOR', 'LOGIC_EDITOR', \
	                      'SEQUENCE_EDITOR', 'NLA_EDITOR', 'DOPESHEET_EDITOR', \
								 'GRAPH_EDITOR','TIMELINE', 'OUTLINER', 'IMAGE_EDITOR' ]
	_mnuCmdList       = []
	_timer            = None
	_r                = mathutils.Matrix.Rotation(math.pi/2, 4, 'X')            # fuer Bones-Kinder in transformObject()
	_t                = mathutils.Matrix.Translation(mathutils.Vector((0,1,0))) # fuer Bones-Kinder in transformObject()
	_ri               = mathutils.Matrix.Rotation(math.pi/2, 4, 'X').inverted()            # fuer Bones-Kinder in transformObject()
	_ti               = mathutils.Matrix.Translation(mathutils.Vector((0,0,1))).inverted() # fuer Bones-Kinder in transformObject()
	_rt               = _r @ _t                                                            # fuer Bones-Kinder in transformObject()
	_riti             = _ri @ _ti                                                          # fuer Bones-Kinder in transformObject()
	
	def getLastRegion(self):
		return self._lastRegion
	
	#----------------------------------------------------------------------------

	def __init__(self, blenderPlugin, sharedStdDataList):
		log(gCf(), self, "Constructing ...")
		self._blenderPlugin     = blenderPlugin
		self._sharedStdDataList = sharedStdDataList
#		self._sharedData        = blenderPlugin
		self.publishBlenderPluginObject(blenderPlugin)
		SpaceControlMenu.bl_label = "SpaceController" # init setzt Object-Move-Mode zurueck, auch Label muss beim Laden von Dateien zurueckgesetzt werden
		self.registerAll()
		self.createMnuFile()
		self.loadMnuFile()
		log(gCf(), self, "Constructed.")

	#-------------------------------------- -------------------------------------
	
	# Macht allen Singleton-Klassen das BlenderPlugin-Objekt bekannt.
	# fse, 09.01.15
	
	def publishBlenderPluginObject(self, blenderPlugin):
		log(gCf(), self, "Invoked.")
		SpaceControlMenu            .setBlenderPlugin(SpaceControlMenu            , blenderPlugin)
		
		RotCenterMenu               .setBlenderPlugin(RotCenterMenu               , blenderPlugin)
		RotCenterDefaultScPluginOp  .setBlenderPlugin(RotCenterDefaultScPluginOp  , blenderPlugin)
		RotCenterCursorScPluginOp   .setBlenderPlugin(RotCenterCursorScPluginOp   , blenderPlugin)
		RotCenterSelectionScPluginOp.setBlenderPlugin(RotCenterSelectionScPluginOp, blenderPlugin)

		ReduceSensScPluginOp        .setBlenderPlugin(ReduceSensScPluginOp        , blenderPlugin)
		AffectOnly3dViewScPluginOp  .setBlenderPlugin(AffectOnly3dViewScPluginOp  , blenderPlugin)

		BindDevicesMenu             .setBlenderPlugin(BindDevicesMenu             , blenderPlugin)
		AutoSendGapScPluginOp       .setBlenderPlugin(AutoSendGapScPluginOp       , blenderPlugin)
		RememberViewScPluginOp      .setBlenderPlugin(RememberViewScPluginOp      , blenderPlugin)
		BindDevice0Op               .setBlenderPlugin(BindDevice0Op               , blenderPlugin)
		BindDevice1Op               .setBlenderPlugin(BindDevice1Op               , blenderPlugin)
		BindDevice2Op               .setBlenderPlugin(BindDevice2Op               , blenderPlugin)
		BindDevice3Op               .setBlenderPlugin(BindDevice3Op               , blenderPlugin)
		BindDevice4Op               .setBlenderPlugin(BindDevice4Op               , blenderPlugin)
		BindDevice5Op               .setBlenderPlugin(BindDevice5Op               , blenderPlugin)
		BindDevice6Op               .setBlenderPlugin(BindDevice6Op               , blenderPlugin)
		BindDevice7Op               .setBlenderPlugin(BindDevice7Op               , blenderPlugin)
		BindDevice8Op               .setBlenderPlugin(BindDevice8Op               , blenderPlugin)
		BindDevice9Op               .setBlenderPlugin(BindDevice9Op               , blenderPlugin)

		ObjectMoveModeScPluginOp    .setBlenderPlugin(ObjectMoveModeScPluginOp    , blenderPlugin)
		MultiRotCentersOmmScPluginOp.setBlenderPlugin(MultiRotCentersOmmScPluginOp, blenderPlugin)
		GenKeyframesOmmScPluginOp   .setBlenderPlugin(GenKeyframesOmmScPluginOp   , blenderPlugin)

		StartScPluginOp             .setBlenderPlugin(StartScPluginOp             , blenderPlugin)
		StopScPluginOp              .setBlenderPlugin(StopScPluginOp              , blenderPlugin)
		UnregisterScPluginOp        .setBlenderPlugin(UnregisterScPluginOp        , blenderPlugin)
		ModalBlenderTimerOp         .setBlenderPlugin(ModalBlenderTimerOp         , blenderPlugin)
	
	#-------------------------------------- -------------------------------------
	
	# Registriert die Singleton-Klassen (Menuepunkte und Menues) bei Blender.
	# fse, 09.01.15
	# fse, 12.11.20: INFO_HT_header durch VIEW3D_MT_editor_menus ersetzt.

	def registerAll(self):
		log(gCf(), self, "Invoked.")
		bpy.utils.register_class(ReduceSensScPluginOp)
		bpy.utils.register_class(AffectOnly3dViewScPluginOp)
		bpy.utils.register_class(AutoSendGapScPluginOp)
		bpy.utils.register_class(RememberViewScPluginOp)
		bpy.utils.register_class(ObjectMoveModeScPluginOp)
		bpy.utils.register_class(MultiRotCentersOmmScPluginOp)
		bpy.utils.register_class(GenKeyframesOmmScPluginOp)
		bpy.utils.register_class(RotCenterMenu)
		bpy.utils.register_class(RotCenterDefaultScPluginOp)
		bpy.utils.register_class(RotCenterCursorScPluginOp)
		bpy.utils.register_class(RotCenterSelectionScPluginOp)
		bpy.utils.register_class(BindDevicesMenu)
		bpy.utils.register_class(BindDevice0Op)
		bpy.utils.register_class(BindDevice1Op)
		bpy.utils.register_class(BindDevice2Op)
		bpy.utils.register_class(BindDevice3Op)
		bpy.utils.register_class(BindDevice4Op)
		bpy.utils.register_class(BindDevice5Op)
		bpy.utils.register_class(BindDevice6Op)
		bpy.utils.register_class(BindDevice7Op)
		bpy.utils.register_class(BindDevice8Op)
		bpy.utils.register_class(BindDevice9Op)
		bpy.utils.register_class(StartScPluginOp)
		bpy.utils.register_class(StopScPluginOp)
		bpy.utils.register_class(UnregisterScPluginOp)
		bpy.utils.register_class(SpaceControlMenu)
		bpy.utils.register_class(MessageOp)
#		bpy.types.INFO_HT_header.append(SpaceControlMenu.showMenu)
		bpy.types.VIEW3D_MT_editor_menus.append(SpaceControlMenu.showMenu)

		# Shortcut Shift-Ctrl-Alt-O fuer "Object Move Mode" definieren, damit wir
		# ihn in "Default.cfg" verwenden koennen. Wenn man unseren Operator
		# verwendet, liegt er in der naechsten Blender-Version wieder woanders.
		# Leider ist es nicht gelungen, den Shortcut auch im Menue anzuzeigen.:
		wMgr = bpy.context.window_manager
		kc = wMgr.keyconfigs.addon
		if kc:
			log(gCf(), self, "Retrieving wMgr.keyconfigs.addon done.")
			km = wMgr.keyconfigs.addon.keymaps.new(name='3D View', space_type='VIEW_3D')
			kmi = km.keymap_items.new(ObjectMoveModeScPluginOp.bl_idname, 'O', 'PRESS', ctrl=True, alt=True, shift=True)
			gShortCuts.append((km, kmi))
		else:
			log(gCf(), self, "Retrieving wMgr.keyconfigs.addon failed!")
			pass

	#-------------------------------------- -------------------------------------
	
	# Erzeugt die MNU-Datei, falls noch nicht vorhanden.
	# Dies ist ein alternativer Ansatz, der als Basis die Liste der Keymaps
	# in der Keyconfiguration heranzieht, um alle Operatoren und deren
	# Tastenkuerzel zu ermitteln.
	# Hat sich aber auch nicht besser bewaehrt als der andere: Viele Dubletten,
	# Liste ist unsortiert und unuebersichtlich (wie im User-Preferences-Dialog
	# in Blender im Reiter "Input").
	# fse, 09.06.15
	
#	def createMnuFile(self):
#		mnuFilePath = self._blenderPlugin.getMnuFilePath()
#		if os.path.exists(mnuFilePath):
#			return
#		
#		print("BlenderProxy.createMnuFile(): Creating file '{}'.".format(mnuFilePath))
#		kc = bpy.context.window_manager.keyconfigs[0]
#		groupListList = self.readOutAllKeymapsFromKeyconfig(kc)
#		f = open(mnuFilePath, 'w')
#		self.writeModes(f)
#		self.writeAllOperators(f, groupListList)
#		f.close()
#
#	#------------------------- ------------------------- ------------------------
#
#	def readOutAllKeymapsFromKeyconfig(self, kc):
#		groupListList = []
#		for km in kc.keymaps:
#			print("  ", km.name)
#			groupList = self.readOutAllKeymapitemsFromKeymap(km)
#			groupListList.append(groupList)
#		return groupListList
#
#	#------------------ ------------------- ------------------- -----------------
#
#	def readOutAllKeymapitemsFromKeymap(self, km):
#		groupList = [ km.name ]
#		for kmi in km.keymap_items:
#			opStr = kmi.name
#			opWithStrippedSample = kmi.idname
#			mod = ""
#			if kmi.shift:
#				mod += "Shift "
#			if kmi.ctrl:
#				mod += "Ctrl "
#			if kmi.alt:
#				mod += "Alt "
#			comment = kmi.value + " " + mod + " " + kmi.type
#			opStrOpSmplCmmnt     = [opStr, opWithStrippedSample, comment]
#			print(opStrOpSmplCmmnt)
#			groupList.append(opStrOpSmplCmmnt)
#		return groupList
	
	#-------------------------------------- -------------------------------------
	
	# Erzeugt die MNU-Datei, falls noch nicht vorhanden.
	# fse, 21.01.15
	
	def createMnuFile(self):
		mnuFilePath = self._blenderPlugin.getMnuFilePath()
		if os.path.exists(mnuFilePath):
			return
		
		log(gCf(), self, "Creating file '{}'.".format(mnuFilePath))
		groupListList = self.readOutAllOperators()
		f = open(mnuFilePath, 'w')
		self.writeModes(f)
		self.writeAllOperators(f, groupListList)
		f.close()

	#------------------------- ------------------------- ------------------------
	
	# Liest saemtliche Blender-Operatoren aus und gruppiert sie in eine Liste,
	# die die Gruppen enthaelt. Gruppen sind z. B. "mesh" (bpy.ops.mesh) oder
	# "view3d" (bpy.ops.view3d), aber auch unsere eigene Gruppe "sc_op"
	# (bpy.ops.sc_op).
	# fse, 21.01.15
	
	def readOutAllOperators(self):
		groupListList = []
		for groupStr in dir(bpy.ops):
			groupList = self.createGroupList(groupStr)
			groupListList.append(groupList)
		return groupListList
		
	#------------------ ------------------- ------------------- -----------------
	
	# Erzeugt die Liste aller Gruppen ("mesh", "view3d", "sc_op", ...); jede
	# Guppe ist wieder eine Liste, die die Operatoren enthaelt.
	# Ein typischer Operator-String, den man mit _get_doc.__str__() aus z. B.
	# dir(bpy.ops.mesh.primitive_cube_add) geholt hat, sieht so aus: 
	#    # Construct a cube mesh
	#    bpy.ops.mesh.primitive_cube_add(radius=1, view_align=False,
	#    enter_editmode=False, location=(0, 0, 0), rotation=(0, 0, 0),
	#    layers=(False, False, False, False, False, False, False, False, False,
	#    False, False, False, False, False, False, False, False, False, False,
	#    False))>
	# Nach einem '#' folgt ein Kommentar bis '\n', dann folgt der Operator mit
	# einer beispielhaften Parameterliste mit Default-Parametern.
	# fse, 21.01.15
	# fse, 26.03.18: Es hat sich herausgestellt, dass es auch ein paar
	#                mehrzeilige Kommentare gibt, sichtbar z. B. bei
	#                eval("bpy.ops.texture.patern_rename"). Daher wird ab sofort
	#                statt nach dem ersten CR nach "bps.ops." gesucht und alles
	#                davor als Kommentar gewertet.

	def createGroupList(self, grpStr):
		groupList = [ grpStr ]
		groupStr = "bpy.ops." + grpStr
		log(gCf(), self, "groupStr: " + str(groupStr))		
		groupEvl = eval(groupStr)
		for opStr in dir(groupEvl):
			operatorStr          = groupStr + "." + opStr
			operatorVal          = eval(operatorStr)
			line                 = operatorVal._get_doc.__str__()
			hashPos              = line.find('#')
#			crPos                = line.find('\n')
#			comment              = line[hashPos+2:crPos]
#			opWithSample         = line[crPos+1:-1]
			opPos                = line.find('bpy.ops.')
			comment              = line[hashPos+2:opPos-1].replace('\n', ' ')
			opWithSample         = line[opPos:-1]
			opWithStrippedSample = self.stripDefaultArgs(opWithSample)
			#print(opWithSample)
			#print(opWithStrippedSample)
			opStrOpSmplCmmnt     = [opStr, opWithStrippedSample, comment]
			groupList.append(opStrOpSmplCmmnt)
		#groupList.sort()
		return groupList
		
	#-------------- --------------- --------------- --------------- -------------

	# Entfernt alle Default-Parameter aus der Parameterliste eines Operators.
	# Aus
	#    op(abc, def, ghi = 'test', jkl = (False, True, False), xyz)
	# wird
	#    op(abc, def, xyz)
	# fse, 21.01.15
	
	def stripDefaultArgs(self, op):
		argList    = self.extractArgList(op)
		opStripped = self.replaceArgs(op, argList)
		return opStripped
		
	#----------- ------------- ------------ ------------ ------------- ----------
	
	# Liest die Argumente eines Operators in eine Liste aus.
	# Beispiel:
	#   op(abc=2, def, ghi = 'test', jkl = (False, True, False), uvw, xyz=3)
	# Ergebnis ist
	#   ['abc=2', ' def', " ghi = 'test'", ' jkl = (False, True, False)', ' uvw', ' xyz=3'].
	# fse, 21.01.15

	def extractArgList(self, op):
		argList    = []
		pntr     = 0 # Zeiger auf Character im String
		cnt      = 0 # Zaehler fuer Klammerebene
		startPos = 0 # Startposition eines Parameters
		for c in op:
			# oeffnende Klammern zaehlen:
			if c == '(':
				cnt += 1 
				if cnt == 1:
					# nur innerhalb der ersten Klammerebene sind Parameter, nach der
					# Klammer startet der Parameter:
					startPos = pntr + 1 
			# ein Parameter endet vor einem Komma oder einer schliessenden Klammer;
			# letzteres aber nur, wenn die Klammerebene 1 ist:
			if (c == ',' or c == ')') and cnt == 1:
				# ein Parameter
				arg = op[startPos:pntr] 
				argList.append(arg)
				startPos = pntr + 1
			# auf schliessende Klammern pruefen:
			if c == ')':
				cnt -= 1
			pntr += 1 # Stringzeiger inkrementieren
		return argList

	#----------- ------------- ------------ ------------ ------------- ----------
	
	# Ersetzt alle Default-Argumente im Operator.
	# Beispiel unten:
	#   op(abc=2, def, ghi = 'test', jkl = (False, True, False), uvw, xyz=3)
	# wird zu 
	#   op(def, uvw)
	# fse, 21.01.15

	def replaceArgs(self, op, argList):
		for arg in argList:
			if '=' in arg:
				# Default-Argumente enthalten '=', nur diese entfernen
				log(gCf(), self, arg)
				op = op.replace(arg, "")
		# Argumente entfernt, es bleiben aber jetzt ueberfluessige Kommas zurueck:
		#   op(, def,,, uvw,)
		# die muessen entfernt werden:
		while ",," in op:
			op = op.replace(",,", ",") # => op(, def, uvw,)
		op = op.replace("(, ", "(")   # => op(def, uvw,)
		op = op.replace(",)", ")")    # => op(def, uvw)
		return op

	#------------------------- ------------------------- ------------------------
	
	def writeModes(self, f):
		f.write("# SpaceControl menu file containing Blender's operators\n")
		f.write("\n")
		f.write("[Modes]\n")
		f.write("3D View Full    = 1\n")
		f.write("Animation       = 2\n")
		f.write("Compositing     = 3\n")
		f.write("Default         = 4\n")
		f.write("Game Logic      = 5\n")
		f.write("Motion Tracking = 6\n")
		f.write("Scripting       = 7\n")
		f.write("UV Editing      = 8\n")
		f.write("Video Editing   = 9\n")
		f.write("\n")

	#------------------------- ------------------------- ------------------------
	
	def writeAllOperators(self, f, groupListList):
		w = 40
		f.write("[MenuBar]\n")
		for groupList in groupListList:
			grp = groupList[0]
			line = "0" + grp.ljust(w) + "= {}\n"
			f.write(line)
			for opStrOpSmplCmmnt in groupList:
				if opStrOpSmplCmmnt == grp:
					continue
				opStr    = opStrOpSmplCmmnt[0]
				opSmpl   = opStrOpSmplCmmnt[1]
				comment  = opStrOpSmplCmmnt[2]
				line = "1" + opStr.ljust(w) + "= {}" + opSmpl + "|" + comment + "\n"
				f.write(line)
				
	#-------------------------------------- -------------------------------------
	
	def loadMnuFile(self):
		self._mnuCmdList =  []
		f = open(self._blenderPlugin.getMnuFilePath(), 'r')
		for line in f:
			if line[0] == '1':
				ketPos  = line.find('}')
				pipePos = line.find('|')
				opSmpl  = line[ketPos+1:pipePos]
				self._mnuCmdList.append(opSmpl)
		f.close()
	
	#----------------------------------------------------------------------------
	
	def __del__(self):
		log(gCf(), self, "Destructing ...")
		self.unregisterAll()
		log(gCf(), self, "Destructed.")

	#-------------------------------------- -------------------------------------

	def unregisterAll(self):
		log(gCf(), self, "Invoked.")
		for km, kmi in gShortCuts:
			km.keymap_items.remove(kmi)
		gShortCuts.clear()

#		bpy.types.INFO_HT_header.remove(SpaceControlMenu.showMenu)
		bpy.types.VIEW3D_MT_editor_menus.remove(SpaceControlMenu.showMenu)
		bpy.utils.unregister_class(MessageOp)
		bpy.utils.unregister_class(SpaceControlMenu)
		bpy.utils.unregister_class(UnregisterScPluginOp)
		bpy.utils.unregister_class(StopScPluginOp)
		bpy.utils.unregister_class(StartScPluginOp)
		bpy.utils.unregister_class(RotCenterSelectionScPluginOp)
		bpy.utils.unregister_class(RotCenterCursorScPluginOp)
		bpy.utils.unregister_class(RotCenterDefaultScPluginOp)
		bpy.utils.unregister_class(RotCenterMenu)
		bpy.utils.unregister_class(BindDevice0Op)
		bpy.utils.unregister_class(BindDevice1Op)
		bpy.utils.unregister_class(BindDevice2Op)
		bpy.utils.unregister_class(BindDevice3Op)
		bpy.utils.unregister_class(BindDevice4Op)
		bpy.utils.unregister_class(BindDevice5Op)
		bpy.utils.unregister_class(BindDevice6Op)
		bpy.utils.unregister_class(BindDevice7Op)
		bpy.utils.unregister_class(BindDevice8Op)
		bpy.utils.unregister_class(BindDevice9Op)
		bpy.utils.unregister_class(BindDevicesMenu)
		bpy.utils.unregister_class(GenKeyframesOmmScPluginOp)
		bpy.utils.unregister_class(MultiRotCentersOmmScPluginOp)
		bpy.utils.unregister_class(ObjectMoveModeScPluginOp)
		bpy.utils.unregister_class(RememberViewScPluginOp)
		bpy.utils.unregister_class(AutoSendGapScPluginOp)
		bpy.utils.unregister_class(AffectOnly3dViewScPluginOp)
		bpy.utils.unregister_class(ReduceSensScPluginOp)

	#----------------------------------------------------------------------------
	
	# Erzeugt fuer ein neues Geraet einen neuen Eintrag im Untermenue
	# "Bind Selection to Device".
	# fse, 07.08.17

	def addDevice(self, stdData):
		idx   = stdData.devIdx
		serNo = stdData.serNo
		op    = "BindDevice" + str(idx) + "Op"  # Namen zusammensetzen: z. B. BindDevice1Op
		bpy.utils.unregister_class(eval(op))    # altes Element entfernen
		eval(op).bl_label=serNo                 # Seriennummer als Label verwenden
		bpy.utils.register_class(eval(op))      # und Element neu registrieren
		self._blenderPlugin.logStdDataList(gCf())
	
	#----------------------------------------------------------------------------

	# Entfernt den Eintrag eines abgesteckten Geraetes im Untermenue
	# "Bind Selection to Device".
	# fse, 07.08.17

	def removeDevice(self, stdData):
		idx = stdData.devIdx
		op  = "BindDevice" + str(idx) + "Op"    # Namen zusammensetzen: z. B. BindDevice1Op
		bpy.utils.unregister_class(eval(op))    # altes Element entfernen
		eval(op).bl_label="Device " + str(idx)  # Seriennummer entfernen
		bpy.utils.register_class(eval(op))      # und Element neu registrieren
		self._blenderPlugin.logStdDataList(gCf())
	
	#----------------------------------------------------------------------------
	
	def setTimer(self, timer):
		self._timer = timer
	
	def getTimer(self):
		return self._timer
	
	#----------------------------------------------------------------------------
	
	# Richtet einen Handler fuer das Scene-Update-Post-Ereignis ein. Siehe
	# Kommentar bei BlenderPlugin.startPluginFromInit().
	# fse, 12.01.15
	# fse, 20.11.20: sceneUpdatePost durch depsgraph_update_post ersetzt.
	# fse, 13.12.20: Obsolet; das Plug-in wird jetzt direkt gestartet.
	
	def addSceneUpdatePostHandler(self):
		if scSceneUpdatePostHandler in bpy.app.handlers.depsgraph_update_post:
			log(gCf(), self, "scSceneUpdatePostHandler is already hooked to bpy.app.handlers.scene_update_post!")
			pass
		else:
			bpy.app.handlers.depsgraph_update_post.append(scSceneUpdatePostHandler)
			log(gCf(), self, "scSceneUpdatePostHandler added.")

	#----------------------------------------------------------------------------

	# Entfernt den Handler fuer das Scene-Update-Post-Ereignis. Siehe Kommentar
	# bei BlenderPlugin.startPluginFromInit().
	# fse, 12.01.15
	# fse, 20.11.20: sceneUpdatePost durch depsgraph_update_post ersetzt.
	# fse, 13.12.20: Obsolet; das Plug-in wird jetzt direkt gestartet.

	def removeSceneUpdatePostHandler(self):
		if scSceneUpdatePostHandler in bpy.app.handlers.depsgraph_update_post:
			bpy.app.handlers.depsgraph_update_post.remove(scSceneUpdatePostHandler)
			log(gCf(), self, "scSceneUpdatePostHandler() removed.")
		else:
			log(gCf(), self, "scSceneUpdatePostHandler() is not hooked to bpy.app.handlers.load_post!")
			pass

	#----------------------------------------------------------------------------
	
	# Registriert und startet den Timer, der die Kappendaten aus dem
	# Shared-Data-Objekt holt und auswertet und meldet unsere Hooks bei Blender
	# an.
	# fse, 04.12.14
	
	def startBlenderProxy(self):
		log(gCf(), self, "Invoked.")
		self._lastArea = None
		self._last3dViewArea = None
		self.startTimer()
		self.hookHandlers()
	
	#-------------------------------------- -------------------------------------
	
	# Startet unseren Timer im ModalBlenderTimerOp.
	# fse, 19.12.14
	
	def startTimer(self):
		if self._blenderPlugin.getIsTimerRunning():
			log(gCf(), self, "Timer is running already!")
			pass
		else:
			if self.isOperatorRegistered(ModalBlenderTimerOp.bl_idname):
				log(gCf(), self, "ModalBlenderTimerOp is registered already!")
				bpy.utils.unregister_class(ModalBlenderTimerOp)
			bpy.utils.register_class(ModalBlenderTimerOp)
			bpy.ops.wm.modal_spacecontrol_timer_operator()

	#------------------------- ------------------------- ------------------------
	
	# Prueft, ob ein Operator namens bl_idname vorhanden ist.
	# \param bl_idname Name des Operators wie in der Klasse definiert
	# \return true, wenn Operator vorhanden, sonst false
	# fse, 18.12.14
	
	def isOperatorRegistered(self, bl_idname):
		ret = False
		# getattr(bpy.ops.wm, "name") scheint nicht zu funktionieren, also
		# umstaendlich, z. B. fuer "wm.modal_spacecontrol_timer_operator":
		pntIdx   = bl_idname.find('.')   # pntIdx = 2
		category = bl_idname[0:pntIdx]   # = "wm"
		idname   = bl_idname[pntIdx+1:]  # = "modal_spacecontrol_timer_operator"
		category = "bpy.ops." + category # = "bpy.ops.wm"
		attrs    = dir(eval(category))   # = alle Attribute in bpy.ops.wm
		if idname in attrs:
			ret = True
		return ret
		
	#-------------------------------------- -------------------------------------
	
	# Meldet unsere Handler fuer das Behandeln der beiden Datei-Lade-Ereignisse
	# am System an. Siehe Kommentar bei scLoadPreHandler().
	# fse, 19.12.14
	
	def hookHandlers(self):
		if scLoadPreHandler in bpy.app.handlers.load_pre:
			log(gCf(), self, "scLoadPreHandler is already hooked to bpy.app.handlers.load_pre!")
			pass
		else:
			bpy.app.handlers.load_pre.append(scLoadPreHandler)
			log(gCf(), self, "scLoadPreHandler hooked")
		if scLoadPostHandler in bpy.app.handlers.load_post:
			log(gCf(), self, "scLoadPostHandler is already hooked to bpy.app.handlers.load_post!")
			pass
		else:
			bpy.app.handlers.load_post.append(scLoadPostHandler)
			log(gCf(), self, "scLoadPostHandler hooked")
			
	
	#----------------------------------------------------------------------------
	
	# Beendet den Timer und meldet ihn bei Blender wieder ab, ebenso unsere
	# beiden Hooks.
	# fse, 19.12.14

	def stop(self):
		log(gCf(), self, "Invoked.")
		self.unhookHandlers()
		self.stopTimer()

	#-------------------------------------- -------------------------------------
	
	# Meldet unsere Handler fuer das Behandeln der beiden Datei-Lade-Ereignisse
	# bei Blender ab. Siehe Kommentar bei scLoadPreHandler().
	# fse, 19.12.14
	
	def unhookHandlers(self):
		log(gCf(), self, "Invoked.")
		if scLoadPostHandler in bpy.app.handlers.load_post:
			bpy.app.handlers.load_post.remove(scLoadPostHandler)
		else:
			log(gCf(), self, "scLoadPostHandler() is not hooked to bpy.app.handlers.load_post!")
			pass
		if scLoadPreHandler in bpy.app.handlers.load_pre:
			bpy.app.handlers.load_pre.remove(scLoadPreHandler)
		else:
			log(gCf(), self, "scLoadPreHandler() is not hooked to bpy.app.handlers.load_pre!")
			pass
		
	#-------------------------------------- -------------------------------------
	
	# Beendet den Timer aus ModalBlenderTimerOp und meldet den Operator bei
	# Blender ab.
	# fse, 19.12.14
	
	def stopTimer(self):
		log(gCf(), self, "Invoked.")
#		if self.isOperatorRegistered(ModalBlenderTimerOp.bl_idname):
		if self._blenderPlugin.getIsTimerRunning():
			log(gCf(), self, str(ModalBlenderTimerOp.getTimerContext(ModalBlenderTimerOp)))
			ModalBlenderTimerOp.cancel(ModalBlenderTimerOp, ModalBlenderTimerOp.getTimerContext(ModalBlenderTimerOp))
			bpy.utils.unregister_class(ModalBlenderTimerOp)
		else:
			log(gCf(), self, "Timer is not running yet!")
			pass
		
	#----------------------------------------------------------------------------
	
	# Setzt die Parameter MoveMode, LockHorizon und Hover. Kopiert aus
	# ScGui.interpretMoveMode() aus "ScGui.java".
	# \param moveMode (0: objscreen, 1: objorigin, 2: camstandard, 3: camtarget,
	#                  10-13: Fix Horizon, 100-103: Schwebe,
	#                  110-113: Lock Horizon und Schwebe)
	# fse, 03.12.14

	def setMoveMode(self, moveMode):
		if moveMode >= 100:
			self._isHoverMode = True;
			moveMode -= 100
		else:
			self._isHoverMode = False;
		if moveMode >= 10:
			self._isHorizonLock = True;
			moveMode -= 10;
		else:
			self._isHorizonLock = False;
		# moveMode hat jetzt Werte aus 0..3 und entspricht genau dem Index der
		# Move-Mode-Combo-Box aus dem Advanced-Settings-Dialog und den IDs
		# OBJ_SCREEN usw. aus ScAdvancedSettings:
		self._moveMode = moveMode;
		log(gCf(), self, "_isHorizonLock: {}".format(self._isHorizonLock))
		log(gCf(), self, "_isHoverMode  : {}".format(self._isHoverMode))
		log(gCf(), self, "_moveMode     : {}".format(self._moveMode))
	
	#----------------------------------------------------------------------------
	
	# Setzt den Parameter _isAutoFit im BlenderProxy.
	# \param isAutoFit kommt aus Konfigurationsdatei
	# fse, 07.07.16
	
	def setIsAutoFit(self, isAutoFit):
		self._isAutoFit = isAutoFit
		
	#----------------------------------------------------------------------------
	
	def toggleReduceSensMode(self):
		self._blenderPlugin.isSlowDownMode = not self._blenderPlugin.isSlowDownMode
		
	#----------------------------------------------------------------------------
	
	def toggle3dViewMode(self):
		self._blenderPlugin.isOnly3dViewActive = not self._blenderPlugin.isOnly3dViewActive
		
	#----------------------------------------------------------------------------
	
	def toggleObjectMoveMode(self):
		self._blenderPlugin.isObjectMoveMode = not self._blenderPlugin.isObjectMoveMode
		if self._blenderPlugin.isObjectMoveMode:
			SpaceControlMenu.bl_label = "SpaceController-O"
		else:
			SpaceControlMenu.bl_label = "SpaceController"
		bpy.utils.unregister_class(SpaceControlMenu)    # altes Element entfernen
		bpy.utils.register_class(SpaceControlMenu)      # und Element neu registrieren

	#----------------------------------------------------------------------------
	
	def toggleRememberView(self):
		self._blenderPlugin.isRememberView = not self._blenderPlugin.isRememberView

	#----------------------------------------------------------------------------
	
	def toggleMultiRotCentersMmm(self):
		self._blenderPlugin.isOmmMultiRotCenter = not self._blenderPlugin.isOmmMultiRotCenter

	#----------------------------------------------------------------------------

	def toggleGenKeyframesOmm(self):
		self._blenderPlugin.isGenKeyframes = not self._blenderPlugin.isGenKeyframes

	#----------------------------------------------------------------------------

	def setRotCenterToDefault(self):
		self._blenderPlugin.rotCenter = 'DEFAULT'

	#----------------------------------------------------------------------------

	def setRotCenterToCursor(self):
		self._blenderPlugin.rotCenter = 'CURSOR'

	#----------------------------------------------------------------------------

	def setRotCenterToSelection(self):
		self._blenderPlugin.rotCenter = 'SELECTION'

	#----------------------------------------------------------------------------

	def setSpecialKeyEvent(self, event):
		self.blenderProxy.setSpecialKeyEvent(moveMode)
		
	#----------------------------------------------------------------------------
	
	# Der BlenderProxy-Timer ermittelt kontinuierlich die aktuellen
	# Mauskoordinaten, um die Region zu finden, ueber der der Mauszeiger gerade
	# steht. Sie werden hier in den Proxy uebertragen.
	# fse, 12.01.15
	
	def setMousePos(self, x, y):
		self._mouseX = x
		self._mouseY = y

	#----------------------------------------------------------------------------
	
	# Holt die aktuellen Geraetedaten aus dem Shared-Data-Objekt und behandelt
	# die Daten entsprechend.
	# Die Funktion wird zyklisch vom Timer ModalBlenderTimerOp aufgerufen und
	# ermittelt auch den Modus, in dem sich Blender gerade befindet. Wir
	# definieren den Modus als den Typ des Screens (= Layout), der gerade aktiv
	# ist, z. B. "Scripting" oder "Default", und den man mit der Drop-Down-Liste
	# im "Info"-Header einstellen kann.
	# fse, 04.12.14
	# fse, 08.07.16: Im Zuge der Doppelklicks auf den Tasten musste auch das
	#                Auto-Fit vom Daemon in die Plug-ins verlagert werden. Hier 
	#                zu beobachtender Effekt: Wenn das Fit ausgefuehrt wird,
	#                bevor das Front, Right, Top usw. auf dem Bildschirm beendet
	#                ist (Blender animiert das), wird dies abgebrochen und das
	#                Fit gemacht. Ein sleep() reicht aber nicht, weil Blender
	#                das Kommando nicht waehrend der Timer-Routine ausfuehrt,
	#                sondern nur ausserhalb. Daher wird jetzt ein Zaehler
	#                eingefuehrt (_doAutoFitDelay), der bei jedem Aufruf von
	#                getAndHandleData() dekrementiert wird, sofern er noch
	#                groesser ist als 0; sobald er auf 0 steht, wird das Fit
	#                ausgefuehrt. handleSpecialKeyEvent() setzt _doAutoFitDelay
	#                auf 15, wenn _isAutoFit true. Experimente haben gezeigt,
	#                dass das Blender anscheinend auch bei grossen Szenen genug
	#                Zeit fuer Front, Right, Top usw. gibt, bevor dann das Fit
	#                gemacht wird (10 ist nicht ausreichend).

	def getAndHandleData(self):
		stdData = None # damit der Obfuscator die Variable findet
#		log(gCf(), self, "Invoked.")
		for stdData in self._sharedStdDataList:
			if stdData.isHandled:
				if not self._blenderPlugin is None:
					self._blenderPlugin.setBlenderMode(bpy.context.screen.name)
				continue
			else:
				pass
				stdData.isHandled = True
#			if not stdData.isDtaNull:
#				stdData.log(gCf())
			x     = stdData.x
			y     = stdData.y
			z     = stdData.z
			a     = stdData.a
			b     = stdData.b
			c     = stdData.c
			event = stdData.event
			area  = self.getActiveArea()
			if self._doAutoFitDelay > 0:
				self._doAutoFitDelay -= 1
				log(gCf(), self, "_isDoAutoFit active ...")
				override = self.setContextOverride(area)
				if self._doAutoFitDelay <= 0:
					self.handleDevFit(area, override)
			if event != self._blenderPlugin.NOTHING_CHANGED:
				self.handleEvent(event, area)
			if stdData.isLastDtaNull and stdData.isDtaNull:
				pass # wenn die aktuellen Daten und die vorherigen 0 sind, nichts tun (sonst blockieren wir die Bewegung mit der normalen Maus)
			else:
				deviceProxy      = self._blenderPlugin.daemonProxy._devDic[stdData.serNo]
				selectionList    = deviceProxy._selectionList
				isDeviceBound    = len(selectionList)				
				isObjectMoveMode = self._blenderPlugin.isObjectMoveMode
				if isObjectMoveMode or isDeviceBound:
					if isDeviceBound:
						if self._blenderPlugin.isRememberView:
							region3d = deviceProxy.getRegion()
						else:
							region3d = self.getActiveRegion(True)
						self.transformObjects(area, region3d, x, y, z, a, b, c, selectionList)	
					else:
						isMoved = self.transformObjectsOrSubobjects(area, x, y, z, a, b, c)
						if not isMoved:
							self.transformView(area, x, y, z, a, b, c)
				else:
					self.transformView(area, x, y, z, a, b, c)

	#-------------------------------------- -------------------------------------
	
	# Ermittelt die Area, in der sich der Mauszeiger gerade befindet. In Blender
	# enthaelt das Hauptfenster einen "Screen" (=Layout), der aus mehreren
	# Areas besteht. Der Scripting-Screen besteht z. B. aus den sechs des Typs
	# (areas[x].type) 'INFO', 'PROPERTIES', 'VIEW_3D', 'CONSOLE', 'OUTLINER' und
	# 'TEXT_EDITOR', sofern man keine weiteren erzeugt hat (z. B. durch Teilung
	# des 3D-Views in mehrere).
	# \return die Area, in der der Mauszeiger steht
	# fse, 16.12.14
	
	def getActiveArea(self):
		ret = self._lastArea
		for area in bpy.context.screen.areas:
			log(gCf(), self, "area: " + area.type)
			if area.type in self._areaTypes:
				mx = self._mouseX
				my = self._mouseY
				ax = area.x
				ay = area.y
				aw = area.width
				ah = area.height
				if mx >= ax and mx < ax + aw and my >= ay and my < ay + ah:
					#print("area.type: {}, mx: {}, my: {}, ax: {}, ay: {}, aw: {}, ah: {}".format(area.type, mx, my, ax, ay, aw, ah))
					if self._blenderPlugin.isOnly3dViewActive and area.type != 'VIEW_3D':
						pass
					else:
						ret = self._lastArea = area
						if area.type == 'VIEW_3D':
							self._last3dViewArea = area
		# Sehr seltsamer Effekt: Nachdem der Obfuscator (nur unter Windows) die
		# folgende Log-Zeile entfernt hat, stuerzt Blender regelmaessig ab, wenn
		# man im UV-Editor die (relativ grosse) Datei
		# "C:\Users\fs\Documents\SC_Workspace\Blender_Models\LogoAnim\SpaceControl.png"
		# zoomt. Der Obfuscator behaelt sie daher ab sofort durch Abfrage auf
		# Uebereinstimmung der Zeile drin; s. auch Kommentare in readFile() in
		# "obfuscator.py":
		#log(gCf(), self, "activeArea: " + str(ret))    
		if ret != None:
			log(gCf(), self, "activeArea: " + ret.type)
		return ret
		
	#-------------------------------------- -------------------------------------
	
	# Behandelt ein vom Geraet ausgeloestes Ereignis.
	# \param event Ereignis-Code wie in event_type.h des Daemons definiert.
	# \param area  die Area, ueber der der Mauszeiger gerade steht
	# fse, 04.12.14
	
	def handleEvent(self, event, area):
#		log(gCf(), self, "event: {}".format(event))
		override = self.setContextOverride(area)
		self._blenderPlugin.event = self._blenderPlugin.NOTHING_CHANGED
		if (event >= self._blenderPlugin.DEV_FRONT and event <= self._blenderPlugin.DEV_FIT) or (event >= self._blenderPlugin.DEV_BACK and event <= self._blenderPlugin.DEV_BOTTOM):
			self.handleSpecialKeyEvent(event, area, override)
		if event >= self._blenderPlugin.APPL_FUNC_START:
			self.handleKeyEvent(event, override)

	#------------------------- ------------------------- ------------------------
	
	# Wenn der Mauszeiger nicht ueber dem 3D-View steht, stimmt der sog. Context
	# nicht. Daher muss er hier kuenstlich gesetzt werden.
	# fse, 28.01.15
	
	def setContextOverride(self, area):
		override = {}
		if area != None:
			region = None
			for region in area.regions:
				if region.type == 'WINDOW':
					break
			override["window"] = bpy.context.window
			override["scene" ] = bpy.context.scene
			override["screen"] = bpy.context.screen
			override["area"  ] = area
			if region != None:
				override["region"] = region
			override["blend_data"] = bpy.context.blend_data
			override["edit_object"] = bpy.context.edit_object
			override["active_object"] = bpy.context.active_object
		return override
		
	#------------------------- ------------------------- ------------------------
	
	# Behandelt die speziellen Kommandos DEV_FRONT, DEV_RIGHT, DEV_TOP und
	# DEV_FIT. 
	# \param event Ereignis-Code wie in event_type.h des Daemons definiert.
	# fse, 04.12.14
	# fse, 08.07.16: Auto-Fit wird jetzt vom Plug-in durchgefuehrt, nicht mehr
	#                vom Daemon. Siehe Kommentar bei getAndHandleData().

	def handleSpecialKeyEvent(self, event, area, override):
		log(gCf(), self, "event: {}".format(event))
		if   event == self._blenderPlugin.DEV_FIT:
			self.handleDevFit  (area, override)
		else:
			if   event == self._blenderPlugin.DEV_FRONT:
				self.handleDevFrontRightTop(area, override, 'FRONT')
			elif event == self._blenderPlugin.DEV_RIGHT:
				self.handleDevFrontRightTop(area, override, 'RIGHT')
			elif event == self._blenderPlugin.DEV_TOP:
				self.handleDevFrontRightTop(area, override, 'TOP')
			elif event == self._blenderPlugin.DEV_BACK:
				self.handleDevFrontRightTop(area, override, 'BACK')
			elif event == self._blenderPlugin.DEV_LEFT:
				self.handleDevFrontRightTop(area, override, 'LEFT')
			elif event == self._blenderPlugin.DEV_BOTTOM:
				self.handleDevFrontRightTop(area, override, 'BOTTOM')
			else:
				log(gCf(), self, "Unknown event '{}'!".format(event))
				pass
			if self._isAutoFit == True:
				self._doAutoFitDelay = 15 # Flag, um Auto-Fit zu starten

	#------------------ ------------------- ------------------- -----------------
	
	# Fuehrt die Fit-Funktion in den diversen Fenstertypen aus.
	# \param area     Area, in der der Mauszeiger gerade steht
	# \param override Context-Override-Map fuer die Operatoren
	# fse, 16.12.14
	
	def handleDevFit(self, area, override):
		if   area.type == 'VIEW_3D':
			self.handleDevFitIn3DView(override)
		elif area.type == 'TIMELINE':
			bpy.ops.time.view_all(override)
		elif area.type == 'GRAPH_EDITOR':
			bpy.ops.graph.view_all(override)
		elif area.type == 'DOPESHEET_EDITOR':
			bpy.ops.action.view_all(override)
		elif area.type == 'NLA_EDITOR':
			bpy.ops.nla.view_all(override)
		elif area.type == 'IMAGE_EDITOR':
			bpy.ops.image.view_all(override)
		elif area.type == 'SEQUENCE_EDITOR':
			bpy.ops.sequencer.view_all(override)
		elif area.type == 'CLIP_EDITOR':
			bpy.ops.clip.view_all(override)
		elif area.type == 'NODE_EDITOR':
			try:
				bpy.ops.node.view_all(override) # Exception, wenn Fenster leer, daher Exception-Handler
			except Exception:
				pass
		elif area.type == 'LOGIC_EDITOR':
			bpy.ops.view2d.reset(override)
		else:
			log(gCf(), self, "Area type '{}' not supported!".format(area.type))
			pass
			
	#-------------- --------------- --------------- --------------- -------------
	
	# Fuehrt die Fit-Funktion aus.
	# Zunaechst wird die originale Blender-Funktion view_all ausgefuehrt,
	# anschliessend werden die Positionen aller Objekte in der Szene ermittelt,
	# die kleinsten und groessten vorhandenen x-, y- und z-Werte herausgesucht
	# und auf jeder Raumachse die Mitte zwischen Minimum und Maximum berechnet;
	# dies wird der neue Drehpunkt.
	# \param override Context-Override-Map fuer die Operatoren
	# fse, 02.06.15: Bug 875 ("Beim FIT sollte der Drehpunkt auch im Bildschirm
	#                zentriert werden").
	# fse, 16.11.15: Bug 875 ("Beim FIT sollte der Drehpunkt auch im Bildschirm
	#                zentriert werden"): Fit in leerer Szene fuehrt nicht mehr
	#                zum Absturz.
	# fse, 08.07.16: Es wird jetzt ueber bpy.context.visible_objects statt ueber
	#                bpy.context.scene.objects iteriert. Grund: Sonst gehen auch
	#                Objekte ein, die evtl. auf einem anderen Layer liegen, und
	#                der Drehpunkt ist nicht da, wo man ihn erwartet.

	def handleDevFitIn3DView(self, override):
		# Fitten mit Blender-Bordmitteln:
		override['region'] = self.getActiveRegion(False)
		try:
			bpy.ops.view3d.view_all(override, center = False)
		except:
			pass
			log(gCf(), self, "Wrong region type!")
		# Drehpunkt neu setzen auf Mitte aller Objekte:
		xList = []
		yList = []
		zList = []
		if len(bpy.context.visible_objects) == 0:
			xList.append(0)
			yList.append(0)
			zList.append(0)
		else:
			for obj in bpy.context.visible_objects:
				xList.append(obj.location.x)
				yList.append(obj.location.y)
				zList.append(obj.location.z)
		self._target.x = (max(xList) - min(xList))/2
		self._target.y = (max(yList) - min(yList))/2
		self._target.z = (max(zList) - min(zList))/2
		self._dist = 0 # sonst huepft im isSlowDownMode die Szene stark bei Translation nach Fit aus grosser Distanz

	#------------------ ------------------- ------------------- -----------------
	
	# Fuehrt die Funktionen Front, Right und Top im 3D-View aus.
	# \param area     Area, in der der Mauszeiger gerade steht
	# \param override Context-Override-Map fuer die Operatoren
	# fse, 16.12.14
	
	def handleDevFrontRightTop(self, area, override, cmd):
		if   area.type == 'VIEW_3D':
			bpy.ops.view3d.view_axis(override, type = cmd)
		else:
			log(gCf(), self, "Command only available in VIEW_3D area!")
			pass
	
	#------------------------- ------------------------- ------------------------
	
	# Behandelt die Blender-Funktionen, wenn sie z. B. durch Tastendruck
	# ausgeloest werden.
	# \param event Ereignis-Code wie in event_type.h des Daemons definiert; hier
	#              immer >= APPL_FUNC_START
	# fse, 04.12.14
	# fse, 09.06.15: Bug 876 ("Blender Funktionen ausgefuehrt vom Geraet aus
	#                funktionieren nicht.")

	def handleKeyEvent(self, event, override):
		event -= self._blenderPlugin.APPL_FUNC_START
		log(gCf(), self, "Handling event '{}'...".format(event))
		cmd = self._mnuCmdList[event]
		log(gCf(), self, "Handling cmd '{}'...".format(cmd))
		# Inhalt von override:
		#{'edit_object'  : None,
		# 'scene'        : bpy.data.scenes['Scene'], 
		# 'active_object': bpy.data.objects['Cube'],
		# 'area'         : bpy.data.screens['Scripting']...Area, 
		# 'blend_data'   :<bpy_struct, BlendData at 0x000000FEDB000DF8>, 
		# 'screen'       : bpy.data.screens['Scripting'], 
		# 'window'       :<bpy_struct, Window at 0x000000FEDAF5D978>, 
		# 'region'       : bpy.data.screens['Scripting']...Region}
		# als ersten Parameter override einfuegen:
		override.pop('edit_object') # dieses Element entfernen, weil z. B. primitive_monkey_add() Blender sonst in einen "komischen" Edit-Mode schaltet
		braPos = cmd.find("(")
		cmd = cmd[:braPos+1] + "override, " + cmd[braPos+1:]
		log(gCf(), self, "Executing operator '{}' ...".format(cmd))
		# Manche Befehle starten das Plug-in ueber scLoadPreHandler() neu, was es
		# beim Aufruf hier zum Absturz bringt. Wir sammeln solche Befehle in der
		# Liste forbiddenCmdList, die ergaenzt werden kann, falls noch weitere
		# bekannt werden:
		forbidden = False
		forbiddenCmdList = ["wm.open_mainfile", "wm.read_homefile"]
		for forbiddenCmd in forbiddenCmdList:
			if cmd.startswith("bpy.ops." + forbiddenCmd):
				forbidden = True
				break
		if forbidden:
			print("Executing operator '" + cmd + "' failed: It cannot be executed by this mechanism - use a keyboard macro instead, please.")
			bpy.ops.error.message('INVOKE_DEFAULT', Failed = cmd)
		else:
			try:
				eval(cmd)
			except:
				print("Executing operator '{}' failed!".format(cmd))
				bpy.ops.error.message('INVOKE_DEFAULT', Failed = cmd)

	#-------------------------------------- -------------------------------------
	
	# Bewegt im Object-Move-Mode alle gerade selektierten Objekte.
	# \param area     die Area, in der sich der Mauszeiger befindet
	# \param region3d die Region, ueber der der Mauszeiger steht
	# \param x        x-Translation im Kamerakoordinatensystem
	# \param y        y-Translation im Kamerakoordinatensystem
	# \param z        z-Translation im Kamerakoordinatensystem
	# \param a        a-Translation im Kamerakoordinatensystem
	# \param b        b-Translation im Kamerakoordinatensystem
	# \param c        c-Translation im Kamerakoordinatensystem
	# fse, 28.01.15
	
	def transformObjects(self, area, region3d, x, y, z, a, b, c, selectedObjects):
		if region3d == None:
			return

		#pruefen, ob wir im Pose-Mode sind:
		obj = bpy.context.active_pose_bone
		isPoseMode = obj != None
	
		# gemeinsames Rotationszentrum (Mittelpunkt) aller gerade
		# selektierten Objekte ermitteln, wenn nicht alle Objekte um ihren
		# eigenen Mittelpunkt rotieren sollen:
		if isPoseMode:
			rotCenter = obj.head  # im Pose-Mode ist Rotationszentrum immer der Kopf des aktiven Knochens
		else:
			if not self._blenderPlugin.isOmmMultiRotCenter:
				rotCenter = mathutils.Vector((0,0,0))
				for obj in selectedObjects:
					rotCenter += obj.location
				rotCenter /= len(selectedObjects)

		# alle selektierten Objekte bewegen:
		if isPoseMode:
			self.transformObject(region3d, obj, x, y, z, a, b, c, rotCenter) # im Pose-Mode ist immer nur ein Knochen aktiv (= selected), ...
		else:
			for obj in selectedObjects: # ... sonst koennen viele Objekte selektiert sein
				log(gCf(), self, "obj: " + str(obj))
				objParent = obj.parent
				hasParent = objParent != None
				isParentSelected = objParent in selectedObjects
				if hasParent and isParentSelected:
					continue # Kinder nicht bewegen, weil schon durch Vater mitbewegt
				if self._blenderPlugin.isOmmMultiRotCenter:
					rotCenter = obj.location
				self.transformObject(region3d, obj, x, y, z, a, b, c, rotCenter)

		# evtl. Keyframe bei Animation erzeugen:
		curScene = bpy.context.scene
		activeKeyingSet = curScene.keying_sets.active
		if self._blenderPlugin.isGenKeyframes and curScene.tool_settings.use_keyframe_insert_auto:
			if activeKeyingSet == None:
				errTxt = "Select the keying set 'Location', 'Rotation' or 'LocRot' in the timeline's drop down list!"
				log(gCf(), self, "Failed: " + errTxt)
				bpy.ops.error.message('INVOKE_DEFAULT', Failed = "No keying set defined!")
			else:
				override = self.setContextOverride(area)
				override["selected_objects"] = selectedObjects
				bpy.ops.anim.keyframe_insert(override)
	
	#------------------------- ------------------------- ------------------------

	# Bewegt im Object-Move-Mode das Objekt objct.
	# \param objct          das zu bewegende Objekt
	# \region3d             die Region, ueber der der Mauszeiger steht
	# \param x              x-Translation im Kamerakoordinatensystem
	# \param y              y-Translation im Kamerakoordinatensystem
	# \param z              z-Translation im Kamerakoordinatensystem
	# \param a              a-Translation im Kamerakoordinatensystem
	# \param b              b-Translation im Kamerakoordinatensystem
	# \param c              c-Translation im Kamerakoordinatensystem
	# \param rotCenterInWld Rotationszentrum in Weltkoordinaten
	# fse, 28.01.15
	# fse, 02.03.18: Groesserer Umbau und Funktionen ausgelagert, um diese auch
	#                Edit-Modus verwenden zu koennen.	

	def transformObject(self, region3d, objct, x, y, z, a, b, c, rotCenterInWld):
		# Kappendaten skalieren:
		divTrans = 1000
		divRot   = 4000
		x       /= divTrans;
		y       /= divTrans;
		z       /= divTrans;
		a       /= divRot;
		b       /= divRot;
		c       /= divRot;
		
		# Transformationsmatrix erzeugen:
		log(gCf(), self, "------------------------------")
		isObjctPoseBone = bpy.context.active_pose_bone != None
		mat             = self.calcHierarchieMatrix(objct, isObjctPoseBone)
		resMat          = self.calcResultMatrix(isObjctPoseBone, mat, region3d, objct, x, y, z, a, b, c, rotCenterInWld)
		
		# Transformationsmatrix anwenden:
		if isObjctPoseBone:
			objct.matrix       = resMat
		else:
			objct.matrix_basis = resMat # .matrix_local
		log(gCf(), self, "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^")

	#------------------ ------------------- ------------------- -----------------
	
	# Durchlaeuft die Vater-Kind-Hierarchie des Objekts objct und akkumuliert
	# alle Transformationsmatrizen.
	# Es gibt verschiedene "parent_types":
	# - Unkritisch ist "'OBJECT'; unter Beruecksichtigung der Matrix zur Zeit
	#   der Adoption eines Kinds (matrix_parent_inverse) kann man die Matrizen
	#   miteinander multiplizieren.
	# - Beim Typ 'BONE' muss aber aus unbekannten Gruenden eine Korrektur
	#   angebracht werden:
	#   Die obj.matrix_parent_inverse des Objekts ist offensichtlich um pi/2 
	#   um die x-Achse gedreht und mit einem Offset von 1 auf der y-Achse
	#   versehen; das wird hier wieder durch die Matrix self._rt ausgeglichen.
	#   Leider funktioniert das bei Bones als Eltern nur, wenn der Bone
	#   seinerseits nicht verdreht ist. Denn man beobachtet, dass auch
	#   objParent.matrix_basis anscheinend verdreht ist. Wenn man den Bone z. B.
	#   um pi/2 um die y-Achse dreht, sieht die Matrix aus wie nach einer Drehung
	#   um -pi/2 um die z-Achse; nach einer Drehung um pi/2 um die z-Achse wie
	#   nach einer Drehung um pi/2 um die y-Achse. Hier ist also noch irgendein
	#   Koordinatensystem im Spiel.
	# - Der Typ 'VERTEX' wird derzeit noch nicht behandelt.
	#
	# \param objct           das zu bewegende Objekt
	# \param isObjctPoseBone True, wenn objct ein Pose-Bone ist; wir also im
	#                        Pose-Mode sind und in einer Armatur einen Bone
	#                        bewegen.
	# fse, 01.03.18
	# fse, 29.11.20: Matrix-Multiplikationsoperator "*" durch "@" ersetzt.

	def calcHierarchieMatrix(self, objct, isObjctPoseBone):
		mat = mathutils.Matrix.Identity(4)
		obj = objct
		cnt = -1
		while True:
			cnt += 1
			if isObjctPoseBone:
				objParent = bpy.context.active_object # Vater eines Pose-Bones ist immer die Armature; und die ist das aktive Objekt
			else:
				objParent  = obj.parent
			hasParent  = objParent != None
			log(gCf(), self, "cnt: " + str(cnt) + ", obj: " + str(obj) + ", objParent: " + str(objParent))
			log(gCf(), self, "isObjctPoseBone: " + str(isObjctPoseBone) + ", hasParent: " + str(hasParent))
			if hasParent:
				if isObjctPoseBone:
					objParentType = bpy.context.active_object.parent_type
				else:
					objParentType = obj.parent_type
				log(gCf(), self, "objParentType: " + str(objParentType))
				if objParentType == 'BONE':
					if isObjctPoseBone:
						mat = objParent.matrix_basis @ (self._rt@bpy.context.active_object.matrix_parent_inverse) @ mat # akkumulieren
					else:
						mat = objParent.matrix_basis @ (self._rt@obj.matrix_parent_inverse) @ mat # akkumulieren
				else:
					if objParentType == 'VERTEX':
#						idx    = obj.parent_vertices[0]
#						v = objParent.data.vertices[idx]
						pass
					else:
						# Basis-Matrix des Vorfahren und die aktuelle Matrix zum
						# Zeitpunkt der Adoption beruecksichtigen:
						if isObjctPoseBone:
							mat = objParent.matrix_basis @ bpy.context.active_object.matrix_parent_inverse @ mat # akkumulieren
						else:
							mat = objParent.matrix_basis @ obj.matrix_parent_inverse @ mat # akkumulieren
				isObjctPoseBone = False # Parent kann nicht auch Pose-Bone sein
				obj = objParent
			else:
				break
		return mat

	#------------------ ------------------- ------------------- -----------------

	# Berechnet die Transformationsmatrix, um ein Objekt gemaess den Kappendaten
	# zu bewegen. Korrekturen wegen einer eventuellen Eltern-Hierarchie wurden
	# bereits in calcHierarchieMatrix() berechnet und in mat uebergeben.
	# \param isObjctPoseBone True, wenn objct ein Pose-Bone ist; wir also im
	#                        Pose-Mode sind und in einer Armatur einen Bone
	#                        bewegen.
	# \param mat             falls objct ein Kind ist, die bisher akkumulierte
	#                        Transformationsmatrix der Eltern-Hierarchie; sonst
	#                        die Einheitsmatrix
	# \param region3d        die Region, ueber der der Mauszeiger steht
	# \param objct           das zu bewegende Objekt
	# \param x               x-Translation im Kamerakoordinatensystem
	# \param y               y-Translation im Kamerakoordinatensystem
	# \param z               z-Translation im Kamerakoordinatensystem
	# \param a               a-Translation im Kamerakoordinatensystem
	# \param b               b-Translation im Kamerakoordinatensystem
	# \param c               c-Translation im Kamerakoordinatensystem
	# \param rotCenterInWld  Rotationszentrum in Weltkoordinaten
	# fse, 02.03.18
	# fse, 29.11.20: Matrix-Multiplikationsoperator "*" durch "@" ersetzt.

	def calcResultMatrix(self, isObjctPoseBone, mat, region3d, objct, x, y, z, a, b, c, rotCenterInWld):
		viewMat           = region3d.view_matrix
		viewMat          @= mat # akkumulierte Matrix mit dem View verheiraten
		viewMatInv        = viewMat.inverted()
		trnVecInCam       = mathutils.Vector((x, y, z, 0))
		trnVecInWld       = viewMatInv @ trnVecInCam
		trnMatInWld       = mathutils.Matrix.Translation((trnVecInWld.x, trnVecInWld.y, trnVecInWld.z, 1))
		trnMatObjInWldPos = mathutils.Matrix.Translation( rotCenterInWld)
		trnMatObjInWldNeg = mathutils.Matrix.Translation(-rotCenterInWld)
		rotVecInCam       = mathutils.Vector((a, b, c, 0))
		# auf Horizont-Modus pruefen; ist nur sinnvoll, wenn man eine
		# Kamera bewegt und deren Sicht im View angezeigt wird ("Camera Persp",
		# region_3d.view_perspective = 'CAMERA'):
		if self._isHorizonLock and region3d.view_perspective == 'CAMERA':
			# Im Horizontmodus darf in Weltkoordinaten nur noch um Blenders z-Achse rotiert werden:
			rotVecInWld = viewMatInv @ rotVecInCam # diesen in Weltkoordinaten transformieren,
			rotVecInWld.x = 0                      # nur Blenders,
			rotVecInWld.y = 0                      # z-Achse erhalten
			rotVecInCam = viewMat @ rotVecInWld    # zurueck in Kamerasystem transformieren;
			rotVecInCam.x = a                      # um unsere x-Achse der Kamera darf man aber ruhig kippen, daher a unveraendert lassen
		rotVecInWld  = viewMatInv @ rotVecInCam
		rotAMatInWld = mathutils.Matrix.Rotation(rotVecInWld.x, 4, 'X')
		rotBMatInWld = mathutils.Matrix.Rotation(rotVecInWld.y, 4, 'Y')
		rotCMatInWld = mathutils.Matrix.Rotation(rotVecInWld.z, 4, 'Z')
		rotMatInWld  = rotAMatInWld @ rotBMatInWld @ rotCMatInWld
		if objct == None:
			resMat    = trnMatInWld @ trnMatObjInWldPos @ rotMatInWld @ trnMatObjInWldNeg
		else:
			if isObjctPoseBone:
				objMat = objct.matrix
			else:
				objMat = objct.matrix_basis  # .matrix_local
			resMat    = trnMatInWld @ trnMatObjInWldPos @ rotMatInWld @ trnMatObjInWldNeg @ objMat 
		return resMat

	#----------------------------------------------------------------------------
	
	# Zur Zeit unbenutzt.
	#
	# Analysiert das Objekt obj. Ein Objekt kann folgendes sein:
	# - Ein Bone, wenn sein Typ 'ARMATURE' ist und es einen Vater hat.
	# - Ein Root-Bone (=der unterste Bone einer Armatur), wenn sein Typ
	#   'ARMATURE' ist und es keinen Vater hat.
	# - Eine Armatur, wenn der Typ 'ARMATURE' ist und es beim letzten Aufruf
	#   kein Root-Bone war.
	# - Ein anderes Objekt (z. B. Mesh) sonst.
	# Problem: Armaturen und Bones haben beide den Typ 'ARMATURE', wenn man ihn
	# ueber obj.id_data.type ermittelt, sie lassen sich also nicht unterscheiden.
	# Bones haben aber kein Element obj.type. Daher der wenig performante
	# Weg ueber obj.type und Exception-Handling, wenn's den type nicht gibt.
	# 
	# Folgende Objekthierarchien koennen vorkommen:
	#
	#                                 Cube
	#                                  ^
	#                                  |
	#           Armature2 -> Bone1 -> Bone2 -> Bone3
	#               ^
	#               |
	#              Cone     Circle
	#               ^         ^
	#               |         |
	#  Torus -> Armature1 -> Bone1 -> Bone2
	#
	# Probleme:
	# - Die "object.parent"-Funktion liefert nicht immer den wahren (bzw. hier
	#   benoetigten Vater. Z. B. liefert "Cube.parent" nicht "Bone2", sondern
	#   "Armature2". 
	# - Bones haben keine "object.type"-Funktion, wir muessen aber zwischen
	#   Bones und Armatures unterscheiden.
	# Beides soll diese Funktion ausgleichen, indem sie den "wahren" Vater und
	# den "wahren" Typ zurueckliefert.
	# \param obj das Objekt, das analysiert werden soll
	# \retval resultList: resultList[0] = isBone, true wenn obj ein Bone und kein
	#                                     Root-Bone (=der erste in einer Armatur)
	#                                     ist
	#	                   resultList[1] = isRootBone, true, wenn obj Root-Bone
	#	                   resultList[2] = isArmature, true, wenn obj Armature
	#	                   resultList[3] = hasParent,  true, wenn obj Vater hat
	# fse, 05.03.15
	
	def analyseAndGetTrueParentObject(self, obj, resultList):
		isBone = isRootBone = isArmature = hasParent = isParentBone = False
		try:
			objType = obj.type
		except Exception:
			if obj.id_data.type == 'ARMATURE':
				objType = 'BONE'
			else:
				objType = "unknown type"
		isArmature = objType == 'ARMATURE' # and (isLastObjRootBone or not isLastObjBone)
		objParent = obj.parent
		hasParent = objParent != None
		if hasParent:
			if objType == 'BONE':
				isBone = True
			else:
				if obj.parent_type == 'BONE':
					objParent = self.findParentBone(obj, objParent)
					isParentBone = True
		else:
			isRootBone = objType == 'BONE'
			if isRootBone:
				objParent  = obj.id_data # Armatur als Vater angeben (obwohl hasParent = False)
				hasParent  = True
		resultList[0] = isBone
		resultList[1] = isRootBone
		resultList[2] = isArmature
		resultList[3] = hasParent
		resultList[4] = isParentBone
#		log(gCf(), self, "obj.name: {}, type: {}, isBone: {}, isRootBone: {}, isArmature: {}, hasParent: {}, isParentBone: {}".format(obj.name, type, isBone, isRootBone, isArmature, hasParent, isParentBone))
		return objParent

	#-------------------------------------- -------------------------------------
	
	# Ermittelt den Bone, an dem das Objekt obj haengt. Die richtige Armatur muss
	# bekannt sein und in armature uebergeben werden.
	# Eigentlich bestand die Hoffnung, dass man auf diese Weise in unserem
	# Object-Move-Mode auch die Objekte korrekt bewegen kann, die direkt an einem
	# Pose-Bone haengen. Denn deren object.parent-Funktion liefert nicht den
	# Bone, sondern die Armatur zurueck, was als Ursache fuer die Probleme
	# vermutet wurde. Leider ist das dennoch nicht gelungen, 
	# \param obj      das Objekt, das an dem gesuchten Bone haengt
	# \param armature die Armatur, in der der gesuchte Bone ist und an dem obj
	#                 haengt
	# \return der Bone, an dem obj haengt
	# fse, 05.03.15

	def findParentBone(self, obj, armature):
		boneName = obj.parent_bone
		parentBone = armature.pose.bones[boneName]
		return parentBone
		
	#-------------------------------------- -------------------------------------
	
	# Bewegt im Object-Move-Mode die selektierten Elemente. Das sind in
	# Blenders 'OBJECT'-Mode die ganzen Objekte (alles, was sich im "Add"-Menue
	# befindet, also Meshes, Curves, Texts, Armatures usw.) oder in
	# Blenders 'EDIT'-Mode deren Einzelteile (Vertices bei Meshes, Bones bei
	# Armatures).
	# \param area die Area, ueber der der Mauszeiger gerade steht
	# \param x    x-Translation im Kamerakoordinatensystem
	# \param y    y-Translation im Kamerakoordinatensystem
	# \param z    z-Translation im Kamerakoordinatensystem
	# \param a    a-Translation im Kamerakoordinatensystem
	# \param b    b-Translation im Kamerakoordinatensystem
	# \param c    c-Translation im Kamerakoordinatensystem
	# \return     true, wenn selektierte Objekte bewegt wurden, 
	#             false, wenn nichts selektiert war
	# fse, 16.02.18

	def transformObjectsOrSubobjects(self, area, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")
		region3d = self.getActiveRegion(True)
		obj = bpy.context.active_object
		if obj == None or area.type != 'VIEW_3D':
			isMoved = False
		else:
			modus = obj.mode;
			if modus == 'EDIT':
				isMoved = self.transformSubsInEditMode(area, region3d, x, y, z, a, b, c)
			else:
				isSelection = len(bpy.context.selected_objects)
				if isSelection:
					region3d = self.getActiveRegion(True)
					self.transformObjects(area, region3d, x, y, z, a, b, c, bpy.context.selected_objects)
					isMoved = True
				else:
					isMoved = False
		return isMoved

	#------------------------- ------------------------- ------------------------

	# Bewegt im Object-Move-Mode die selektierten Elemente, wenn Blender im
	# 'EDIT'-Mode ist. Wenn das aktive Objekt ein Mesh ist, werden die
	# selektierten Vertices bewegt, bei einer Armature die Bones usw.
	# \param region3d die Region, ueber der der Mauszeiger steht
	# \param x        x-Translation im Kamerakoordinatensystem
	# \param y        y-Translation im Kamerakoordinatensystem
	# \param z        z-Translation im Kamerakoordinatensystem
	# \param a        a-Translation im Kamerakoordinatensystem
	# \param b        b-Translation im Kamerakoordinatensystem
	# \param c        c-Translation im Kamerakoordinatensystem
	# \return         true, wenn selektierte Objekte bewegt wurden, 
	#                 false, wenn nichts selektiert war
	# fse, 16.02.18

	def transformSubsInEditMode(self, area, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked. region3d: " + str(region3d))
		if  area.type == 'VIEW_3D':
			typ = bpy.context.active_object.type
			if typ   == 'MESH':
				isMoved = self.transformMeshSubs      (region3d, x, y, z, a, b, c)
			elif typ == 'ARMATURE':
				isMoved = self.transformArmatureSubs  (region3d, x, y, z, a, b, c)
			elif typ == 'CURVE' or typ == 'SURFACE':
				isMoved = self.transformCurveTypes    (region3d, x, y, z, a, b, c)
			elif typ == 'LATTICE':
				isMoved = self.transformLatticeSubs   (region3d, x, y, z, a, b, c)
			elif typ == 'META':
				isMoved = self.transformMetaSubs      (region3d, x, y, z, a, b, c)
			else:
				isMoved = False
		else:
			# in 2D-Views funktioniert leider noch nichts; es muesste ungefaehr so
			# laufen:
			if area.type == 'IMAGE_EDITOR':
				isMoved = self.transformVertInUvEditor(region3d, x, y, z, a, b, c)
			isMoved = False
		return isMoved
		
	#------------------ ------------------- ------------------- -----------------

	# Bewegt im Edit-Modus die selektierten Vertices entsprechend der
	# Kappenauslenkung.
	# \param region3d      die Region, ueber der der Mauszeiger steht
	# \param x             x-Translation im Kamerakoordinatensystem
	# \param y             y-Translation im Kamerakoordinatensystem
	# \param z             z-Translation im Kamerakoordinatensystem
	# \param a             a-Translation im Kamerakoordinatensystem
	# \param b             b-Translation im Kamerakoordinatensystem
	# \param c             c-Translation im Kamerakoordinatensystem
	# fse, 16.02.18

	def transformMeshSubs(self, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")
		obj = bpy.context.active_object
		
		# selektierte Vertices holen; das funktioniert leider nur im
		# 'OBJECT'-Mode, daher muss auf diesen umgeschaltet werden:
		modus = bpy.context.active_object.mode; # aktuellen Modus merken (kann nur 'EDIT' sein)
		bpy.ops.object.mode_set(mode='OBJECT')
		vectors = []
		for v in obj.data.vertices:
			if v.select:
				vectors.append(v.co) # Koordinatenvektor eine Vertex speichern
				
		# alle Vertices bewegen:
		if vectors:
			self.transformVectors(region3d, obj, x, y, z, a, b, c, vectors)
			isMoved = True
		else:
			isMoved = False
			
		# auf vorherigen Modus zurueckschalten:
		bpy.ops.object.mode_set(mode=modus)
		return isMoved

	#-------------- --------------- --------------- --------------- -------------

	# Transformiert die Vektoren in der Liste vectors entsprechend der
	# Kappenauslenkung. Die Vektoren sind dabei relativ zum lokalen
	# Koordinatensystem des Objekts obj definiert.
	# \param region3d die Region, ueber der der Mauszeiger steht
	# \param obj      das Objekt, in dessen Koordinatensystem die Vektoren
	#                 definiert sind
	# \param x        x-Translation im Kamerakoordinatensystem
	# \param y        y-Translation im Kamerakoordinatensystem
	# \param z        z-Translation im Kamerakoordinatensystem
	# \param a        a-Translation im Kamerakoordinatensystem
	# \param b        b-Translation im Kamerakoordinatensystem
	# \param c        c-Translation im Kamerakoordinatensystem
	# \param vectors  Liste mit zu transformierenden Vektoren 
	# fse, 19.02.18
	# fse, 29.11.20: Matrix-Multiplikationsoperator "*" durch "@" ersetzt.

	def transformVectors(self, region3d, obj, x, y, z, a, b, c, vectors):
		if region3d == None:
			return
		
		divTrans = 1000
		divRot   = 4000
		x /= divTrans;
		y /= divTrans;
		z /= divTrans;
		a /= divRot;
		b /= divRot;
		c /= divRot;

		# Rotationszentrum berechnen:
		rotCenter = mathutils.Vector((0,0,0))
		for v in vectors:
			log(gCf(), self, "v: " + str(v))
			rotCenter +=  v
			log(gCf(), self, "rotCenter: " + str(rotCenter))
		rotCenter /= len(vectors) # Mittel aller Vektoren
		log(gCf(), self, "=> rotCenter: " + str(rotCenter))
		
		# Transformationsmatrizen ermitteln:
		objMat         = obj.matrix_basis  # setzt obj in die Welt
		objMatInv      = objMat.inverted() # setzt Weltpunkte ins System von obj
		rotCenterInWld = objMat @ rotCenter
		mat            = self.calcHierarchieMatrix(obj, False)
		trnMat         = self.calcResultMatrix(False, mat, region3d, None, x, y, z, a, b, c, rotCenterInWld)
#		trnMat         = self.calcTrnMat2(region3d, None, x, y, z, a, b, c, rotCenterInWld)

		# Kopie der Vektorliste anlegen, weil sonst bereits transformierte
		# Elemente die spaeter zu transformierenden beeinflussen:
		vectorsCopy = copy.deepcopy(vectors)

		# alle Punkte in der Kopie transformieren:
		for v in vectorsCopy:
			vertInWld    = objMat @ v
			newVertInWld = trnMat @ vertInWld
			newVertInObj = objMatInv @ newVertInWld
			v.x          = newVertInObj.x
			v.y          = newVertInObj.y
			v.z          = newVertInObj.z

		# Im zweiten Schritt alle echten Vektoren mit den neuen
		# Koordinaten versehen:
		for i in range(0, len(vectors)):
			vectors[i].x = vectorsCopy[i].x
			vectors[i].y = vectorsCopy[i].y
			vectors[i].z = vectorsCopy[i].z

	#----------- ------------- ------------ ------------ ------------- ----------

	# Obsolet!
	
	# Berechnet die Transformationsmatrix, um einen Punkt in der Welt so zu
	# bewegen, wie es die im Kamerasystem vorliegenden Kappendaten verlangen.
	# Macht im Prinzip das gleiche wie transformObject(), daher unschoene
	# Code-Duplizierung.
	# \param region3d       die Region, ueber der der Mauszeiger steht
	# \param x              x-Translation im Kamerakoordinatensystem
	# \param y              y-Translation im Kamerakoordinatensystem
	# \param z              z-Translation im Kamerakoordinatensystem
	# \param a              a-Translation im Kamerakoordinatensystem
	# \param b              b-Translation im Kamerakoordinatensystem
	# \param c              c-Translation im Kamerakoordinatensystem
	# \param rotCenterInWld zu verwendendes Rotationszentrum in Weltkoordinaten
	# \return die Transformationsmatrix
	# fse, 13.02.18
	# fse, 29.11.20: Matrix-Multiplikationsoperator "*" durch "@" ersetzt.

	def calcTrnMat2(self, region3d, objct, x, y, z, a, b, c, rotCenterInWld):
		divTrans = 1000
		divRot   = 4000
		x /= divTrans;
		y /= divTrans;
		z /= divTrans;
		a /= divRot;
		b /= divRot;
		c /= divRot;
		
		viewMat           = region3d.view_matrix
		mat               = mathutils.Matrix.Identity(4)
		obj = objct
		cnt          = -1
		while True:
			cnt += 1
			objType    = obj.type
			isArmature = objType == 'ARMATURE' # and (isLastObjRootBone or not isLastObjBone)
			objParent  = obj.parent
			hasParent  = objParent != None
			log(gCf(), self, "cnt: " + str(cnt) + ", obj: " + str(obj) + ", objParent: " + str(objParent))
			log(gCf(), self, "isArmature: " + str(isArmature) + ", hasParent: " + str(hasParent))
			if hasParent:
				isParentArmature = objParent.type == 'ARMATURE'
				if isParentArmature and obj.parent_type == 'BONE':
					mat = obj.parent.matrix_basis @ (self._rt@obj.matrix_parent_inverse) @ mat # akkumulieren
				else:
					log(gCf(), self, "cnt: " + str(cnt) + ": !isParentBone.")
					# das hier sind die Elemente normaler Objekt-Hierarchien:
					# Basis-Matrix des Vorfahrs und seine aktuelle Matrix zum
					# Zeitpunkt der Adoption beruecksichtigen:
					mat = objParent.matrix_basis @ obj.matrix_parent_inverse @ mat # akkumulieren
					#print("used parent: {}, mat = objParent.matrix_basis @ obj.matrix_parent_inverse @ mat:".format(objParent.name))
					#print(objParent.matrix_basis)
					#print(obj.matrix_parent_inverse)
				obj = objParent
			else:
				log(gCf(), self, "cnt: " + str(cnt) + ": !hasParent. ------------")
				break
		viewMat          @= mat # akkumulierte Matrix mit dem View verheiraten
		viewMatInv        = viewMat.inverted()
		trnVecInCam       = mathutils.Vector((x, y, z, 0))
		trnVecInWld       = viewMatInv @ trnVecInCam
		trnMatInWld       = mathutils.Matrix.Translation((trnVecInWld.x, trnVecInWld.y, trnVecInWld.z, 1))
		trnMatObjInWldPos = mathutils.Matrix.Translation( rotCenterInWld)
		trnMatObjInWldNeg = mathutils.Matrix.Translation(-rotCenterInWld)
		rotVecInCam       = mathutils.Vector((a, b, c, 0))
		rotVecInWld       = viewMatInv @ rotVecInCam
		rotAMatInWld      = mathutils.Matrix.Rotation(rotVecInWld.x, 4, 'X')
		rotBMatInWld      = mathutils.Matrix.Rotation(rotVecInWld.y, 4, 'Y')
		rotCMatInWld      = mathutils.Matrix.Rotation(rotVecInWld.z, 4, 'Z')
		rotMatInWld       = rotAMatInWld @ rotBMatInWld @ rotCMatInWld
		resMat            = trnMatInWld @ trnMatObjInWldPos @ rotMatInWld @ trnMatObjInWldNeg
		return resMat

	#------------------ ------------------- ------------------- -----------------

	# Bewegt im Edit-Modus die selektierten Bones einer Armature entsprechend der
	# Kappenauslenkung.
	# \param region3d      die Region, ueber der der Mauszeiger steht
	# \param x             x-Translation im Kamerakoordinatensystem
	# \param y             y-Translation im Kamerakoordinatensystem
	# \param z             z-Translation im Kamerakoordinatensystem
	# \param a             a-Translation im Kamerakoordinatensystem
	# \param b             b-Translation im Kamerakoordinatensystem
	# \param c             c-Translation im Kamerakoordinatensystem
	# fse, 16.02.18

	def transformArmatureSubs(self, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")
		obj = bpy.context.active_object
		
		# Erst einmal die Indizes der selektierten Heads und Tails in der Armature
		# holen; das funktioniert leider nur im 'OBJECT'-Mode, daher muss auf
		# diesen umgeschaltet werden:
		modus = bpy.context.active_object.mode;
		bpy.ops.object.mode_set(mode='OBJECT')
		selHeadIdxList = [] # die Indizes der selektierten Heads der Bones
		selTailIdxList = [] # die Indizes der selektierten Tails der Bones
		bnes = obj.data.bones
		for i in range(0, len(bnes)):
			if bnes[i].select_head:
				selHeadIdxList.append(i)
			if bnes[i].select_tail:
				selTailIdxList.append(i)
		bpy.ops.object.mode_set(mode=modus) # auf vorherigen Modus zurueckschalten:
				
		# Falls selektierte Heads oder Tails gefunden wurden, diese in einer
		# weiteren Liste zusammenfuehren und transformieren:
		if (selHeadIdxList or selTailIdxList):
			selHeadsAndTails = [] # die selektierten Heads und Tails der Bones
#		bpy.ops.object.mode_set(mode='EDIT')
			for i in selHeadIdxList:
				selHeadsAndTails.append(obj.data.edit_bones[i].head)
			for i in selTailIdxList:
				selHeadsAndTails.append(obj.data.edit_bones[i].tail)
			self.transformVectors(region3d, obj, x, y, z, a, b, c, selHeadsAndTails)
			isMoved = True
		else:
			isMoved = False
		return isMoved

	#------------------ ------------------- ------------------- -----------------

	# Bewegt im Edit-Modus die selektierten Kontrollpunkte von Beziers und Nurbs
	# entsprechend der Kappenauslenkung.
	# \param region3d      die Region, ueber der der Mauszeiger steht
	# \param x             x-Translation im Kamerakoordinatensystem
	# \param y             y-Translation im Kamerakoordinatensystem
	# \param z             z-Translation im Kamerakoordinatensystem
	# \param a             a-Translation im Kamerakoordinatensystem
	# \param b             b-Translation im Kamerakoordinatensystem
	# \param c             c-Translation im Kamerakoordinatensystem
	# fse, 10.02.18

	def transformCurveTypes(self, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")
		typ = bpy.context.active_object.data.splines[0].type
		if typ == 'BEZIER':
			isMoved = self.transformBezierSubs(region3d, x, y, z, a, b, c)
		elif typ == 'NURBS':
			isMoved = self.transformNurbsSubs(region3d, x, y, z, a, b, c)
		else:
			isMoved = False
		return isMoved
		
	#-------------- --------------- --------------- --------------- -------------

	# Bewegt im Edit-Modus die selektierten Kontrollpunkte von Beziers 
	# entsprechend der Kappenauslenkung.
	# \param region3d      die Region, ueber der der Mauszeiger steht
	# \param x             x-Translation im Kamerakoordinatensystem
	# \param y             y-Translation im Kamerakoordinatensystem
	# \param z             z-Translation im Kamerakoordinatensystem
	# \param a             a-Translation im Kamerakoordinatensystem
	# \param b             b-Translation im Kamerakoordinatensystem
	# \param c             c-Translation im Kamerakoordinatensystem
	# fse, 20.02.18

	def transformBezierSubs(self, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")	
		obj = bpy.context.active_object
		
		# selektierte Control-Points und Handles holen:
		selectedCps   = [cp for cp in obj.data.splines[0].bezier_points if cp.select_control_point]
		selectedCpsLh = [lh for lh in obj.data.splines[0].bezier_points if lh.select_left_handle]
		selectedCpsRh = [rh for rh in obj.data.splines[0].bezier_points if rh.select_right_handle]
		selectedObjects = []
		
		for cp in selectedCps:
			selectedObjects.append(cp.co)
		for lh in selectedCpsLh:
			selectedObjects.append(lh.handle_left)
		for rh in selectedCpsRh:
			selectedObjects.append(rh.handle_right)
		# alle Vertices bewegen:
		if selectedObjects:
			self.transformVectors(region3d, obj, x, y, z, a, b, c, selectedObjects)
			isMoved = True
		else:
			isMoved = False
		return isMoved
	
	#-------------- --------------- --------------- --------------- -------------

	# Bewegt im Edit-Modus die selektierten Kontrollpunkte von Nurbs
	# entsprechend der Kappenauslenkung.
	# \param region3d      die Region, ueber der der Mauszeiger steht
	# \param x             x-Translation im Kamerakoordinatensystem
	# \param y             y-Translation im Kamerakoordinatensystem
	# \param z             z-Translation im Kamerakoordinatensystem
	# \param a             a-Translation im Kamerakoordinatensystem
	# \param b             b-Translation im Kamerakoordinatensystem
	# \param c             c-Translation im Kamerakoordinatensystem
	# fse, 20.02.18

	def transformNurbsSubs(self, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")	
		obj = bpy.context.active_object
		
		# selektierte Punkte:
		vectors3d = []
		vectors4d = []
		for p in obj.data.splines[0].points:
			if p.select:
				vectors4d.append(p.co)
#				v = p.co.resized(3) # funktioniert nicht, daher
				v = mathutils.Vector((p.co.x,p.co.y,p.co.z))
				vectors3d.append(v)
				log(gCf(), self, "vectors3d[0]: " + str(vectors3d[0]))

		# alle selektierten Punkte bewegen:
		if vectors4d:
			self.transformVectors(region3d, obj, x, y, z, a, b, c, vectors3d)
			for i in range(0, len(vectors4d)):
				vectors4d[i].x = vectors3d[i].x
				vectors4d[i].y = vectors3d[i].y
				vectors4d[i].z = vectors3d[i].z
			isMoved = True
		else:
			isMoved = False
		return isMoved

	#-------------- --------------- --------------- --------------- -------------

	# Bewegt im Edit-Modus die selektierten Punkte von Lattices
	# entsprechend der Kappenauslenkung.
	# \param region3d      die Region, ueber der der Mauszeiger steht
	# \param x             x-Translation im Kamerakoordinatensystem
	# \param y             y-Translation im Kamerakoordinatensystem
	# \param z             z-Translation im Kamerakoordinatensystem
	# \param a             a-Translation im Kamerakoordinatensystem
	# \param b             b-Translation im Kamerakoordinatensystem
	# \param c             c-Translation im Kamerakoordinatensystem
	# fse, 20.02.18

	def transformLatticeSubs(self, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")
		obj = bpy.context.active_object
		
		# selektierte Points holen:
		vectors = []
		vectorsCopy = []
		for p in obj.data.points:
			if p.select:
				vectors.append(p) # Koordinatenvektor eines Points speichern
				v = mathutils.Vector((p.co_deform.x,p.co_deform.y,p.co_deform.z))
				log(gCf(), self, "v: " + str(v))
				vectorsCopy.append(v)
		
		# alle Points bewegen:
		if vectors:
			self.transformVectors(region3d, obj, x, y, z, a, b, c, vectorsCopy)
			for i in range(0, len(vectorsCopy)):
				vectors[i].co_deform.x = vectorsCopy[i].x
				vectors[i].co_deform.y = vectorsCopy[i].y
				vectors[i].co_deform.z = vectorsCopy[i].z
			isMoved = True
		else:
			isMoved = False
			
		return isMoved

	#-------------- --------------- --------------- --------------- -------------

	# Bewegt im Edit-Modus den ersten selektierten Metaball
	# entsprechend der Kappenauslenkung.
	# \param region3d      die Region, ueber der der Mauszeiger steht
	# \param x             x-Translation im Kamerakoordinatensystem
	# \param y             y-Translation im Kamerakoordinatensystem
	# \param z             z-Translation im Kamerakoordinatensystem
	# \param a             a-Translation im Kamerakoordinatensystem
	# \param b             b-Translation im Kamerakoordinatensystem
	# \param c             c-Translation im Kamerakoordinatensystem
	# fse, 20.02.18

	def transformMetaSubs(self, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")
		obj = bpy.context.active_object
		
		# selektierte Metaballs holen: leider gibt es keine "select"-Property;
		# daher nehmen wir notgedrungen die Property "active", die immer nur
		# den ersten selektierten Metaball liefert, oder den zuletzt selektierten,
		# wenn gerade gar keiner selektiert ist:
		vectors = [ obj.data.elements.active.co ]
		
		# alle Points bewegen:
		if vectors:
			self.transformVectors(region3d, obj, x, y, z, a, b, c, vectors)
			isMoved = True
		else:
			isMoved = False
			
		return isMoved
	
	#------------------ ------------------- ------------------- -----------------

	# Funktioniert noch nicht, da unbekannt, wie man im UV-Editor die
	# selektierten Vertices ermittelt. Der auskommentierte Code stammt
	# aus dem Operator-Mesh-UV-Template, ermittelt aber die selektierten
	# Vertices des zu behandelnden Objekts.
	
	# Bewegt UV-Editor die selektierten Vertices entsprechend der
	# Kappenauslenkung.
	# \param region3d      die Region, ueber der der Mauszeiger steht
	# \param x             x-Translation im Kamerakoordinatensystem
	# \param y             y-Translation im Kamerakoordinatensystem
	# \param z             z-Translation im Kamerakoordinatensystem
	# \param a             a-Translation im Kamerakoordinatensystem
	# \param b             b-Translation im Kamerakoordinatensystem
	# \param c             c-Translation im Kamerakoordinatensystem
	# fse, 16.02.18

	def transformVertInUvEditor(self, region3d, x, y, z, a, b, c):
		log(gCf(), self, "Invoked.")
#		obj = bpy.context.active_object
#		me = obj.data
#		bm = bmesh.from_edit_mesh(me)
#
#		uv_layer = bm.loops.layers.uv.verify()
#		bm.faces.layers.tex.verify()  # currently blender needs both layers.
#
#		# adjust UVs
#		log(gCf(), self, "start f loop: ")
#		for f in bm.faces:
#			log(gCf(), self, "start l loop: ")
#			for l in f.loops:
#				luv = l[uv_layer]
#				if luv.select:
#					# apply the location of the vertex as a UV
#					log(gCf(), self, "luv.uv: " + str(luv.uv))
##					luv.uv = l.vert.co.xy
#			log(gCf(), self, "end l loop: ")
#		log(gCf(), self, "end f loop: ")
	#	bmesh.update_edit_mesh(me)
		pass
		
	#-------------------------------------- -------------------------------------
	
	# Bewegt die Szene in einem 2D- oder 3D-Fenster in Blender.
	# \param area  die Area, in der sich der Mauszeiger befindet
	# \param x     x-Translation im Kamerakoordinatensystem
	# \param y     y-Translation im Kamerakoordinatensystem
	# \param z     z-Translation im Kamerakoordinatensystem
	# \param a     a-Translation im Kamerakoordinatensystem
	# \param b     b-Translation im Kamerakoordinatensystem
	# \param c     c-Translation im Kamerakoordinatensystem
	# fse, 04.12.14

	def transformView(self, area, x, y, z, a, b, c):
		if area == None:
			return
		if area.type == 'VIEW_3D':
			self.transformView3d(area, x, y, z, a, b, c)
		else:
			self.transformView2d(area, x, y, z)
			
	#------------------------- ------------------------- ------------------------

	# Bewegt die Szene in einem 3D-Fenster in Blender.
	# In Wahrheit wird natuerlich die Kamera bewegt.
	# bpy.data.window_managers[0].windows[0].screen.areas[2].spaces[0].region_3d:
	# window_managers[0]: Window-Manager gibt es anscheinend immer nur
	#                     einen mit dem Index 0.
	# windows: Gibt es mehrere, das Hauptfenster hat den Index 0. Wenn man ein
	#          weiteres oeffnet (z. B. das User-Preferences-Fenster) hat dieses
	#          den Index 1.
	# screen: Gibt das aktuelle Layout des Hauptfensters an, wie in der
	#         Drop-Down-Liste des Info-Headers angegeben (z. B. "Scripting" oder
	#         "Default"). Daran aendert sich auch nichts, wenn man bei offenem
	#         User-Preferences-Fenster das windows[1] abfragt.
	# areas: Ein Screen besteht aus mehreren Areas. Der Scripting-Screen besteht
	#        z. B. aus den sechs des Typs (areas[x].type) 'INFO', 'PROPERTIES'
	#        'VIEW_3D', 'CONSOLE', 'OUTLINER' und 'TEXT_EDITOR', sofern man keine
	#        weiteren erzeugt hat (z. B. durch Teilung des 3D-Views in mehrere).
	# Area.regions: Eine Aerea kann mehrere Regionen haben. Der 3D-View besteht
	#               z. B. aus den Regionen 'HEADER', 'TOOLS', 'TOOL_PROPS', 'UI'
	#               und 'WINDOW'. Wenn man in den Quad-View wechselt, kommen drei
	#               weitere 'WINDOW'-Elemente hinzu.
	# spaces: Scheint es immer nur einen zu geben, der direkt vom Typ der Area
	#         abhaengt. Eine 'VIEW_3D'-Aerea hat einen "SpaceView3D"-Space.
	# SpaceView3D.view_matrix: Unser SpaceView3D hat dann schliesslich ein
	#                          Element "view_matrix". Diese Matrix wandelt die
	#                          Koordinaten eines Punkts im Weltkoordinatensystem
	#                          in die des Kamerasystems um.
	# SpaceView3D.region_quadviews: In der Quad-View-Darstellung enthaelt diese
	#                               Liste vier Elemente, wobei jedes wieder eine
	#                               view_matrix hat.
	# \param area  die Area, in der sich der Mauszeiger befindet
	# \param x     x-Translation im Kamerakoordinatensystem
	# \param y     y-Translation im Kamerakoordinatensystem
	# \param z     z-Translation im Kamerakoordinatensystem
	# \param a     a-Translation im Kamerakoordinatensystem
	# \param b     b-Translation im Kamerakoordinatensystem
	# \param c     c-Translation im Kamerakoordinatensystem
	# fse, 04.12.14
	# fse, 29.11.20: Matrix-Multiplikationsoperator "*" durch "@" ersetzt.

	def transformView3d(self, area, x, y, z, a, b, c):
		self._region3d = self.getActiveRegion(True)
		#log(gCf(), self, "self._region3d: " + str(self._region3d))
		if self._region3d == None:
			return
		if self._blenderPlugin.isSlowDownMode and self._region3d.is_perspective:
			fact = 1/max(self._dist, 0.5)
		else:
			fact = 0.3
		divTrans = self._blenderPlugin.transSensBase * fact
		divRot   = max(self._blenderPlugin.rotSensBase * fact, 4000)
		x /= divTrans;
		y /= divTrans;
		z /= divTrans;
		a /= divRot;
		b /= divRot;
		c /= divRot;
		self._viewMat    = self._region3d.view_matrix
		self._viewMatInv = self._viewMat.inverted()
		camPos           = self._viewMatInv @ self._origin # Position der Kamera im Weltkoordinatensystem
		#print("camPos: {}".format(camPos))
		moveMode         = self._moveMode
		self.getTargetPoint(camPos)
		if moveMode == self._blenderPlugin.OBJ_SCREEN:
			targetPntInWld = self._target
			targetPntInWld.w = 1
			self._region3d.view_location = targetPntInWld.to_3d()
			self.moveCam(area, -x, -y, -z, -a, -b, -c, targetPntInWld, self._isHoverMode, self._isHorizonLock)
		elif moveMode == self._blenderPlugin.OBJ_ORIGIN:
			targetPntInWld = self._origin
			self._region3d.view_location = targetPntInWld.to_3d()
			self.moveCam(area, -x, -y, -z, -a, -b, -c, targetPntInWld, self._isHoverMode, self._isHorizonLock)
		elif moveMode == self._blenderPlugin.CAM_STANDARD:
			targetPntInWld = camPos
			self._region3d.view_location = targetPntInWld.to_3d()
			self.moveCam(area,  x,  y,  z,  a,  b,  c, targetPntInWld, self._isHoverMode, self._isHorizonLock)
		else:
			targetPntInWld = self._target
			targetPntInWld.w = 1
			self._region3d.view_location = targetPntInWld.to_3d()
			self.moveCam(area,  x,  y,  z,  a,  b,  c, targetPntInWld, self._isHoverMode, self._isHorizonLock)		

	#------------------ ------------------- ------------------- -----------------
	
	# Ermittelt die aktive Region, also die Region unter dem Mauszeiger.
	# Der Screen (Layout) besteht aus mehreren Areas, keine oder mehrere davon
	# koennen vom Typ 'VIEW_3D' sein. Eine Area hat einen Space vom selben Typ,
	# von dem sich diverse Klassen ableiten, die die verschiedenen Editoren
	# implementieren; wozu ein Space sonst noch gut ist, ist unklar.
	# spaces[0] und spaces.active scheint dasselbe zu sein. spaces[0] hat
	# jedenfalls ein Element region_3d, das wiederum ein Element view_matrix
	# besitzt. Diese Matrix rechnet einen Punkt aus dem Weltkoordinatensystem in
	# einen Punkt des Kamerakoordinatensystems um. Falls der View in der
	# Darstellung mit vier Quadranten ist, greift man auf die Quadranten mit
	# region_quadviews[x] zu; dies sind ebenfalls Elemente region_3d.
	# spaces[0].region_3d und spaces[0].region_quadviews[3] ist wohl dasselbe
	# Objekt.
	#
	# Geometrie:
	#
	# SpaceView3D besteht entweder aus einem Fenster oder den vier 
	# Quadranten und ist als eine Area in einen Screen eingebettet:
	#
	#  +-------------------------+             +-------------------------+
	#  |                         |             |            |            |
	#  |                         |             |   quad 1   |   quad 3   |
	#  |                         |             |            |            |
	#  |                         |             |------------+------------|
	#  |                         |             |            |            |
	#  |                         |             |   quad 0   |   quad 2   |
	#  |                         |             |            |            |
	#  |-------------------------| Header      |-------------------------| Header
	#  +-------------------------+             +-------------------------+
	#
	# Ziel ist es, die region_3d zu ermitteln, ueber der der Mauszeiger gerade
	# steht. Das ist anscheinend nicht einfach moeglich, daher etwas
	# komplizierter:
	# Die aktuelle Mauszeigerposition kann man ueber ein Maus-Event ermitteln,
	# das uns unser Timer-Operator ModalBlenderTimerOp liefert. Die Koordinaten
	# sind relativ zum Screen und werden in self._mouseX und self._mouseY
	# abgelegt. Die Position und Groesse einer Area kann man abfragen und dadurch
	# berechnen, ob sich der Mauszeiger in der Area befindet. Die Groesse und
	# Position eines Quadranten ist nur deswegen bekannt, weil alle vier
	# Quadranten grundsaetzlich gleich gross sind und je ein Viertel der Area
	# (abzueglich der 'Header'-Fusszeile) beanspruchen.
	# \param isRegionView3D wenn true, wird eine Region des Typs RegionView3D
	#                       zurueckgegeben, sonst eine vom Typ Region
	# \return die Region unter dem Mauszeiger
	# fse, 10.12.14
	
	def getActiveRegion(self, isRegionView3dNeeded):
		ret = self._lastRegion
		for area in bpy.context.screen.areas:
			if area.type == 'VIEW_3D':
				mx = self._mouseX
				my = self._mouseY
				ax = area.x
				ay = area.y
				aw = area.width
				ah = area.height
				if mx >= ax and mx < ax + aw and my >= ay and my < ay + ah:
					isQuadView = len(area.spaces.active.region_quadviews) > 0
					if isQuadView:
						if isRegionView3dNeeded:
							ret = self.getQuadviewActiveRegionView3D(area, mx, my, ax, ay, aw, ah)
							if ret != None:
								self._lastRegion = ret
								break
						else:
							ret = self.getQuadviewActiveRegion(area, mx, my)
							if ret != None:
								break
					else:
						if isRegionView3dNeeded:
							ret = self._lastRegion = area.spaces.active.region_3d
						else:
							for region in [r for r in area.regions if r.type == 'WINDOW']:
								ret = region # brauchen wir nur fuer Fit, daher _lastRegion nicht veraendern
		log(gCf(), self, "activeRegion: " + str(ret))
		return ret
				
	#-------------- --------------- --------------- --------------- -------------
	
	# Eine Quadview-Area ist in vier gleichgrosse Quadranten unterteilt, so dass
	# sich die Groessen der Quadranten einfach durch Vierteilung ermitteln
	# lassen. Unten muss man allerdings noch den Header abziehen.
	# \param area die Area, die gerade untersucht wird
	# \param mx   x-Position des Mauszeigers
	# \param my   y-Position des Mauszeigers
	# \param ax   x-Position der Area
	# \param ay   y-Position der Area
	# \param aw   Breite der Area
	# \param ah   Hoehe der Area
	
	def getQuadviewActiveRegionView3D(self, area, mx, my, ax, ay, aw, ah):
		ret = None
		header = next(r for r in area.regions if r.type == 'HEADER')
		hh  = header.height
		qw  = aw/2                  #  +-------------------------+
		qh  = (ah-hh)/2             #  |            |            |
		q0x = ax                    #  |   quad 1   |   quad 3   |
		q0y = ay + hh               #  |            |            |
		q1x = ax                    #  |------------+------------|
		q1y = ay + qh + hh          #  |            |            |
		q2x = ax + qw               #  |   quad 0   |   quad 2   |
		q2y = ay + hh               #  |            |            |
#		q3x = ax + qw               #  |-------------------------| Header
#		q3y = ay + qh + hh          #  +-------------------------+
		if   mx >= q0x and mx < q0x + qw and my >= q0y and my < q0y + qh:
			ret = area.spaces.active.region_quadviews[0]
		elif mx >= q1x and mx < q1x + qw and my >= q1y and my < q1y + qh:
			ret = area.spaces.active.region_quadviews[1]
		elif mx >= q2x and mx < q2x + qw and my >= q2y and my < q2y + qh:
			ret = area.spaces.active.region_quadviews[2]
		else:
			ret = area.spaces.active.region_quadviews[3]
		return ret
	
	#-------------- --------------- --------------- --------------- -------------
	
	# Im Quadview-Modus bildet sich nicht nur eine neue Liste
	# active.region_quadviews[x] mit Regionen vom Typ "RegionView3D" aus, sondern
	# auch die vorhandene Liste area.regions erweitert sich um 3 Regionen vom Typ
	# "Region". Dieser Typ wird gebraucht, wenn man mit einem Operator ein Fit
	# machen will.
	# \param area die Area, die gerade untersucht wird
	# \param mx   x-Position des Mauszeigers
	# \param my   y-Position des Mauszeigers
	# fse, 16.12.14
	
	def getQuadviewActiveRegion(self, area, mx, my):
		ret = None
		for region in area.regions:
			x = region.x
			y = region.y
			w = region.width
			h = region.height
			if mx >= x and mx < x + w and my >= y and my < y + h:
				ret = region
				break
		return ret
	
	#------------------ ------------------- ------------------- -----------------

	# Ermittelt den neuen Target-Punkt (Punkt, auf den die Kamera blicken soll),
	# der sich nach Anwendung der Kappenauslenkung ergibt.
	# Die Richtung, in die die Kamera schaut, kann man leicht durch den
	# umgedrehten Normalen-Vektor der Kamera (0,0,1) (transformiert in
	# Weltkoordinaten) ermitteln. Das Problem ist aber der Abstand zum Zielpunkt,
	# der sich ja irgendwo auf der Achse in den Bildschirm hinein befinden kann.
	# Ausserdem soll sich dieser Abstand bei Rotationen automatisch so einpendeln,
	# dass in den Rotationsmodi um die Bildschirmmitte schliesslich die Kamera
	# einen solchen Abstand einnimmt, dass auch ohne gleichzeitige Translation
	# das Objekt in der Mitte bleibt.
	# Das wird hier durch Verwendung des "vorherigen" Target-Punktes erreicht:
	# - Die Kamera hat sich gegenueber dem Zeitpunkt des vorherigen 
	#   Target-Punktes bewegt und wurde rotiert.
	# - Der vorherige Target-Punkt hat einen Abstand zur aktuellen Position der
	#   Kamera.
	# - Aus dem aktuellen Normalen-Vektor wird die Blickrichtung der Kamera
	#   (= die Bildschirmmitte) ermittelt.
	# - Als neuer Target-Punkt wird die Stelle auf dem Sehstrahl verwendet, die
	#   die Entfernung zur Kamera hat wie die Kamera zum vorherigen Target-Punkt.
	# fse, 02.12.14
	# fse, 29.11.20: Matrix-Multiplikationsoperator "*" durch "@" ersetzt.

	def getTargetPoint(self, camPos):
		camPos.w = 0
		#print("_target1: {}".format(self._target))
		camMinusTarget = camPos - self._target
		camMinusTarget.w = 0
#		print("  self._target: {}".format(self._target))
#		print("        camPos: {}".format(camPos))
#		print("camMinusTarget: {}".format(camMinusTarget))
		self._dist = camMinusTarget.length   # Abstand Kamera von Target-Punkt
		#print("dist: {}".format(self._dist))
		normal = self._viewMatInv @ self._uz # _uz: Einheitsvektor in z-Richtung (Kamerasystem, also Vektor aus Bildschirm heraus) => normal: dieser Vektor im Weltkoordinatensystem
		#print("normal: {}".format(normal))
		normal *= self._dist                 # normal ist jetzt so lang wie Abstand zw. Kamera und Target
		#print("normal: {}".format(normal))
		if self._region3d.is_perspective:
			self._target = camPos - normal
		else:
			pass
			self._target = camPos + normal    # im Orthomodus sind eigenartigerweise alle Kamerakoordinaten negativ, wenn sie sich im Blender-View im positiven Oktanten befindet?!
		#print("_target2: {}".format(self._target))
		camPos.w = 1

	#------------------ ------------------- ------------------- -----------------

	# Wir rotieren und verschieben die Kamera in ihrem eigenen Koordinatensystem,
	# indem eine entsprechende Matrix erzeugt wird (resMat). Diese wird dann
	# mittels der inversen View-Matrix in Weltkoordinaten transformiert, die
	# Invertierung davon ist dann die neue View-Matrix.
	# Auf Blenders View-Matrix kann mittels
	# bpy.data.window_managers[0].windows[0].screen.areas[2].spaces[0].region_3d.view_matrix
	# oder besser mittels
	# bpy.context.screen.areas[x].spaces[0].region_3d.view_matrix
	# zugegriffen werden. Diese Matrix wandelt einen Punkt P, der in
	# Weltkoordinaten vorliegt, gemaess
	#  P' = view_matrix @ P
	# in denselben Punkt im Kamerasystem um.
	# Umgekehrt wandelt
	#  P = view_matrix.inverted() @ P'
	# die Koordinaten eines Punktes P' in Weltkoordinaten zurueck.
	# \param x                    x-Translation im Kamerakoordinatensystem
	# \param y                    y-Translation im Kamerakoordinatensystem
	# \param z                    z-Translation im Kamerakoordinatensystem
	# \param a                    a-Translation im Kamerakoordinatensystem
	# \param b                    b-Translation im Kamerakoordinatensystem
	# \param c                    c-Translation im Kamerakoordinatensystem
	# \param targetPntInWldSystem Rotationszentrum im Weltkoordinatensysem
	# \param isToHover            true, wenn im Schwebemodus
	# \param isHorizonLocked      true, wenn Horizont fix
	# fse, 02.12.14
	# fse, 03.03.15: Bug 840 ("Wechsel von Ortho. zu Persp. View fuehrt zu
	#                extremem Heranzoomen des Objekts"): Was hier passiert, ist
	#                mir nicht klar. Jedenfalls behebt die untige Aenderung
	#                tempMat[2][3] = -self._region3d.view_distance (statt = 0)
	#                das Problem. Allerdings darf dann im Ortho-Modus
	#                self.isSlowDownMode nicht gesetzt sein, weil die Bewegung
	#                dann viel zu langsam wird. Daher wird in transformView3d()
	#                die Verlangsamung jetzt nur noch aktiviert, wenn wir im
	#                Perspektivenmodus sind (ist auch logisch, im Ortho-Modus
	#                gibt es ja eigentlich keine Entfernung zur Kamera).
	# fse, 29.11.20: Matrix-Multiplikationsoperator "*" durch "@" ersetzt.

	def moveCam(self, area , x, y, z, a, b, c, targetPntInWldSystem, isToHover, isHorizonLocked):
		# evtl. Target-Punkt patchen:
		if self._blenderPlugin.rotCenter != 'DEFAULT':
			self.patchTargetPoint(area, targetPntInWldSystem)
		# Target-Punkt in Kamerakoordinaten umrechnen:
		targetPntInCamSystem = self._viewMat @ targetPntInWldSystem
		negTargetPntInCamSystem = -targetPntInCamSystem
		negTargetPntInCamSystem.w = 1
		# Translationsmatrizen mit Target und negativem Target erzeugen
		# (das Target liegt im Kamerasystem immer auf der z-Achse mit Abstand
		# view_distance):
		trnTargetMat    = mathutils.Matrix.Translation(   targetPntInCamSystem)
		trnNegTargetMat = mathutils.Matrix.Translation(negTargetPntInCamSystem)
		# Translationsvektor erzeugen:
		trnVecInCamSystem = mathutils.Vector((x, y, z, 0))
		if isToHover:
			# Im Schwebemodus muss Blenders z-Koordinate in der Welt unveraendert bleiben:
			trnVecInWldSystem = self._viewMatInv @ trnVecInCamSystem # Translation in Weltkoordinaten transformieren
			trnVecInWldSystem.z = 0                                  # dort die z-Komponente auf 0 setzen
			trnVecInCamSystem = self._viewMat @ trnVecInWldSystem    # und zurueck ins Kamerasystem transformieren
			#trnVecInWldSystem.resize_3d()
			#self._region3d.view_location += trnVecInWldSystem
		trnVecInCamSystem.w = 1
		if isHorizonLocked:
			# Im Fix-Horizontmodus darf in Weltkoordinaten nur noch um Blenders z-Achse rotiert werden:
			rotVecInCamSystem = mathutils.Vector((a, b, c, 0))       # Kappendaten in Rotationsvektor uebertragen
			rotVecInWldSystem = self._viewMatInv @ rotVecInCamSystem # diesen in Weltkoordinaten transformieren
			rotVecInWldSystem.x = 0                                  # nur Blenders
			rotVecInWldSystem.y = 0                                  # z-Achse erhalten
			rotVecInCamSystem = self._viewMat @ rotVecInWldSystem    # zurueck in Kamerasystem transformieren
			#a = rotVecInCamSystem.x # um unsere x-Achse der Kamera darf man aber ruhig kippen, daher a unveraendert lassen
			b = rotVecInCamSystem.y
			c = rotVecInCamSystem.z
		# drei Rotationsmatrizen und eine Translationsmatrix mit Kappendaten erzeugen:
		rotAMat = mathutils.Matrix.Rotation(a, 4, 'X')
		rotBMat = mathutils.Matrix.Rotation(b, 4, 'Y')
		rotCMat = mathutils.Matrix.Rotation(c, 4, 'Z')
		trnMat = mathutils.Matrix.Translation(trnVecInCamSystem)
		# Verknuepfen, in Weltkoordinaten transformieren und zurueckinvertieren;
		# zunaechst wird das System in den Rotationspunkt verschoben, dann die
		# Rotationen durchgefuehrt, dann das System zurueckgeschoben und
		# schliesslich die Translation gemacht:
		resMat = trnTargetMat @ rotAMat @ rotBMat @ rotCMat @ trnNegTargetMat @ trnMat
		tempMat = self._viewMatInv @ resMat
		tempMat.invert()
		if self._region3d.is_perspective:
			self._region3d.view_distance = self._dist # damit auch das Mausrad weiterhin wie gewohnt zoomt; sonst stimmt die automatisch angepasste Empfindlichkeit nicht mehr
		else:
			self._region3d.view_distance += z # im Ortho-Modus muss man eigenartigerweise die distance aendern, nicht den Zoom-Faktor
			#print("view_distance: {}".format(self._region3d.view_distance))
#			tempMat[2][3] = 0 # seltsam: dieses Element waechst sonst ins Unendliche, und die Szene wird unsichtbar
			tempMat[2][3] = -self._region3d.view_distance # Bug 840: diese Zuweisung statt 0 behebt das Problem
		self._region3d.view_matrix = tempMat

	#-------------- --------------- --------------- --------------- -------------
	
	# Fuer den Fall, dass der Anwender ein festes Rotationszentrum eingestellt
	# hat, wird dieses hier eingestellt.
	# \param  area die aktive Area
	# \retval targetPntInWldSystem das neue Rotationszentrum
	# fse, 12.12.20: area.spaces.active.cursor_location durch
	#                bpy.context.scene.cursor.location ersetzt.
	
	def patchTargetPoint(self, area, targetPntInWldSystem):
		if self._blenderPlugin.rotCenter == 'CURSOR':
			targetPntInWldSystem[0:3] = bpy.context.scene.cursor.location
		elif self._blenderPlugin.rotCenter == 'SELECTION':
			isSomethingSelected = len(bpy.context.selected_objects) > 0
			if isSomethingSelected:
				temp = mathutils.Vector((0,0,0))
				for obj in bpy.context.selected_objects:
					temp +=  obj.location
				temp /= len(bpy.context.selected_objects)
				targetPntInWldSystem[0:3] = temp

	#------------------------- ------------------------- ------------------------
	
	# Bewegt die Szene in allen unterstuetzten 2D-Fenstern.
	# Die diversen Editoren unterscheiden sich hinsichtlich der Operatoren, mit
	# denen man in den Fenstern pannen und zoomen kann. Die Operatoren sind
	# context-abhaengig und eigentlich dazu vorgesehen, aus einem Menue des
	# entsprechenden Editors aufgerufen zu werden. Daher muessen wir den context
	# passend zum Area-Typ einstellen.
	# \param x   x-Translation im Kamerakoordinatensystem
	# \param y   y-Translation im Kamerakoordinatensystem
	# \param z   z-Translation im Kamerakoordinatensystem
	# fse, 16.12.14
	# fse, 23.11.16: Der Clip-Editor hat die drei Untermodi "CLIP", "GRAPH" und
	#                "DOPESHEET", in die man mit den quadratischen Icons im
	#                Header umschalten kann. Die letzten beiden Editoren sehen
	#                genauso aus wie der Graph- bzw. der Dopesheet-Editor. Die im
	#                Clip-Editor verwendeten Funktionen bpy.ops.clip.view_pan()
	#                und bpy.ops.clip.view_zoom() funktionieren aber nur im
	#                "CLIP"-Untermodus und fuehren in den anderen beiden zum
	#                Absturz, so dass man im Clip-Editor mit
	#                   subType = area.spaces[0].view 
	#                pruefen muss, ob man im "CLIP"-Untermodus ist.
	#                Leider reicht es offenbar nicht, einfach den Typ zu patchen,
	#                damit man auch in den andern beiden Untermodi pannen und
	#                zoomen kann; es bewegt sich leider nichts. Aber wenigstens
	#                stuerzt das Skript nicht mehr ab.
	
	def transformView2d(self, area, x, y, z):
		override = self.setContextOverride(area)
		areaType = area.type                # type ist erst einmal area.type
		if areaType == 'CLIP_EDITOR':
			subType = area.spaces[0].view
#			print("BlenderProxy.transformView2d(): subType: {}".format(subType))
			if subType == 'CLIP':
				areaType == 'CLIP_EDITOR'     # Zuweisung ueberfluessig, aber aus Symmetriegruenden ...
			elif subType == 'GRAPH':
				areaType = 'GRAPH_EDITOR'     # type patchen; funktioniert zwar nicht, so ist aber die Idee dokumentiert
			elif subType == 'DOPESHEET':
				areaType = 'DOPESHEET_EDITOR' # type patchen; funktioniert zwar nicht, so ist aber die Idee dokumentiert
			else:
				log(gCf(), self, "Unknown subType '{}' detected!".format(subType))
				pass
		if areaType == 'CLIP_EDITOR':
			ofs = ( -x/50, -y/50 )
			z = 1 + z/20000;
			bpy.ops.clip.view_pan(override, offset=ofs)
			bpy.ops.clip.view_zoom(override, factor=z)
		elif areaType == 'NODE_EDITOR'      or areaType == 'LOGIC_EDITOR' \
		  or areaType == 'SEQUENCE_EDITOR'  or areaType == 'NLA_EDITOR'   \
		  or areaType == 'DOPESHEET_EDITOR' or areaType == 'GRAPH_EDITOR' \
		  or areaType == 'TIMELINE':
			bpy.ops.view2d.pan(override, deltax=-x/50, deltay=-y/50)
			bpy.ops.view2d.zoom(override, deltax=z/200, deltay=z/200)
		elif areaType == 'OUTLINER':
			if y != 0:
				isUp = y > 0
				bpy.ops.outliner.scroll_page(override, up = isUp)
		elif areaType == 'IMAGE_EDITOR':
			ofs = ( -x/50, -y/50 )
			z = 1 + z/20000;
			bpy.ops.image.view_pan(override, offset=ofs)
			bpy.ops.image.view_zoom(override, factor=z)
		else:
			pass
	
	#----------------------------------------------------------------------------

	def test(self):
		log(gCf(), self, "Invoked.")
		self._blenderPlugin.test()

	#----------------------------------------------------------------------------
	
	# Loescht alle Referenzen auf das BlenderPlugin-Objekt. Sonst werden die
	# Destruktoren nicht aufgerufen.
	# fse, 09.01.15

	def clean(self):
		log(gCf(), self, "Invoked.")
		self.publishBlenderPluginObject(None)
#		self._sharedData = None
		del self._blenderPlugin

#-------------------------------------------------------------------------------

# Handler fuer das bpy.app.handlers.load_pre-Ereignis; beendet unseren Timer.
# Beim Laden einer Datei ruft Blender automatisch und ohne dass man es
# verhindern koennte die cancel()-Funktion unseres ModalBlenderTimerOp auf.
# In Folge funktioniert dann nichts mehr.
# Der Vorschlag des Blender-Supports, keinen Timer zu verwenden, sondern
# stattdessen einen Hook zu verwenden, scheiterte daran, dass man in den
# Handlern fuer die bpy.app.handlers.scene_update_post-Ereignisse anscheinend
# keine Operatoren aufrufen kann. Einen solchen haetten wir z. B. gebraucht, um
# die Mauskoordinaten zu erfragen. Das fuehrt aber zum Absturz von Blender, weil
# irgendeine Rekursionstiefe zu hoch wird.
# Daher verwenden wir jetzt Handler fuer die Ereignisse
# bpy.app.handlers.load_pre und bpy.app.handlers.load_post. Beim ersten beenden
# wir den Timer regulaer und starten ihn wieder beim zweiten.
# fse, 19.12.14
# fse, 19.01.15: Timer beenden reicht nicht, auch der Thread muss beendet
#                werden weil Blender sonst bei File/New... abstuerzt.
#                Daher wird jetzt mit haltScBlenderPlugin() tabula rasa gemacht,
#                wobei anschliessend der Hook fuer das load_post-Ereignis aber
#                wieder eingerichtet werden muss.

@persistent
def scLoadPreHandler(dummy):
	haltScBlenderPlugin()
	# haltScBlenderPlugin() hat den Hook entfernt => neu einrichten:
	bpy.app.handlers.load_post.append(scLoadPostHandler)
	log(gCf(), None, "Stopped.")
	
#-------------------------------------------------------------------------------

# Handler fuer das bpy.app.handlers.load_post-Ereignis; startet unseren Timer
# wieder. S. Kommentar bei scLoadPreHandler().
# fse, 19.12.14
# fse, 19.01.15: Startet das ganze Plug-in wieder neu, nur Timer hat nicht
#                gereicht.

@persistent
def scLoadPostHandler(dummy):
	global gIsUnregisterMenuEntry
	runScBlenderPlugin(gIsUnregisterMenuEntry)
	log(gCf(), None, "Started.")
	
#-------------------------------------------------------------------------------

# Handler fuer das Scene-Update-Post-Ereignis. Wird nur einmal gebraucht, um
# das Add-On verzoegert direkt beim Registrieren zu starten. Siehe Kommentar 
# bei BlenderPlugin.startPluginFromInit().
# fse, 12.01.15
# fse, 13.12.20: Obsolet; das Plug-in wird jetzt direkt gestartet.

@persistent
def scSceneUpdatePostHandler(dummy):
	global gBlenderPlugin
	gBlenderPlugin.removeSceneUpdatePostHandler()
	gBlenderPlugin.startPlugin()
	log(gCf(), None, "Started.")
	
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

class BlenderPlugin:
	_name                = "BlenderPlugin"
	_version             = ""
#	_sharedStdDataList   = [] # Jedes neue Device haengt hier ein neues Element an. Nach devDataList ...
	devDataList          = [] # ... umbenannt, damit leichterer Zugriff von Game-Engine aus moeglich.
	                          # Achtung: In den anderen Klassen heisst die Liste immer noch _sharedStdDataList.
	
	# diverse Pfade:
	_SPACECONTROL_DIR = "SpaceController"
	_APPLS_DIR        = "Applications"
	if sys.platform == "win32":
		_USER_HOME = os.environ['USERPROFILE']
	else:
		_USER_HOME = os.environ['HOME']
	_APPLS_DIR_PATH   = _USER_HOME + os.path.sep + _SPACECONTROL_DIR + os.path.sep + _APPLS_DIR
	_INI_FILE_NAME    = "Blender.ini"
	_INI_FILE_PATH    = _APPLS_DIR_PATH + os.path.sep + _INI_FILE_NAME
	_MNU_FILE_NAME    = "Blender.mnu"
	_MNU_FILE_PATH    = _APPLS_DIR_PATH + os.path.sep + _MNU_FILE_NAME

	# Move-Modes, Kopien aus ScAdvancecSettings.java:
	OBJ_SCREEN              = 0
	OBJ_ORIGIN              = 1
	CAM_STANDARD            = 2
	CAM_TARGET              = 3
	
	# Daemon-Events, Kopien aus ScDllWrapper.java:
	NOTHING_CHANGED         = -1
	ADV_SETTINGS_CHANGED    = 0x20001 # = 131073
	DEV_PARS_CHANGED        = 0x20002 # = 131074
	DEV_FRONT               = 0x20009
	DEV_RIGHT               = 0x2000a
	DEV_TOP                 = 0x2000b
	DEV_FIT                 = 0x2000c
	EVT_APPL_IN_FRGRND      = 0x20013
	DEV_BACK                = 0x20016;
	DEV_LEFT                = 0x20017;
	DEV_BOTTOM              = 0x20018;
	APPL_FUNC_START         = 0x20020;
	
	scDll  = None
	devIdx = 0 # ctypes.c_int(0)
	x      = 0 # ctypes.c_short(0)
	y      = 0 # ctypes.c_short(0)
	z      = 0 # ctypes.c_short(0)
	a      = 0 # ctypes.c_short(0)
	b      = 0 # ctypes.c_short(0)
	c      = 0 # ctypes.c_short(0)
	traLmh = 0 # ctypes.c_int(0)
	rotLmh = 0 # ctypes.c_int(0)
	event  = 0 # ctypes.c_int(NOTHING_CHANGED)
	tvSec  = 0 # ctypes.c_long(0)
	tvUsec = 0 # ctypes.c_long(0)
	
#	moveMode             = OBJ_SCREEN
	isToHover            = False
	isHorizonLocked      = False
	isTimerKilled        = False
	isTimerRunning       = False

	_isUnregisterMenuEntry = True # true, wenn der "Unregister"-Menue-Eintrag vorhanden sein soll
	daemonProxy  = None
	blenderProxy = None

	# Save-Parameter:
	timerInterval       = 0.02
	transSensBase       = 10000
	rotSensBase         = 40000
	rotCenter           = 'DEFAULT' # 'CURSOR', 'SELECTION'
	isSlowDownMode      = True
	isOnly3dViewActive  = False
	isAutoSendGap       = False
	isRememberView      = True
	isObjectMoveMode    = False
	isOmmMultiRotCenter = False
	isGenKeyframes      = False    

	#----------------------------------------------------------------------------

	def __init__(self, isUnregisterMenuEntry):
		log(gCf(), self, "Constructing ...")
		self._version = self.makeVersion()
		self.saveIniFile(True)
		self.loadIniFile()
		self.daemonProxy  = DaemonProxy(self, self.devDataList)
		self.blenderProxy = BlenderProxy(self, self.devDataList)
		self._isUnregisterMenuEntry = isUnregisterMenuEntry
		# DLL-Objekt definieren:
		if sys.platform == "win32":
			# Windows:
			if platform.architecture()[0] == "32bit":
				self.scDll = ctypes.CDLL("spc_ctrlr_32.dll")
			else:
				self.scDll = ctypes.CDLL("spc_ctrlr_64.dll") # mit 64-Bit-Blender funktioniert auch WinDLL statt CDLL
		elif sys.platform == "darwin":
			# OS X:
			self.scDll = ctypes.CDLL("/Applications/SpaceControl/lib/libspc_ctrl_64.dylib")
		else:
			# Linux:
#			libPath = "<installDir>"
#			libPath = "/opt/SpaceControl/lib"
#			libPath = "/home/fs/opt/SpaceControl/lib"
#			if platform.architecture()[0] == "32bit":
#				lib = libPath + "/libspc_ctrl_32.so"
#			else:
#				lib = libPath + "/libspc_ctrl_64.so"
#			self.scDll = ctypes.CDLL(lib)
			self.scDll = ctypes.CDLL("/opt/SpaceControl/lib/libspc_ctrl.so")
#			self.scDll = ctypes.CDLL("libspc_ctrl.so") # <= funktioniert unter Debian nicht
		log(gCf(), self, "Constructed.")
	
	#-------------------------------------- -------------------------------------
	
	# Erzeugt aus bl_info["version"] einen Versions-String.
	# fse, 04.02.15
	
	def makeVersion(self):
		v = bl_info["version"]
		return str(v[0]) + "." + str(v[1]) + "." + str(v[2])
	
	#-------------------------------------- -------------------------------------
	
	# Speichert die Datei "blender.ini".
	# fse, 31.01.15 
	
	def saveIniFile(self, isToCheckForExistence):
		iniFilePath = self._INI_FILE_PATH
		if isToCheckForExistence and os.path.exists(iniFilePath):
			return
		
		log(gCf(), self, "Saving file '{}'.".format(iniFilePath))
		w = 20
		lines = []
		lines.append("# Configuration file for SpaceControl's Blender Plug-in")
		lines.append("")
		lines.append("[PROGRAM]")
		lines.append("\tname                = " +     self._name               .ljust(w) + "# plug-in's name")
		lines.append("\tversion             = " +     self._version            .ljust(w) + "# plug-in's version number")
		lines.append("\ttimerInterval       = " + str(self.timerInterval      ).ljust(w) + "# interval for getting data from the SpaceController")
		lines.append("\ttransSensBase       = " + str(self.transSensBase      ).ljust(w) + "# base sensitivity divisor for translation")
		lines.append("\trotSensBase         = " + str(self.rotSensBase        ).ljust(w) + "# base sensitivity divisor for rotation")
		lines.append("\trotCenter           = " +     self.rotCenter           .ljust(w) + "# rotation center ('DEFAULT', 'CURSOR', 'SELECTION')")
		lines.append("\tisSlowDownMode      = " + str(self.isSlowDownMode     ).ljust(w) + "# if True sensitivity is decreased near the rotation center and increased far from it")
		lines.append("\tisOnly3dViewActive  = " + str(self.isOnly3dViewActive ).ljust(w) + "# if True the device will affect only 3D View areas")
		lines.append("\tisAutoSendGap       = " + str(self.isAutoSendGap      ).ljust(w) + "# if True the plug-in adjusts the device's send gap automatically to avoid lagging")
		lines.append("\tisRememberView      = " + str(self.isRememberView     ).ljust(w) + "# if True the plug-in remembers the active (split) view when a selection is bound to a device")
		lines.append("\tisObjectMoveMode    = " + str(self.isObjectMoveMode   ).ljust(w) + "# if True the controller moves selected objects instead of the whole scene")
		lines.append("\tisOmmMultiRotCenter = " + str(self.isOmmMultiRotCenter).ljust(w) + "# if True each selected object rotates around its own center in multiselections")
		lines.append("\tisGenKeyframes      = " + str(self.isGenKeyframes     ).ljust(w) + "# if True keyframes are automatically created while moving")
		f = open(iniFilePath, 'w')
		for line in lines:
			f.write(line + "\n")
		f.close()

	#-------------------------------------- -------------------------------------

	# Laedt die Datei "blender.ini".
	# fse, 31.01.15
	
	def loadIniFile(self):
		iniFilePath = self._INI_FILE_PATH
		log(gCf(), self, "Loading file '{}'.".format(iniFilePath))
		iniFilePath = self._INI_FILE_PATH
		lines = []
		f = open(iniFilePath, 'r')
		for line in f:
			lines.append(line)
		f.close()
		self._name               =       self.getParamVal(lines, "name"               )
		self._version            =       self.getParamVal(lines, "version"            )
		self.timerInterval       = float(self.getParamVal(lines, "timerInterval"      ))
		self.transSensBase       = int  (self.getParamVal(lines, "transSensBase"      ))
		self.rotSensBase         = int  (self.getParamVal(lines, "rotSensBase"        ))
		self.rotCenter           =       self.getParamVal(lines, "rotCenter"          )
		self.isSlowDownMode      =       self.getParamVal(lines, "isSlowDownMode"     ) == "True"
		self.isOnly3dViewActive  =       self.getParamVal(lines, "isOnly3dViewActive" ) == "True"
		self.isAutoSendGap       =       self.getParamVal(lines, "isAutoSendGap"      ) == "True"
		self.isRememberView      =       self.getParamVal(lines, "isRememberView"     ) == "True"
		self.isObjectMoveMode    =       self.getParamVal(lines, "isObjectMoveMode"   ) == "True"
		self.isOmmMultiRotCenter =       self.getParamVal(lines, "isOmmMultiRotCenter") == "True"
		self.isGenKeyframes      =       self.getParamVal(lines, "isGenKeyframes"     ) == "True"
		self.isObjectMoveMode = False # zu riskant, den schalten wir beim Start besser immer ab
	
	#------------------------- ------------------------- ------------------------

	# Durchlaeuft alle Elemente der Liste lines und sucht darin per Regex nach
	# Parameternamen param.
	# Gesucht wird nach beliebig vielen Leerzeichen, dann param, dann wieder
	# beliebig viele Leerzeichen, dann "=", dann wieder beliebig viele
	# Leerzeichen, dann alle Zeichen bis zum, aber ohne "#". Letztere
	# enthalten den Parameterwert; sie werden zu einer Gruppe zusammengefasst,
	# endstaendige Leerzeichen entfernt und als String zurueckgeliefert.
	# \param lines Liste mit allen Zeilen der "blender.ini"
	# \param param der gesuchte Parametername
	# \return der Wert des Parameters als String
	# fse, 31.01.15
	
	def getParamVal(self, lines, param):
		rxpStr = r'\s*(' + param + r')\s*=\s*([^#]*)'
		rxp = re.compile(rxpStr)
		for line in lines:
			res = rxp.match(line)
			if res != None:
				#print("BlenderPlugin.getParamVal(): " + res.group(1) + ": <" + res.group(2).rstrip() + ">")
				return res.group(2).rstrip()
	
	#----------------------------------------------------------------------------

	def __del__(self):
		log(gCf(), self, "Destructing ...")
		self.saveIniFile(False)
		del self.blenderProxy
		del self.daemonProxy
		log(gCf(), self, "Destructed.")
		
	#----------------------------------------------------------------------------
	
	# Wird vom Konstruktor eines DeviceProxys aufgerufen und macht dem System den
	# Speicherbereich bekannt, in dem der BlenderProxy mittels getAndHandleData()
	# die Kappendaten abholen kann.
	# Ausserdem wird das SpaceController-Menue entsprechend angepasst.
	# fse, 03.08.17
	
	def addDevice(self, stdData):
		self.devDataList.append(stdData)
		self.blenderProxy.addDevice(stdData)

	#----------------------------------------------------------------------------
	
	# Sucht in der Liste der Speicherbereiche fuer die Kappendaten der
	# Device-Proxies nach dem Element mit dem Device-Index devIdx und loescht
	# dieses Element, falls es gefunden wird.
	# Ausserdem wird das SpaceController-Menue entsprechend angepasst.
	# fse, 03.08.17
	
	def removeDevice(self, stdData):
		self.devDataList.remove(stdData)
		self.blenderProxy.removeDevice(stdData)
	
	#----------------------------------------------------------------------------
	
	# Gibt zu Kontrollzwecken die Liste der Standard-Kappendaten-Elemente aus.
	# Jeder Device-Proxy haengt dort seinen Kappendaten-Speicherbereich an.
	# fse, 03.08.17
	
	def logStdDataList(self, currentFrame):
		log(currentFrame, self, "Elements in BlenderPlugin.devDataList[]: -----")
		for stdData in self.devDataList:
			log(currentFrame, self, "devIdx: " + str(stdData.devIdx) + ", serNo: " + str(stdData.serNo) + ", isDtaNull: " + str(stdData.isDtaNull) + ", isHandled: " + str(stdData.isHandled))
			pass
		log(currentFrame, self, "                                         ^^^^^")
			
	#----------------------------------------------------------------------------

	def printStdData(self):
		currentTime = datetime.datetime.now().time()
		if self.x != 0 or self.y != 0 or self.z != 0 or self.a != 0 or self.b != 0 or self.c != 0 or self.event != -1:
			print(currentTime.isoformat(), ": ", self.x, self.y, self.z, self.a, self.b, self.c, self.event)
			pass

	#----------------------------------------------------------------------------

	def test(self):
		log(gCf(), self, "BlenderPlugin.test(): Invoked.")
		pass
		
	#----------------------------------------------------------------------------
	
	# Liefert die Anzahl der angeschlossenen Geraete zurueck.
	# fse, 02.08.17
	
	def getDevNum(self):
		return len(self.daemonProxy._devDic)
	
	#----------------------------------------------------------------------------
	
	# Setzt im Device-Proxy mit der Seriennummer serNo die Liste der Objekte,
	# die das Geraet per Move-Mode bewegen soll.
	# fse, 02.08.17
	
	def setSelectionList(self, serNo, selectionList):
		lastRegion = self.blenderProxy.getLastRegion()
		self.daemonProxy.setDevProxySelectionListAndRegion(serNo, selectionList, lastRegion)
		log(gCf(), self, "lastRegion: " + str(lastRegion))

	#----------------------------------------------------------------------------

	def clean(self):
		self.blenderProxy.clean()
		self.daemonProxy.clean()
		pass

	def getIsThreadRunning(self):
		return self.daemonProxy.getIsRunning()

	def getIsTimerRunning(self):
		return self.isTimerRunning

	def startPlugin(self):
		log(gCf(), self, "Invoked.")
		if self.daemonProxy != None:
			self.daemonProxy.startDaemonProxy()
		if self.blenderProxy != None:
			self.blenderProxy.startBlenderProxy()

	def stopPlugin(self):
		log(gCf(), self, "----------------------------------------------")
		log(gCf(), self, "Stopping Plugin ...")
		self.setBlenderMode("unknown")
		cnt = 0
		while self.daemonProxy.getIsDeviceListModifying() and cnt < 10:
			cnt += 1
			log(gCf(), self, "Waiting for modifying device list to finish ({})...".format(cnt))
			time.sleep(0.3)
		self.saveIniFile(False)
		self.blenderProxy.stop()
		self.daemonProxy.stop()
		log(gCf(), self, "Plugin stopped.")

# fse, 13.12.20: Obsolet; das Plug-in wird jetzt direkt gestartet.
	def removeSceneUpdatePostHandler(self):
		self.blenderProxy.removeSceneUpdatePostHandler()
		
	def toggleReduceSensMode(self):
		self.blenderProxy.toggleReduceSensMode()
		
	def toggle3dViewMode(self):
		self.blenderProxy.toggle3dViewMode()
		
	def toggleAutoSendGap(self):
		self.isAutoSendGap = not self.isAutoSendGap
		if self.isAutoSendGap:
			self.daemonProxy.triggerSendDelayAdjustForAllDevices()

	def toggleRememberView(self):
		self.blenderProxy.toggleRememberView()

	def toggleObjectMoveMode(self):
		self.blenderProxy.toggleObjectMoveMode()

	def toggleMultiRotCentersMmm(self):
		self.blenderProxy.toggleMultiRotCentersMmm()

	def toggleGenKeyframesOmm(self):
		self.blenderProxy.toggleGenKeyframesOmm()

	def setMoveMode(self, moveMode):
		self.blenderProxy.setMoveMode(moveMode)

	def setIsAutoFit(self, isAutoFit):
		self.blenderProxy.setIsAutoFit(isAutoFit)

	def setSpecialKeyEvent(self, event):
		self.blenderProxy.setSpecialKeyEvent(event)
		
	def setRotCenterToDefault(self):
		self.blenderProxy.setRotCenterToDefault()

	def setRotCenterToCursor(self):
		self.blenderProxy.setRotCenterToCursor()

	def setRotCenterToSelection(self):
		self.blenderProxy.setRotCenterToSelection()

	def setBlenderMode(self, modus):
		if self.daemonProxy != None:
			self.daemonProxy.setBlenderMode(modus)
		
	def setIsTimerRunning(self, isTimerRunning):
		self.isTimerRunning = isTimerRunning
		
	def getIsAutoSendGap(self):
		return self.isAutoSendGap

	def getIsRememberView(self):
		return self.isRememberView

	def getIsObjectMoveMode(self):
		return self.isObjectMoveMode

	def getIsGenKeyframes(self):
		return self.isGenKeyframes

	def setTimer(self, timer):
		self.blenderProxy.setTimer(timer)
	
	def getTimer(self):
		return self.blenderProxy.getTimer()

	def setIsUnregisterMenuEntry(self, isThere):
		self._isUnregisterMenuEntry = isThere
		
	def getIsUnregisterMenuEntry(self):
		return self._isUnregisterMenuEntry
	
	def getMnuFilePath(self):
		return self._MNU_FILE_PATH

	#----------------------------------------------------------------------------
	
	# Wenn man Blender bei noch laufendem Plug-in schliesst, wird
	# unregister() aus __init__.py nicht aufgerufen. In Folge laeuft der
	# DaemonProxy-Thread weiter, und Blender wird nicht richtig beendet, d. h.
	# das Konsolenfenster bleibt offen (falls offen) und "blender.exe" laeuft im
	# Taskmanager weiter - also Absturz. Ideal waere ein Hook, der beim
	# Schliessen von Blender aufgerufen wird; den gibt's aber nicht.
	# Man beobachtet aber, dass unser ModalBlenderTimerOp-Timer durchaus von
	# Blender beim Schliessen gecancelt wird. Damit hat man so eine Art Hook,
	# indem die dortige cancel()-Funktion in diesem Fall diese Funktion hier
	# aufruft und das Plug-in beendet wird.
	# fse, 13.01.14

	def signalTimerKilled(self):
		haltScBlenderPlugin()
		self.isTimerKilled = True
		
	def getIsTimerKilled(self):
		return self.isTimerKilled
		
	def setIsTimerKilled(self, isKilled):
		self.isTimerKilled = isKilled
		
	#----------------------------------------------------------------------------
	
	# Startet das Plug-in indirekt, indem ein Handler fuer das Updaten der Szene
	# eingerichtet wird, der das Plug-in dann startet.
	# Grund:
	# Eigentlich sollte das Plug-in beim Registrieren des Blender-Add-Ons
	# in __init__.py gestartet werden. Leider wird man mit einer Fehlermeldung
	# "AttributeError: '_RestrictContext' object has no attribute 'scene'" beim
	# Start des BlenderProxy-Timers konfrontiert. Laut 
	# http://wiki.blender.org/index.php/Extensions:2.6/Py/API_Changes gehoert
	# das wohl so, weil irgendein Context beim Start noch nicht zur Verfuegung
	# steht. Daher wird beim Registrieren jetzt ein Handler fuer das
	# Update-Ereignis eingerichtet, der das Plug-in dann mit dieser Verzoegerung
	# startet. Nach einmaligem Aufruf wird der Handler dann sofort wieder
	# entfernt.
	# fse, 12.01.15
	# fse, 13.12.20: Obsolet; das Plug-in wird jetzt direkt gestartet.
	
	def startPluginFromInit(self):
		log(gCf(), None, "startPluginFromInit() invoked.")
		self.blenderProxy.addSceneUpdatePostHandler()

#-------------------------------------------------------------------------------

# Startet das Blender-Plug-in, wird von __init__.py aufgerufen.
# fse, 09.01.15
# fse, 13.12.20: Das Plug-in kann jetzt direkt gestartet werden, Hook nicht
#                mehr noetig.

def runScBlenderPlugin(isUnregisterMenuEntry):
	log(None, None, None)
	log(gCf(), None, "runScBlenderPlugin() invoked.")
	global gBlenderPlugin
	global gIsUnregisterMenuEntry
	gIsUnregisterMenuEntry = isUnregisterMenuEntry
	gBlenderPlugin = BlenderPlugin(isUnregisterMenuEntry)
	# Funktioniert leider nicht, weil sonst Error
	# "AttributeError: '_RestrictContext' object has no attribute 'scene'" beim
	# Start des BlenderProxy-Timers (geht in Blender 2.8 jetzt doch):
	#gBlenderPlugin.startPlugin()
	# Daher etwas umstaendlicher ueber Hook starten:
	gBlenderPlugin.startPluginFromInit()

#-------------------------------------------------------------------------------

# Beendet das Blender-Plug-in, wird von __init__.py aufgerufen.
# fse, 09.01.15

def haltScBlenderPlugin():
	log(gCf(), None, "haltScBlenderPlugin() invoked.  --------------")
	global gBlenderPlugin
	gBlenderPlugin.stopPlugin()
	#gBlenderPlugin.unregisterBlenderProxy()
	gBlenderPlugin.clean()
	del gBlenderPlugin
	log(gCf(), None, "haltScBlenderPlugin() finished. --------------")

#-------------------------------------------------------------------------------

def register():
	log(gCf(), None, "SpaceController Blender add-on registering ...")
	runScBlenderPlugin(False)
	log(gCf(), None, "SpaceController Blender add-on registered.")
 
def unregister():
	log(gCf(), None, "SpaceController Blender add-on unregistering ...")
	haltScBlenderPlugin()
	log(gCf(), None, "SpaceController Blender add-on unregistered")
 
if __name__ == "__main__":
	register()

#-------------------------------------------------------------------------------
#--------------------------------------- ---------------------------------------
#-------------------------- ------------------------- --------------------------
#------------------- ------------------- ------------------- -------------------
#--------------- --------------- --------------- --------------- ---------------

	#----------------------------------------------------------------------------
	#-------------------------------------- -------------------------------------
	#------------------------- ------------------------- ------------------------
	#------------------ ------------------- ------------------- -----------------
	#-------------- --------------- --------------- --------------- -------------
	#----------- ------------- ------------ ------------ ------------- ----------
	#--------- -------- --------- --------- --------- --------- -------- --------
